
/*******************************************************************************
 * Copyright (c) 2003 University Corporation for Advanced Internet
 * Development (UCAID).
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all software files.
 *
 * IN NO EVENT SHALL UCAID BE LIABLE TO ANY PARTY FOR * DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE
 * OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF UCAID HAS BEEN ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * UCAID SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND UCAID HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *****************************************************************************/

/****************************************************************************
 * This file contains the routines necessary to print actions allowed by Akenti 
 * for accessing a resource. It also prints constraints that are of type EXT_AUTH
 * that have not been evaluated by Akenti code, and must be evaluated by some
 * external library. 
 * This file is now a part of libPolicyEngine that must be linked with mod_akenti
 *
 * author: Sushila R.Subramanian
 * date: 23 Feb 2003
 ****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string>

#include "AkentiInit.hpp"
#include "AkentiPrincipal.hpp"
#include "AkentiEngine.hpp"
#include "PolicyContext.hpp"
#include "CapabilityCertificate.hpp"
#include "ConditionalActions.hpp"
#include "AkentiMessage.hpp"
#include "ExpressionParser.hpp"
#include "TreeNode.hpp"
#include "Exception.hpp" 
#include "Error.hpp" 

#include "PrintInfo.hpp"

using namespace std;
using namespace Akenti;

typedef PrintInfo PI;

PI::PrintInfo () {
  unConditionalActions = "";
  conditionalInfo = "";
}

/*------------------------------------------------------------------------
 * getInfo: Entry routine from an external function.
 * This calls the routines necessary to pass information on allowed actions
 * and EXT_AUTH constraints - only if checkAccess returns a code of 104 or 105.
 * It sets the variables "unConditionalActions", and "conditionalInfo" to the 
 * appropriate result values. These must be obtained seperately through "get_blah"
 * routines.
 *------------------------------------------------------------------------*/

void
PI::getInfo (bool retval, CapabilityCertificate *cert, AkentiMessage *msg) {  
  
  if (retval) {
    printActions ((*cert)->getUnConditionalActions());
    //printActions ((*cert)->getConditionalActions());
  }
  
#ifdef SRS_DEBUG
  cerr << "Message code is " << msg->getCode() << endl;
#endif /* SRS_DEBUG */
  
  if ((msg->getCode() == 105) || (msg->getCode() == 104)) {
    printConstraints  ((*cert)->getConditionalActions());
  }
  
}

string 
PI::getActions() {
  return unConditionalActions;
}

string
PI::getConditionalInfo() {
  return conditionalInfo;
}


void
PI::printAllowedActions (const vector <string>& allowedactions) {
#ifdef SRS_DEBUG
  cerr << "ACTIONS:";   
#endif /* SRS_DEBUG */
  
  int size = allowedactions.size();
  if (size == 0) {
#ifdef SRS_DEBUG
    cerr << "NONE" << endl;
#endif /* SRS_DEBUG */
    return;
  }
  
  conditionalInfo = conditionalInfo + "(ACTIONS=" + allowedactions[0];
  for (unsigned int i = 1; i < allowedactions.size(); i++) {
#ifdef SRS_DEBUG
    cerr << " "<< allowedactions[i] ;
#endif /* SRS_DEBUG */
    conditionalInfo = conditionalInfo + " " + allowedactions[i];
  }
#ifdef SRS_DEBUG
  cerr << endl;
#endif /* SRS_DEBUG */
  conditionalInfo = conditionalInfo + ")";
}

void
PI::printAttributeInfos(const vector <AttributeInfo>& attrinfos) {
  
  int size = attrinfos.size();
  if (size == 0) {
    return;
  }
  for (unsigned int i = 0; i < attrinfos.size(); ++i) {
    AttributeInfo attrInfo = attrinfos[i];
#ifdef SRS_DEBUG
    cerr << "ATTRIBUTE:" << attrInfo.getAttribute() << endl;
    cerr << "VALUE:" << attrInfo.getValue() << endl; 
#endif /* SRS_DEBUG */
    if (attrInfo.getType() == 0 ) { // TYPE = SYSTEM
#ifdef SRS_DEBUG
      cerr << "TYPE:SYSTEM" << endl;
#endif /* SRS_DEBUG */
    }
    else if (attrInfo.getType() == 3) { // TYPE = EXT_AUTH
#ifdef SRS_DEBUG
      cerr << "TYPE:EXT_AUTH" << endl; 
#endif /* SRS_DEBUG */
      conditionalInfo = conditionalInfo + "(TYPE=EXT_AUTH):";
    }
  }
}

void 
PI::printConstraints(const vector<ConditionalActions>& actions) {
  
  int size = actions.size();
  
  if (size == 0) {
#ifdef SRS_DEBUG
    cerr << "NONE" << endl << "END_CONDITION" << endl;
#endif /* SRS_DEBUG */
    conditionalInfo = "NONE";
    return;
  }
  
  for (unsigned int i = 0; i < actions.size(); ++i) {
    ConditionalActions conAct = actions[i];
    
    // print attributes first
    printAttributeInfos (conAct.getCondition().getAttributeInfos());
    // print if it is a critical set of attributes
    if (conAct.isCritical()) {
#ifdef SRS_DEBUG
      cerr << "CRITICAL:1"<<endl;
#endif /* SRS_DEBUG */
      conditionalInfo = conditionalInfo + "(CRITICAL=1):";
    }
    else {
#ifdef SRS_DEBUG
      cerr << "CRITICAL:0" <<endl;
#endif /* SRS_DEBUG */
      conditionalInfo = conditionalInfo + "(CRITICAL=0):";
    }
    // print allowed actions 
    printAllowedActions  (conAct.getActions());

#ifdef SRS_DEBUG 
    cerr << "CONSTRAINT=" << conAct.getConstraint() << endl;
#endif /* SRS_DEBUG */

    conditionalInfo = conditionalInfo + "(CONSTRAINT=" + conAct.getConstraint() + ")";
  }
  
}


void 
PI::printActions(const vector<string>& actions) {
#ifdef SRS_DEBUG
  cerr << "The actions granted are: ";
#endif /* SRS_DEBUG */

  int size = actions.size();

  if (size == 0) {
#ifdef SRS_DEBUG 
    cerr << "NONE" << endl;
#endif /* SRS_DEBUG */
    unConditionalActions = "NONE";
    return;
  }

#ifdef SRS_DEBUG
  cerr << actions[0];
#endif /* SRS_DEBUG */
  unConditionalActions = actions[0];

  for (unsigned int i = 1; i < actions.size(); ++i) {
#ifdef SRS_DEBUG
    cerr << " " << actions[i];
#endif /* SRS_DEBUG */
    unConditionalActions = unConditionalActions + " " + actions[i];
  }

#ifdef SRS_DEBUG
  cerr << endl;
#endif /* SRS_DEBUG */
}

void 
PI::printActions(const vector<ConditionalActions>& actions) {
#ifdef SRS_DEBUG 
  cerr << "Conditional actions are: ";
#endif /* SRS_DEBUG */

  int size = actions.size();

  if (size == 0) {
    cerr << "NONE" << endl;
    return;
  }

  cerr << actions[0].toLogString();

  for (unsigned int i = 1; i < actions.size(); ++i) {
    cerr << ", " << actions[i].toLogString();
  }

  cerr << endl;
}
