#include <sstream>
#include "Xerces-Includes.hpp"
#include "AkClientInf.h"
#include "cap.h"
#include "SoapHttpRequest.hpp"
#include "SoapHttpResponse.hpp"
#include "AkentiRequests.hpp"
#include "AkentiResponses.hpp"
#include "CheckAccessRequest.hpp"
#include "CheckAccessResponse.hpp"
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "Socket.hpp"
#include "SSLSocket.hpp"
#include "InputStream.hpp"
#include "OutputStream.hpp"
#include "CertificateFactory.hpp"
#include "Certificate.hpp"
#include "AkentiCertificate.hpp"
#include "CapabilityCertificate.hpp"
#include "DistinguishedName.hpp"
#include "AkentiPrincipal.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "XMLUtil.hpp"

using namespace std;
using namespace Akenti;

void 
convertString(const string& in, char **out) {
   if (out) {
      if (*out) {
          free(*out);
      }
 
      *out = strdup(in.c_str());
   }
} 

int 
akentiClientInit(const char *confFile, char **outStatus) {
  string status = "Init XML Library";
  
  try {
    XMLUtil::init();
    status += "OK ";
    Properties& props = Properties::getSystemProperties();

    status += "loading config file ";
    props.load(confFile);
    status += "OK ";

    if (!Provider::isProviderInstalled()) {
      status += "installing security provider ";
      Provider provider = Provider(new OpenSSLProviderImpl);

      provider->loadDigestAlgorithms();
      provider->loadSignatureAlgorithms();
      provider->seedRandomGenerator("WE GOT TO SEED THE RANDOM GENERATOR");

      Provider::setProvider(provider);
      status += "OK ";
      string sslMode = props.getValue("SSLMode");

      if (StringUtil::equalsIgnoreCase(sslMode, "on")) {
	status += "creating SSLcontext ";
	SSLContext clientContext = provider->createSSLContext(SSLConstants::TLS1, true);

	clientContext->setVerify(SSLConstants::VERIFY_PEER);

	// What happens if these are not set?
	clientContext->setCertificateFile(props.getValue("ClientCertFile"), SSLConstants::PEM);
	clientContext->setPrivateKeyFile(props.getValue("ClientKeyFile"), SSLConstants::PEM);
  
	string path = props.getValue("CertificateAuthorityPath");

	if (path.length() != 0) {
	  path += "/";
	}
  
	path += props.getValue("CertificateAuthorityFile");

	clientContext->setTrustedCAS(path);
	SSLContext::setDefaultSSLClientContext(clientContext);
	status += "OK ";
      }
    }
    
    status += "Reading port ";

    string portString = props.getValue("Port");

    if (portString.length()) {
      istringstream is(portString);
      int port = -1;

      is >> port;

      if (port <= 0) {
        throw Exception("Invalid port");
      }
    }

    status += "OK ";
  } catch(Exception& ex) {
    status += "FAIL";
    convertString(status, outStatus);
    return 1;
  } catch(Error& err) {
    status += "FAIL";
    convertString("SHOULD NEVER HAPPEN" + err.getMessage() , outStatus);
    return -1;
  } catch(...) {
    convertString("UNKNOWN ERROR", outStatus);
    return -1;
  }
  
  convertString(status,outStatus);
  return 0;
}

int 
clientCheckAccess(const AkentiPrincipal& ap, 
                  const char* resource, 
                  CapCert *outCap, 
                  AKMessage *outMsg) {
  bool debug = false;
 
  delete(*outCap); 
  *outCap = NULL;

  Properties& props = Properties::getSystemProperties();
  string sslMode = props.getValue("SSLMode");
  bool sslIsOn = StringUtil::equalsIgnoreCase(sslMode, "on");
  istringstream is(sslIsOn? props.getValue("Port","8443") : props.getValue("Port","8444"));
  int port;

  is >> port;

  string host = props.getValue("Host","localhost");
  Socket sock;

  try {
    sock = sslIsOn? SSLSocket(host, port) : Socket(host, port);
    sock.connect();
  } catch(Exception& ex) {
    *outMsg = new AkentiMessage(201, " " + ex.getMessage());
    return 1;
  }

  CheckAccessRequest checkAccessRequest;

  checkAccessRequest.setID(1);
  checkAccessRequest.setResource(resource);
  checkAccessRequest.setAkentiPrincipal(ap);

  AkentiRequests requests;

  requests.addRequest(checkAccessRequest);

  SoapHttpRequest soapRequest;

  soapRequest.setParams(requests.toXML());

  if (debug) {
    ostringstream* os = new ostringstream();
    OutputStream dos(os);

    soapRequest.toStream(dos);
    cout << os->str() << endl;
  }

  soapRequest.toStream(sock.getOutputStream());

  SoapHttpResponse soapResponse;

  soapResponse.fromStream(sock.getInputStream());

  if (debug) {
    ostringstream* os = new ostringstream();
    OutputStream dos(os);

    soapResponse.toStream(dos);
    cout << os->str() << endl;
  }

  if (soapResponse.getFaultCode() != NONE) {
    *outMsg = new AkentiMessage(202, " " + soapResponse.getFaultMessage());
    return -1;
  }

  AkentiResponses responses;

  responses.fromXML(soapResponse.getReturnValue());

  CheckAccessResponse checkAccessResponse = (CheckAccessResponse&) responses.getResponseAt(0);
  const AkentiMessage& mesg = checkAccessResponse.getAkentiMessage();

  *outMsg = new AkentiMessage(mesg);

  CapabilityCertificate capCert = checkAccessResponse.getCapabilityCertificate();

  if (!capCert) {
    return 1; // we have no rights
  } else {
    *outCap = new CapabilityCertificate(capCert);
    return 0;  // we have some rights
  }
}

int 
remoteCheckAccess(const char *subjectDN, 
                  const char *subjectCA, 
                  const char *resource, 
	          CapCert *outCap, 
                  AKMessage *outMsg) { 
  delete(*outMsg);
  *outMsg = NULL;

  if (!subjectDN || !subjectCA || !resource) {
    *outMsg = new AkentiMessage(200);
    return -1;
  }

  try {
    AkentiPrincipal ap(subjectDN, subjectCA);

    return clientCheckAccess(ap, resource, outCap, outMsg);
  } catch(Exception &ex) {
    *outMsg = new AkentiMessage(205, " " + ex.getMessage());
  } catch(Error &err) {
    *outMsg = new AkentiMessage(206, " " + err.getMessage());
  } catch(...) {
    *outMsg = new AkentiMessage(207);
  }

  return -1;
}

int 
remoteCheckAccessByCertString(const char *subjectCert, 
                              const char *resource, 
	                      CapCert* outCap, 
                              AKMessage *outMsg) {
  delete(*outMsg);
  *outMsg = NULL;

  if (!subjectCert || !resource) {
    *outMsg = new AkentiMessage(200);
    return -1;
  }

  try {
    CertificateFactory fac = CertificateFactory::getInstance();
    Certificate cert = fac->initCertificate("PEM", subjectCert);
    AkentiPrincipal ap = AkentiPrincipal(cert->getPrincipal(), cert->getGuarantor());

    return clientCheckAccess(ap, resource, outCap, outMsg);
  } catch(Exception &ex) {
    *outMsg = new AkentiMessage(205, " " + ex.getMessage());
  } catch(Error &err) {
    *outMsg = new AkentiMessage(206, " " + err.getMessage());
  } catch(...) {
    *outMsg = new AkentiMessage(207);
  }

  return -1;
}

int 
remoteCheckAccessByCertFile(const char *subjectCertFile, 
                            const char *resource, 
	                    CapCert *outCap, 
                            AKMessage *outMsg)  {
  delete(*outMsg);
  *outMsg = NULL;

  if (!subjectCertFile || !resource) {
    *outMsg = new AkentiMessage(200);
    return -1;
  }

  try {
    CertificateFactory fac = CertificateFactory::getInstance();
    Certificate cert = fac->initCertificateFromFile("PEM", subjectCertFile);
    AkentiPrincipal ap = AkentiPrincipal(cert->getPrincipal(), cert->getGuarantor());

    return clientCheckAccess(ap, resource, outCap, outMsg);
  } catch(Exception &ex) {
    *outMsg = new AkentiMessage(205, " " + ex.getMessage());
  } catch(Error &err) {
    *outMsg = new AkentiMessage(206, " " + err.getMessage());
  } catch(...) {
    *outMsg = new AkentiMessage(206);
  }

  return -1;
}

