#include <fstream>
#include "AkentiInit.hpp"
#include "AkentiEngine.hpp"
#include "AkentiPrincipal.hpp"
#include "AkentiMessage.hpp"
#include "DistinguishedName.hpp"
#include "CapabilityCertificate.hpp"
#include "CertificateChain.hpp"
#include "CertificateFactory.hpp"
#include "ExternalEvaluator.hpp"
#include "ExternalEvaluatorImpl.hpp"
#include "AkGlobusEngInf.h"
#include "AkLogger.h"
#include "Logger.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "IOException.hpp"
#include "Exception.hpp"
#include "Error.hpp"

using namespace Akenti;
using namespace std;

namespace Akenti {

class GlobusExternalEvaluator : public ExternalEvaluatorImpl {
  
public:

GlobusExternalEvaluator() {
}

~GlobusExternalEvaluator() {
}

bool canHandleSystemAttribute(const string& attribute) {
  return true;
}
 
bool evaluateSystemAttribute(TokenType op, const AttributeInfo& attrInfo) {
  return false;
}
 
bool canHandleExternalAttribute(const string& attribute) {
  return true;
}

bool evaluateExternalAttribute(TokenType op, const AttributeInfo& attrInfo) {
  return false;
}

}; // end of GlobusExternalEvaluator 

} // end of namespace

int
initLogger(const char* id, const char* conFile, void** logger, char** status) {
  try {
    if (Logger::hasBeenInitialized()) {
      *status = strdup("logger already initilized");
      return 1; 
    }

    Properties props;

    props.load(conFile);

    string message;

    *logger = Logger::init(id, props, message);
    *status = strdup(message.c_str());
    return 0; 
  } catch(IOException& ex) {
    *status = strdup("io error occured when initilizing logger");
  } catch(...) {
    *status = strdup("error initializing logger");
  }

  return -1;
}

int deleteLogger(void* logger) {
  delete((Logger*) logger);
  return 0;
}

void
bLog(int level, const char* key, const char* mesg) { 
  Logger::write(level, key, string("STATUS=BEGIN ") + mesg);
}

void 
iLog(int level, const char* key, const char* mesg) {
  Logger::write(level, key, mesg);
}

void
sLog(int level, const char* key, const char* mesg) {
  Logger::write(level, key, string("STATUS=SUCCESS ") + mesg);
}

void
fLog(int level, const char* key, const char* mesg) {
  Logger::write(level, key, string("STATUS=FAIL ") + mesg);
}

void flush() {
  Logger::flush();
}

int akentiGlobusCheck(const char* subjectDN,
                      const char* subjectCADN,
                      const char* resource,
                      const char* action,
                      const char* confFile,
                      char** outStatus) {
   string status;
   int ret = -1;

   Logger* logger = NULL;

   try {
     if (!AkentiInit::init(confFile, status)) {
       status = "901 akenti library initialization failed";
       goto error; 
     } else if (!Logger::hasBeenInitialized()) {
       logger = Logger::init("AkentiGlobusEngine", status); // KEEP GOING?
     }

     CapabilityCertificate capCert;
     AkentiMessage msg;
     AkentiEngine eng;
     AkentiPrincipal principal(subjectDN, subjectCADN);

     bLog(LOG_AUD, "CALLOUT_GLUE_MODULE", "");

     eng.setExternalEvaluator(new GlobusExternalEvaluator);

     if (!eng.checkAccess(resource, principal, capCert, msg)) {
       ret = 1;
       status = msg.toString();
       goto error;
     }

     string temp = string("DESIRED_ACTION=") + action;

     iLog(LOG_MON, "INFO", temp.c_str());

     const vector<string>& actions = capCert->getUnConditionalActions();

     for (int i = 0; i < actions.size(); ++i) {
        temp = "GRANTED_ACTION=" + actions[i];
        iLog(LOG_MON, "INFO", temp.c_str());

        if (StringUtil::equalsIgnoreCase(actions[i], action)) {
          ret = 0;
          status = "104 action granted";
          sLog(LOG_AUD, "CALLOUT_GLUE_MODULE", ("MESSAGE=" +status).c_str());
          goto exit;
        }
     }

     ret = 1;
     fLog(LOG_AUD, "CALLOUT_GLUE_MODULE", "MESSAGE=310 action not granted");
     goto exit;
   } catch (Exception& ex) {
     status = "901 " + ex.getMessage();
   } catch (Error& err) {
     status = "901 " + err.getMessage();
   } catch (...) {
     status = "901 unknown error";
   }

error:
   fLog(LOG_AUD, "CALLOUT_GLUE_MODULE", ("ERROR=" + status).c_str());
exit:
   *outStatus = strdup(status.c_str());
   delete(logger);
   return ret;
}
