#include <iostream>
#include <sstream>
#include <fstream>
#include "cap.h" 
#include "eval.h" 
#include "Expression.hpp" 
#include "ExpressionEvaluator.hpp" 
#include "CapabilityCertificate.hpp" 
#include "AkentiCertificate.hpp"
#include "Provider.hpp" 
#include "OpenSSLProviderImpl.hpp" 
#include "Exception.hpp" 
#include "Error.hpp" 

using namespace Akenti;
using namespace std;

void initProvider() {
  if (!Provider::isProviderInstalled()) {
    Provider po = Provider(new OpenSSLProviderImpl);

    po->loadSignatureAlgorithms();
    Provider::setProvider(po);
  }
}

int CapCert_initFromString(const char* encType, const char* certString, CapCert* cert)  {
  if (cert) {
    try { 
      CapCert_free(*cert);
      *cert = NULL;
      initProvider();

      istringstream is(certString);
      AkentiCertificate temp = AkentiCertificate::initFromStream(encType, is);

      if (temp.instanceOf<CapabilityCertificate>()) {
        *cert = new CapabilityCertificate((CapabilityCertificate&) temp);
        return 0;
      } else {
        return 1;
      }
    } catch(Exception& ex) {
      return 1;
    } catch(...) {
    }
  }

  return -1;
}

int CapCert_initFromFile(const char* encType, const char* fileName, CapCert* cert)  {
  if (cert) {
    try { 
      CapCert_free(*cert);
      *cert = NULL;
      initProvider();
      AkentiCertificate temp = AkentiCertificate::initFromFile(encType, fileName);
       
      if (temp.instanceOf<CapabilityCertificate>()) {
        *cert = new CapabilityCertificate((CapabilityCertificate&) temp);
        return 0;
      } else {
        return 1;
      }
    } catch(Exception& ex) {
      return 1;
    } catch(...) {
    }
  }

  return -1;
}

int CapCert_writeToBuffer(const char* encType, const CapCert cert, char* buff) {
  try {
    ostringstream os;
    AkentiCertificate::writeToStream(encType, os, *cert);
    string temp = os.str();

    if (!buff) {
      return temp.length();
    }

    int len = temp.length(); 

    temp.copy(buff, len, 0);
    return len;
  } catch(Exception& ex) {
    return 0;
  } catch(...) {
    return -1;
  }
}

int CapCert_writeToFile(const char* encType, const CapCert cert, const char* fileName) {
  try {
    AkentiCertificate::writeToFile(encType, fileName, *cert);
    return 0;
  } catch(Exception& ex) {
    return 1;
  } catch(...) {
    return -1;
  } 
}

int CapCert_toText(const CapCert cert, char* buff) {
  try {
    string temp = (*cert)->toText();

    if (!buff) {
      return temp.length();
    }

    int len = temp.length(); 

    temp.copy(buff, len, 0);
    return len;
  } catch(Exception& ex) {
    return 0;
  } catch(...) {
    return -1;
  }
}

void CapCert_free(CapCert cert) {
  delete(cert);
}

int CapCert_getVersionNumber(const CapCert cert) {
  return (*cert)->getVersionNumber();
}

const AKPrincipal CapCert_getSubject(const CapCert cert) {
  const AkentiPrincipal& ap = (*cert)->getPrincipal();

  return (AKPrincipal) &ap;
}

const PubKey CapCert_getSubjectPublicKey(const CapCert cert) {
  const PublicKey& key = (*cert)->getPublicKey();

  return (PubKey) &key;
}

const char* CapCert_getResource(const CapCert cert) {
  return (*cert)->getResource().c_str();
}

const AKPrincipal CapCert_getIssuer(const CapCert cert) {
  const AkentiIssuer& issuer = (*cert)->getIssuer();
  const AkentiPrincipal& ap = issuer.getIssuer();

  return (AKPrincipal) &ap;
}

const UTCTime CapCert_getNotBefore(const CapCert cert) {
  const UtcTime& utc = (*cert)->getNotBefore();

  return (UTCTime) &utc;
}

const UTCTime CapCert_getNotAfter(const CapCert cert) {
  const UtcTime& utc = (*cert)->getNotAfter();

  return (UTCTime) &utc;
}

int CapCert_checkValidity(const CapCert cert) {
  return (*cert)->checkValidity();
}

int  CapCert_hasGrantedActions(const CapCert cert) {
  return CapCert_numOfGrantedActions(cert) > 0;
}

int CapCert_numOfGrantedActions(const CapCert cert) {
  return (*cert)->getUnConditionalActions().size();
}

const char* CapCert_grantedActionAt(const CapCert cert, int idx) {
  return (*cert)->getUnConditionalActions()[idx].c_str();
}

int CapCert_hasCondActions(const CapCert cert) {
  return CapCert_numOfCondActions(cert) > 0;
}

int
CapCert_numOfCondActions(const CapCert cert) {
  return (*cert)->getConditionalActions().size();
}

const CondActions CapCert_condActionsAt(const CapCert cert, int idx) {
  const vector<ConditionalActions>&
     temp = (*cert)->getConditionalActions();

  return (CondActions) &temp[idx];
}

int AKPrin_getName(const AKPrincipal ap, char* buff) {
  const DistinguishedName& dn = ap->getName();
  string temp = dn.toString();

  if (!buff) {
     return temp.length();
  }
  
  int len = temp.length();

  temp.copy(buff, len, 0);
  return len;
}

int AKPrin_getGuarantor(const AKPrincipal ap, char* buff) {
  const DistinguishedName& dn = ap->getGuarantor();
  string temp = dn.toString();

  if (!buff) {
     return temp.length();
  }
  
  int len = temp.length();

  temp.copy(buff, len, 0);
  return len;
}

int UTCT_toUtcString(const UTCTime utc, char* buff) {
  string temp = utc->toString();
  
  if (!buff) {
     return temp.length();
  }

  int len = temp.length();

  temp.copy(buff, len, 0);
  return len;
}

const char* CondAct_getConstraint(const CondActions condActions) {
  return condActions->getConstraint().c_str();
}

int CondAct_isCritical(const CondActions condActions) {
  return condActions->isCritical();
}

int CondAct_numOfActions(const CondActions condActions) {
  return condActions->getActions().size();
}

const char* CondAct_actionAt(const CondActions condActions, int idx) {
  return condActions->getActions()[idx].c_str();
}

int CondAct_numOfAttrInfos(const CondActions condActions) {
  return condActions->getAttributeInfos().size();
}

const AttrInfo CondAct_attrInfoAt(const CondActions condActions, int idx) {
  const vector<AttributeInfo>& temp = condActions->getAttributeInfos();
  
  return (AttrInfo) &temp[idx]; 
}

AttrType Attr_getType(const AttrInfo info) {
  return info->getType();
}

const char* Attr_getAttributeName(const AttrInfo info) {
  return info->getAttribute().c_str();
}

const char* Attr_getValue(const AttrInfo info) {
  return info->getValue().c_str();
}

namespace Akenti {

class CEvaluator : public ExpressionEvaluator {

public:

  CEvaluator(CondActions condActions, EvalCallBack func, void* state) : idx(0) {
    this->condActions = condActions;
    this->func = func;
    this->state = state;
  }

  int evaluate() {
    Expression expr = Expression(condActions->getConstraint());

    return expr.evaluate(*this);
  }

  bool evaluate(TokenType type, const string& leftOperand, const string& rightOperand) {
    AttrInfo info = CondAct_attrInfoAt(condActions, idx++);

    return func(type, info, state) == 1;
  }

private:

  CondActions condActions;
  int idx;
  EvalCallBack func;
  void* state; 

};

}

int CondAct_evaluate(const CondActions condActions,  EvalCallBack func, void* state) {
  CEvaluator evaluator(condActions, func, state);

  return evaluator.evaluate();
}

int AKMsg_getCode(AKMessage mesg) {
  return mesg->getCode();
}

const char* AKMsg_getDescription(AKMessage mesg) {
  return mesg->getMessage().c_str();
}

void AKMsg_free(AKMessage mesg) {
  delete(mesg);
}

