#include <sstream>
#include <fstream>
#include <algorithm>
#include <cctype>
#include <memory>

#include "AkentiCertificate.hpp"
#include "AkentiCertificateFactory.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "Exception.hpp"
#include "IOException.hpp"
#include "Base64.hpp"
#include "StringUtil.hpp"
#include "AkentiConstants.hpp"

#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

using namespace std;

namespace Akenti {

static const string B_HEADER = "-----BEGIN AKENTI ";
static const string E_HEADER = "-----END AKENTI ";
static const string C_HEADER = " CERTIFICATE-----";

static string 
formHeader(bool begin, const string& type) {
  string ret = begin? B_HEADER : E_HEADER; 
  string temp = type;

  // Shuts the compiler. 
  int (*op)(int) = std::toupper;
  std::transform(type.begin(), type.end(), temp.begin(), op);
  ret += temp;
  ret += C_HEADER;
  return ret;
}
  
static bool 
parseHeader(bool begin, const string& header, string& type) {
  string temp = begin? B_HEADER : E_HEADER;

  int i = header.find(temp);

  // allow for stuff to precede the PEM Header
  if (i == string::npos) {
    return false;
  }

  i += temp.size();

  int j = header.find(C_HEADER);

  if (j <= i || j + C_HEADER.size() != header.size()) {
    return false;
  }

  type = header.substr(i, j - i);
  return true;
}

typedef AkentiCertificate AC;

AC::AkentiCertificate() {
}

AC::AkentiCertificate(AkentiCertificateImpl *impl) 
    : ref_ptr<AkentiCertificateImpl>(impl)  {
  if (!impl) {
    throw Error("AkentiCertificate:constructor:illegal argument");
  }
}

AC::~AkentiCertificate() {
}

bool 
AC::operator==(const AkentiCertificate& cert) const {
  // XXXX

  AkentiCertificateImpl *impl1 = get();
  AkentiCertificateImpl *impl2 = cert.get();

  return  
    impl1->getSignable() == impl2->getSignable() 
                         && 
    impl1->getSignature() == impl2->getSignature();
}

void 
AC::writeObject(AkentiOutputStream& aos) const {
  AkentiCertificateImpl* impl = get();

  aos << impl->getType()
      << ' ';

  impl->writeObject(aos);
}

void 
AC::writeEncoding(const string& enc, ostream& os)const {
  if (enc == "XML") {
    DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(XMLUtil::convert("Core"));
    DOMDocument* doc = impl->createDocument(0, XMLUtil::convert("WORKAROUND"), 0);
    DOMElement* element = doc->createElement(XMLUtil::convert("ak:AkentiCertificate"));

    element->setAttribute(XMLUtil::convert("xmlns:ak"), 
                          XMLUtil::convert(CURRENT_CERT_NAME_SPACE));


    get()->toXML(element);
    XMLUtil::write(os, element);
    return;
  }

  if (enc != "PEM") {
    throw Error(
      "AkentiCertificate:writeEncoding:unsupported encoding:" + enc 
    );
  }

  ostringstream ostream;
  AkentiOutputStream aos(ostream);

  writeObject(aos);
     
  string type = get()->getType();

  // Shuts the compiler. 
  int (*op)(int) = std::toupper;
  std::transform(type.begin(), type.end(), type.begin(), op); 

  os << B_HEADER
     << type
     << C_HEADER
     << '\n' 
     << Base64::encode(ostream.str())
     << '\n' 
     << E_HEADER
     << type
     << C_HEADER
     << '\n';
} 

AkentiCertificate 
AC::readObject(AkentiInputStream& ais) {
  string type;

  ais >> type;

  AkentiCertificateFactory& fac = AkentiCertificateFactory::getAkentiCertificateFactory();
  AkentiCertificate cert(fac.getPrototype(type));

  cert->readObject(ais);
  return cert;
}


AkentiCertificate 
AC::readEncoding(const string& encType, istream& is) {
  if (encType == "XML") {
    auto_ptr<AbstractDOMParser> parser = XMLUtil::parse(is);
    DOMDocument* doc = parser->getDocument();
    DOMElement* root = doc->getDocumentElement();

    XMLUtil::filter(root);

    DOMNodeList *list = root->getElementsByTagName(XMLUtil::convert("Header"));

    if (list->getLength() != 1) {
      throw Exception("No header");
    }

    DOMNode *temp = list->item(0);
    DOMElement *header = (DOMElement*) temp;
    string type = XMLUtil::convert(header->getAttribute(XMLUtil::convert("Type")));
    AkentiCertificateFactory& fac = AkentiCertificateFactory::getAkentiCertificateFactory();
    AkentiCertificate cert(fac.getPrototype(type));

    cert->fromXML(root);
    return cert;
  }

  if (encType != "PEM") {
    throw Error(
      "AkentiCertificate:readEncoding:unsupported encoding:" + encType
    );
  }

  string beginCertType;
  string line;

  do {
    if (!std::getline(is, line)) {
      throw Exception("AkentiCertificate:readPemEncoding:No BEGIN CERTIFICATE TAG found");
    }
    
  }  while (!parseHeader(true, line, beginCertType)); 
    

  if (!std::getline(is, line)) {
    throw Exception(
     "AkentiCertificate:readPemEncoding:No encoded cert found"
    );
  }

  string endCertType;
  string dec;

  for (int i = 64; i < line.size(); i += 65) {
    line[i] = '\n';
  }

  while (!parseHeader(false, line, endCertType)) {
    dec += line;
    dec += '\n';

    if (!std::getline(is, line)) {
      throw Exception(
        "AkentiCertificate:readPemEncoding:No END CERTIFICATE TAG found"
      );
    }
  }

  if (endCertType != beginCertType) {
    throw Exception("AkentiCertificate:readPemEncoding:BEGIN and END tags did not match");
  }

  istringstream stream(Base64::decode(dec));
  AkentiInputStream ais(stream);
  AkentiCertificate cert = readObject(ais);
  string type = cert->getType();

  if (!StringUtil::equalsIgnoreCase(endCertType, type)) {
    throw Exception(
      "AkentiCertificate:readPemEncoding:cert type:" + type + " does not match PEM TAG"
    );
  }
  
  return cert;
}

void
AC::writeToStream(const string& encType, 
                  ostream& os, 
                  const AkentiCertificate& cert) {
  cert.writeEncoding(encType, os);
}

AkentiCertificate 
AC::initFromStream(const string& encType, istream& is) {
  AkentiCertificate cert = readEncoding(encType, is);
  return cert;
}

void
AC::writeToFile(const string& encType, 
                const string& filename, 
                const AkentiCertificate& cert) {
  ofstream os((char *) filename.c_str());

  if (!os) {
    throw IOException(
      "AkentiCertificate:writeToFile:unable to open file:" + filename
    );
  }

  cert.writeToStream(encType, os, cert);
  os.close();
}

AkentiCertificate 
AC::initFromFile(const string& encType, const string& filename) {
  if (encType.size() == 0) {
    return initFromFile(filename);
  }

  ifstream is((char *) filename.c_str());

  if (!is) {
    throw IOException(
      "AkentiCertificate:initFromFile:unable to open file:" + filename
    );
  }
  
  AkentiCertificate cert = initFromStream(encType, is); 

  is.close();
  return cert;
}

AkentiCertificate 
AC::initFromFile(const string& filename) {
  ifstream is1((char *) filename.c_str());
  ostringstream os;
  string line;

  while (std::getline(is1, line)) {
    os << line << '\n';
  }

  is1.close();

  string temp = os.str();
  string encType = (temp.find("-----BEGIN") != -1 ? "PEM" : "XML");
  istringstream is2(temp);
  
  return initFromStream(encType, is2); 
}


string 
AC::paramString() const {
  return get()->paramString();
}

string
AC::toLogString() const {
  return get()->toLogString();
}

AkentiOutputStream& 
operator<<(AkentiOutputStream& os, const AkentiCertificate& cert) {
  cert.writeObject(os);
  return os;
}

AkentiInputStream&
operator>>(AkentiInputStream& ais, AkentiCertificate& cert) {
  cert = AkentiCertificate::readObject(ais);
  return ais;
}

istream& 
operator>>(istream& is, Akenti::AkentiCertificate& cert) {
  cert = Akenti::AkentiCertificate::initFromStream("PEM", is);
  return is;
}

ostream& 
operator<<(ostream& os, const Akenti::AkentiCertificate& cert) {
  Akenti::AkentiCertificate::writeToStream("PEM", os, cert);
  return os;
}

} // end of namespace
