#include "CapabilityCertificateImpl.hpp"
#include "KeyFactory.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "Util.hpp"
#include "StringTokenizer.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Base64.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

using std::endl;

namespace Akenti {

typedef CapabilityCertificateImpl CCI;

CCI::CapabilityCertificateImpl() 
  : AkentiCertificateImpl("Capability"),
    principal(),
    resource("NO_RESOURCE")  {
}

CCI::~CapabilityCertificateImpl() {
}

string
CCI::getHashString() const {
  return 
    resource 
    + Util::getHashDN(principal.getName())
    + Util::getHashDN(principal.getGuarantor());
}

void 
CCI::setResource(const string& resource) {
  this->resource = resource;
}

const string&
CCI::getResource() const {
  return resource;
}

void 
CCI::setPrincipal(const AkentiPrincipal& principal) {
  this->principal = principal;
}

const AkentiPrincipal&
CCI::getPrincipal() const {
  return principal;
}


void 
CCI::setPublicKey(const PublicKey& publicKey) {
  if (publicKey == NULL) {
    throw Error(
      "CapabilityCertificateImpl:setPublicKey:illegal arg"
    );
  }

  this->publicKey = publicKey;
}

const PublicKey&
CCI::getPublicKey() const {
  return publicKey;
}

void 
CCI::setUnConditionalActions(const vector<string>& unConditionalActions) {
  this->unConditionalActions = unConditionalActions;
}

const vector<string>&
CCI::getUnConditionalActions() const {
  return unConditionalActions; 
}

void 
CCI::setConditionalActions(const vector<ConditionalActions>& conditionalActions) {
  this->conditionalActions = conditionalActions;
}

const vector<ConditionalActions>&
CCI::getConditionalActions() const {
  return conditionalActions;
}

void
CCI::doWrite(AkentiOutputStream& drain) const {
  drain << principal;

  if (version == 2) {
    KeyFactory fac = KeyFactory::getInstance();
    string der = fac->getKeyEncoded("DER", publicKey);

    int len = der.length();

    drain << ' ' << len << ' ';
    drain.write(der.data(), len);
  }

  drain << ' ' << resource;

  int size =  unConditionalActions.size();

  drain << ' ' << size;

  for (int i = 0; i < size; ++i) {
    drain << ' ' << unConditionalActions[i];
  }

  size =  conditionalActions.size();

  drain << ' ' << size;

  for (int i = 0; i < size; ++i) {
    drain << ' ' << conditionalActions[i];
  }
}

void 
CCI::doRead(AkentiInputStream& source) {
  conditionalActions.clear();
  unConditionalActions.clear();
  publicKey = PublicKey();

  source >> principal; 

  if (version == 2) {
    int len;

    source >> len;

    if (len <= 0) {
      throw Exception("CapabalityCertificateImpl:doRead: len < 0");
    }

    if (source.get() != ' ') {
      throw Exception("CapabalityCertificateImpl:readObject: expecting blank space");
    }

    string der  = string(len, '\0');

    source.read((char*) der.data(), len);
     
    KeyFactory fac = KeyFactory::getInstance();

    publicKey = fac->initPublicKey("DER", der);
  }

  source >> resource;

  int size;
  
  source >> size;

  if (size < 0) {
    throw Exception("CapabilityCertificateImpl:doRead:size < 0(1)");
  }

  for (int i = 0; i < size; ++i) {
    string unConditionalAction;

    source >> unConditionalAction;
    unConditionalActions.push_back(unConditionalAction);
  }
  
  source >> size;

  if (size < 0) {
    throw Exception("CapabilityCertificateImpl:doRead:size < 0(2)");
  }
  
  for (int i = 0; i < size; ++i) {
    ConditionalActions conditionalActions;

    source >> conditionalActions;
    this->conditionalActions.push_back(conditionalActions);
  }
}

void
CCI::doToXML(DOMElement* capabilityCertInfo) const {
  DOMDocument *doc = capabilityCertInfo->getOwnerDocument();
  DOMElement *temp = doc->createElement(XMLUtil::convert("ResourceName"));

  temp->appendChild(doc->createTextNode(XMLUtil::convert(resource.c_str())));
  capabilityCertInfo->appendChild(temp);

  temp = doc->createElement(XMLUtil::convert("SubjectAndCA"));
  principal.toXML(temp);
  capabilityCertInfo->appendChild(temp);

  temp = doc->createElement(XMLUtil::convert("PublicKey"));
  
  KeyFactory fac = KeyFactory::getInstance();
  string der = Base64::encode(fac->getKeyEncoded("DER", publicKey));

  temp->appendChild(doc->createTextNode(XMLUtil::convert(der.c_str())));
  capabilityCertInfo->appendChild(temp);

  for (int i = 0; i < unConditionalActions.size(); i++) {
    string sb = unConditionalActions[i];
    temp = doc->createElement(XMLUtil::convert("Action"));

    temp->appendChild(doc->createTextNode(XMLUtil::convert(sb.c_str())));
    capabilityCertInfo->appendChild(temp);
  }

  for (int i=0; i < conditionalActions.size(); i++) {
    temp = doc->createElement(XMLUtil::convert("ConditionalAction"));

    conditionalActions[i].toXML(temp);
    capabilityCertInfo->appendChild(temp);
  }
}

void
CCI::doFromXML(const DOMElement* capabilityCertInfo) {
  unConditionalActions.clear();
  conditionalActions.clear();
  publicKey = PublicKey();

  DOMNode *child = capabilityCertInfo->getFirstChild();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("CapabilityCertificateImpl:doFromXML:missing resource");
  }

  DOMNode *node = child->getFirstChild();

  if (node  == NULL) {
    throw Exception("CapabilityCertificateImpl:doFromXML:missing resource");
  }

  resource = XMLUtil::convert(node->getNodeValue());
  child = child->getNextSibling();
  
  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("CapabilityCertificateImpl:doFromXML:missing principal");
  }

  principal.fromXML((DOMElement*) child);
  child = child->getNextSibling();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("CapabilityCertificateImpl:doFromXML:missing publicKey");
  }

  node = child->getFirstChild();

  if (node  == NULL) {
    throw Exception("CapabilityCertificateImpl:doFromXML:missing publicKey");
  }


  // XXX
  string base64 = XMLUtil::convert(node->getNodeValue());

  if (base64[0] == '\n') {
      base64 = base64.substr(1);
  }

  for (int i = 64; i < base64.size(); i += 65) {
      base64[i] = '\n';
  }

  string der(Base64::decode(base64));
  KeyFactory fac = KeyFactory::getInstance();
  publicKey = fac->initPublicKey("DER", der);

  DOMNodeList *actionList = capabilityCertInfo->getElementsByTagName(XMLUtil::convert("Action"));

  for (int i = 0; i < actionList->getLength(); i++) {
    node = actionList->item(i);
    
    // XXXXXXXXX WHAT IS THIS?
    if (node->getParentNode() == (DOMNode*) capabilityCertInfo) {
      string action = XMLUtil::convert(node->getFirstChild()->getNodeValue());

      unConditionalActions.push_back(action);
    }
  }

  DOMNodeList *list = capabilityCertInfo->getElementsByTagName(XMLUtil::convert("ConditionalAction"));

  for (int i=0; i < list->getLength(); i++) {
    ConditionalActions condActions;

    child = list->item(i);
    condActions.fromXML((DOMElement*) child);
    conditionalActions.push_back(condActions);
  }
}

void
CCI::doParamString(ostream& os) const {
  os << getPrincipal().paramString();

  if (publicKey) {
    os << ',' << publicKey->paramString();
  }

  os << ',' << getResource();

  vector<string> unConditionalActions = getUnConditionalActions();
  int size =  unConditionalActions.size();
  
  os << ',';
  os << '[';

  if (size) {
    os << unConditionalActions[0];

    for (int i = 1; i < size; ++i) {
      os << ',' << unConditionalActions[i];
    }
  }

  os << ']';

  vector<ConditionalActions> conditionalActions = getConditionalActions();
  size =  conditionalActions.size();

  os << ',';
  os << '[';

  if (size) {
    os << conditionalActions[0].paramString();

    for (int i = 1; i < size; ++i) {
      os << ',' << conditionalActions[i].paramString();
    }
  }

  os << ']';
}

void 
CCI::doToText(int indentAmt, ostream& os) const {
  os << Util::fText(indentAmt + 3, "Subject:") << endl  
     << Util::fText(indentAmt + 6,
                     "Name: " + getPrincipal().getName().toString()) << endl
     << Util::fText(indentAmt + 6,
                     "CA  : " + getPrincipal().getGuarantor().toString()) << endl
     << Util::fText(indentAmt + 3, "Resource:") << getResource() << endl;  

  const vector<string>& unConditionalActions = getUnConditionalActions();
  int size =  unConditionalActions.size();

  // XXX to print public key 

  if (size) {
    os << Util::fText(indentAmt + 3, "Granted Actions:") << endl;
    os << Util::fText(indentAmt + 6, unConditionalActions[0]);

    for (int i = 1; i < size; ++i) {
      os << ", " << unConditionalActions[i];
    }

    os << endl;
  }

  const vector<ConditionalActions>& conditionalActions = getConditionalActions();

  size =  conditionalActions.size();

  if (size) {
    os << Util::fText(indentAmt + 3, "Conditional Action(s):") << endl;

    for(int i = 0; i < size; i++) {
      os << Util::fText(indentAmt + 6, "") << i + 1 << '.' << endl;

      string critical =  conditionalActions[i].isCritical() ? "true" : "false";

      os << Util::fText(indentAmt + 7, "Critical  : " + critical) << endl;
      os << Util::fText(indentAmt + 7, "Constraint: " + conditionalActions[i].getConstraint()) << endl;

      const vector<string>& conditionActions = conditionalActions[i].getActions();

      int casize = conditionActions.size();  

      if (casize) {
        os << Util::fText(indentAmt + 7, "Actions:") << endl;
        os << Util::fText(indentAmt + 9, conditionActions[0]);

        for (int i = 1; i < casize; ++i) {
          os << ", " << conditionActions[i];
        } 

        os << endl;
      } 

      os << Util::fText(indentAmt + 7, "Attribute Info(s):") << endl;

      const vector<AttributeInfo>& ais = conditionalActions[i].getAttributeInfos();

      for (int i = 0; i < ais.size(); ++i) {
        static string TYPES[] = {"SYSTEM", "X509", "AKENTI", "EXT_AUTH"};
        AttributeInfo::Type type = ais[i].getType();

        os << Util::fText (
          indentAmt + 9, 
	  "attribute = " + ais[i].getAttribute() + " , " +
          "value = " + ais[i].getValue() + 
	  ", type = " + TYPES[type]
        ) << endl;

        if (type == AttributeInfo::X509) { // X509
          const vector<DistinguishedName>& cas = ais[i].getCAS();

          for (int i = 0; i < cas.size(); i++) {
            os << Util::fText(indentAmt + 12, "CA  : " +  cas[i].toString())
               << endl;
          }
        } else if (type == AttributeInfo::AKENTI || type == AttributeInfo::EXT_AUTH) {
          const vector<AkentiPrincipal>& aps = 

            type == AttributeInfo::AKENTI? ais[i].getIssuers() : ais[i].getAuthorities();

          for (int i = 0; i < aps.size(); i++) {
            os << Util::fText(indentAmt + 12,
                              "Name: " + aps[i].getName().toString()) << endl
               << Util::fText(indentAmt + 12,
                             "CA  : " + aps[i].getGuarantor().toString())
               << endl;
          }
        }

        os << endl;
      } 
    }

    os << endl;
  } // end conditional actions.
}

CapabilityCertificateImpl*
CCI::newInstance() const {
  return new CapabilityCertificateImpl;
}

}
