#include <sstream>
#include "PolicyCertificateImpl.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Util.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

using namespace std;

namespace Akenti {

typedef PolicyCertificateImpl PCI;
typedef CertificateAuthorityInfo CAInfo;
typedef UseCondIssuerGroup UCGroup;

PCI::PolicyCertificateImpl()  
  : AkentiCertificateImpl("Policy"),
    resource("NO_RESOURCE_NAME"),
    certAuthInfos(),
    attributeDirs(),
    ucIssuerGroups(1),
    cacheTime(-1) {
}
  
PolicyCertificateImpl*
PCI::newInstance() const {
  return new PolicyCertificateImpl;
}

PCI::~PolicyCertificateImpl() {
}

string 
PCI::getHashString() const {
  return getResource(); 
}

void
PCI::setResource(const string& resource) {
  if (!resource.size()) {
    throw Error("PolicyCertificateImpl:setResource:illegal arg");
  }

  this->resource = resource;
}

const string&
PCI::getResource() const {
  return resource;
}

void
PCI::setCertificateAuthorityInfos(const vector<CAInfo>& certAuthInfos) {
  if (!certAuthInfos.size()) {
    throw Error(
      "PolicyCertificateImpl:setCertificateAuthorityInfos:size = 0"
    );
  }
  
  this->certAuthInfos = certAuthInfos;
}

bool 
PCI::containsCertificateAuthorityInfos() const {
  return certAuthInfos.size();
}

const vector<CertificateAuthorityInfo>&
PCI::getCertificateAuthorityInfos() const {
  return certAuthInfos;
}

void 
PCI::setUseCondIssuerGroups(const vector<UCGroup>& ucIssuerGroups) {
  if (!ucIssuerGroups.size()) {
    throw Error("PolicyCertificateImpl:setUseCondIssuerGroups:size = 0");
  }

  this->ucIssuerGroups = ucIssuerGroups;
}

const vector<UseCondIssuerGroup>& 
PCI::getUseCondIssuerGroups() const {
  return ucIssuerGroups; 
}

void 
PCI::setAttributeDirectories(const vector<Directory>& attributeDirs) {
  if (!attributeDirs.size()) {
    throw Error("PolicyCertificateImpl:setAttributeDirectories:size = 0");
  }

  this->attributeDirs = attributeDirs;
}

const vector<Directory>&
PCI::getAttributeDirectories() const {
  return attributeDirs;
}

void
PCI::setCacheTime(long cacheTime) {
  this->cacheTime = cacheTime;
}

long
PCI::getCacheTime() const {
  return cacheTime;
}

void 
PCI::doWrite(AkentiOutputStream& drain) const {
  drain << resource;

  int size = certAuthInfos.size();
  
  drain << ' ' << size;

  for (int i = 0; i < size; ++i) {
    drain << certAuthInfos[i];
  }

  drain << ' ' << (size = ucIssuerGroups.size());

  for (int i = 0; i < size; ++i) {
    drain << ' ' << ucIssuerGroups[i];
  }

  drain << ' ' << (size = attributeDirs.size());

  for (int i = 0; i < size; ++i) {
    drain << ' ' << attributeDirs[i];
  }

  drain << ' ' << cacheTime;
}

void
PCI::doRead(AkentiInputStream& source) {
  certAuthInfos.clear();
  ucIssuerGroups.clear();
  attributeDirs.clear();

  source >> resource;

  int size;

  source >> size;
  
  if (size < 0) {
    throw Exception("PolicyCertificateImpl:doRead: number of ca infos < 0");
  }

  for (int i = 0; i < size; ++i) {
    CAInfo caInfo;

    source >> caInfo;
    certAuthInfos.push_back(caInfo);
  }

  source >> size;

  if (size <= 0) {
    throw Exception("PolicyCertificateImpl:doRead: number of UseCondGroups <= 0");
  }

  for (int i = 0; i < size; ++i) {
    UCGroup ucGroup;

    source >> ucGroup;
    ucIssuerGroups.push_back(ucGroup);
  }

  source >> size;

  if (size < 0) {
    throw Exception("PolicyCertificateImpl:doRead:number of attrDirs < 0");
  }

  for (int i = 0; i < size; ++i) {
    Directory dir;

    source >> dir;
    attributeDirs.push_back(dir);
  }

  source >> cacheTime;
}

void 
PCI::doToXML(DOMElement* policyCertInfo) const {
  DOMDocument *doc = policyCertInfo->getOwnerDocument();
  DOMElement *temp = doc->createElement(XMLUtil::convert("ResourceName"));

  temp->appendChild(doc->createTextNode(XMLUtil::convert(resource.c_str())));
  policyCertInfo->appendChild(temp);

  for (int i=0; i < certAuthInfos.size(); i++) {
    temp = doc->createElement(XMLUtil::convert("CAInfo"));
    certAuthInfos[i].toXML(temp);
    policyCertInfo->appendChild(temp);
  }

  for (int i=0; i < ucIssuerGroups.size(); i++) {
    temp = doc->createElement(XMLUtil::convert("UseCondIssuerGroup"));
    ucIssuerGroups[i].toXML(temp);
    policyCertInfo->appendChild(temp);
  }

  if (attributeDirs.size() > 0) {
    DOMElement* attrDirs = doc->createElement(XMLUtil::convert("AttrDirs"));

    for (int i=0; i < attributeDirs.size(); i++) {
      temp = doc->createElement(XMLUtil::convert("URL"));
      attributeDirs[i].toXML(temp);
      attrDirs->appendChild(temp);
    }

    policyCertInfo->appendChild(attrDirs);
  }

  if (cacheTime != -1) {
    temp = doc->createElement(XMLUtil::convert("CacheTime"));
    ostringstream os;

    os << cacheTime;
    string time = os.str();  // XXX
    temp->appendChild(doc->createTextNode(XMLUtil::convert(time.c_str())));
    policyCertInfo->appendChild(temp);
  }
}

void
PCI::doFromXML(const DOMElement* policyCertInfo) {
  certAuthInfos.clear();
  ucIssuerGroups.clear();
  attributeDirs.clear();
  cacheTime = -1;

  DOMNode* child = policyCertInfo->getFirstChild();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("PolicyCertificateImpl:doFromXML:missing resource");
  }

  DOMNode* temp = child->getFirstChild();

  if (temp  == NULL) {
    throw Exception("PolicyCertificateImpl:doFromXML:missing resource");
  }

  resource = XMLUtil::convert(temp->getNodeValue());
  DOMNodeList *list = policyCertInfo->getElementsByTagName(XMLUtil::convert("CAInfo"));
  int len = list->getLength();

  for (int i=0; i < len; i++) {
    CertificateAuthorityInfo info;

    child = list->item(i);
    info.fromXML((DOMElement*) child);
    certAuthInfos.push_back(info);
  }

  list = policyCertInfo->getElementsByTagName(XMLUtil::convert("UseCondIssuerGroup"));

  len = list->getLength();

  if (len == 0) {
    throw Exception("PolicyCertificateImpl:doFromXML:missing uc grpups");
  }

  for (int i = 0; i < len; i++) {
    UseCondIssuerGroup ucGroup;

    child = list->item(i);
    ucGroup.fromXML((DOMElement *) child);
    ucIssuerGroups.push_back(ucGroup);
  }

  child = child->getNextSibling();

  if (child == NULL) {
    return;
  } else if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("PolicyCertificateImpl:doFromXML:expecting AttrDirs/CacheTime");
  }

  if (XMLString::compareString(child->getNodeName(), XMLUtil::convert("AttrDirs")) == 0) {
    list = ((DOMElement *) child)->getElementsByTagName(XMLUtil::convert("URL"));
    len = list->getLength();

    if (len == 0) {
      throw Exception("PolicyCertificateImpl:doFromXML:missing AttrDirs");
    }

    for (int i = 0; i < len; i++) {
      Directory dir;
      DOMNode* temp = list->item(i);

      dir.fromXML((DOMElement *) temp);
      attributeDirs.push_back(dir);
    }

    if ((child = child->getNextSibling()) == NULL) {
      return;
    } else if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
      throw Exception("PolicyCertificateImpl:doFromXML:expecting CacheTime Element");
    }
  }

  temp = child->getFirstChild();

  if (temp == NULL) {
    throw Exception("PolicyCertificateImpl:doFromXML:expecting CacheTime");
  }

  string rep = XMLUtil::convert(temp->getNodeValue());
  istringstream is(rep);

  if (!(is >> cacheTime)) {
    throw Exception("PolicyCertificateImpl:doFromXML:cacheTime expecting (int)");
  }
}

void
PCI::doParamString(ostream& os) const {
  os << getResource() << ','
     << "CAINFOS"
     << '='
     << '[';

  const vector<CAInfo>& certAuthInfos = getCertificateAuthorityInfos();
  int size = certAuthInfos.size();

  if (size) {
    os << certAuthInfos[0].paramString();
  }

  for (int i = 1; i < size; ++i) {
    os << ',' << certAuthInfos[i].paramString();
  }

  os << ']'
     << ','
     << "UCGROUPS"
     << '='
     << '[';

  const vector<UCGroup>& ucIssuerGroups = getUseCondIssuerGroups();

  size = ucIssuerGroups.size();

  if (size) {
    os << ucIssuerGroups[0].paramString();
  }

  for (int i = 1; i < size; ++i) {
    os << ',' << ucIssuerGroups[i].paramString();
  }
  
  os << ']'
     << ','
     << "AttrDirs"
     << '='
     << '[';

  const vector<Directory>& attrDirs = getAttributeDirectories();

  size = attrDirs.size();

  if (size) {
    os << attrDirs[0].paramString();
  }

  for (int i = 1; i < size; ++i) {
    os << ',' << attrDirs[i].paramString();
  }

  os << ']'
     << ',';

  os << "cacheTime" 
     << '='
     << getCacheTime();
}

void 
PCI::doToText(int indentAmt, ostream& os) const {
  os << Util::fText(indentAmt + 3, "Resource: " + getResource()) << endl;

  const vector<CertificateAuthorityInfo>& cais = getCertificateAuthorityInfos();  
  int size = cais.size();

  if (size > 0) {
     os << Util::fText(indentAmt + 3, "Trusted CAs: ") << endl;

     for (int i = 0; i < cais.size(); ++i) { 
       os << Util::fText(indentAmt + 6, "Name: " + cais[i].getCa().toString()) 
          << endl;

       const vector<Directory>& idDirs = cais[i].getIdentityDirectories();

       size = idDirs.size();

       if (size > 0) {
         os << Util::fText(indentAmt + 6, "Identity Directories: ") << endl;

         for (int j = 0; j < size; ++j) {
           os << Util::fText(indentAmt + 9, idDirs[j].toString()) << endl;
         } 
       }

       const vector<Directory>& crlDirs = cais[i].getCrlDirectories(); 

       size = crlDirs.size();

       if (size > 0) {
         os << Util::fText(indentAmt + 6, "CRL Directories: ") << endl;

         for (int k = 0; k < size; ++k) {
           os << Util::fText(indentAmt + 9, crlDirs[k].toString()); 
         }
      } 
    }

    os << endl;
  }   

  os << Util::fText(indentAmt + 3, "Stakeholders: ") << endl;

  const vector<UseCondIssuerGroup>& ucgs = getUseCondIssuerGroups(); 

  for (int i = 0; i < ucgs.size(); ++i) {
    os << Util::fText(indentAmt + 6, "Stakeholders Group: ") << endl;

    const vector<AkentiPrincipal>& apis = ucgs[i].getUCIssuers(); 

    for (int j = 0; j < apis.size(); ++j) {  
      os << Util::fText(indentAmt + 9, "Name: " +  
                        apis[j].getName().toString()) << endl
         << Util::fText(indentAmt + 9, "CA  : " + 
	                apis[j].getGuarantor().toString()) << endl;

    }

    os << Util::fText(indentAmt + 9, "UseCondition Directories: ") << endl;

    const vector<Directory>& ucDirs = ucgs[i].getDirectories(); 

    for (int k = 0; k < ucDirs.size(); ++k) {
      os << Util::fText(indentAmt + 12, ucDirs[k].toString()) << endl;
    } 
  }

  const vector<Directory>& attrDirs = getAttributeDirectories();
  size = attrDirs.size();

  if (size > 0) {
    os << Util::fText(indentAmt + 3, "Attribute Directories: ") << endl;

    for (int i = 0; i < attrDirs.size(); ++i) {
      os << Util::fText(indentAmt + 9, attrDirs[i].toString()) << endl;
    }
  }

  long cacheTime = getCacheTime();

  if (cacheTime > -1) {
    os << Util::fText(indentAmt + 3, "Maximum cache time:") 
       << cacheTime 
       << " seconds" 
       << endl;
  }
}

} // end of namespace
