#include <sstream>
#include "CertificateAuthorityInfo.hpp"
#include "CertificateFactory.hpp"
#include "AkentiOutputStream.hpp"
#include "AkentiInputStream.hpp"
#include "Base64.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "XMLUtil.hpp"
#include "Xerces-Includes.hpp"

using std::ostringstream;

namespace Akenti {

typedef CertificateAuthorityInfo CAI;

CAI::CertificateAuthorityInfo() {
}

CAI::CertificateAuthorityInfo(const Certificate& cert, 
                              const vector<Directory>& idDirs,
                              const vector<Directory>& crlDirs)
  : certs(1, cert), idDirs(idDirs), crlDirs(crlDirs) {
  if (!cert) {
    throw Error("CertificateAuthorityInfo:constructor:illegal argument(1)");
  }
  
  ca = cert->getPrincipal();
}

CAI::CertificateAuthorityInfo(const vector<Certificate>& certs, 
                              const vector<Directory>& idDirs,
                              const vector<Directory>& crlDirs)
  : certs(certs), idDirs(idDirs), crlDirs(crlDirs) {
  int size = certs.size();

  if (!size) {
    throw Error("CertificateAuthorityInfo:constructor:illegal argument(2)");
  }
  
  for (int i = 0; i < size; ++i) {
    if (!certs[i]) {
      throw Error("CertificateAuthorityInfo:constructor:illegal argument(3)");
    }
  }

  ca = certs[size - 1]->getPrincipal();
}

CAI::~CertificateAuthorityInfo() {
}

const DistinguishedName&
CAI::getCa() const {
  return ca;
}

const vector<Certificate>&
CAI::getChain() const {
  return certs;
}

const vector<Directory>&
CAI::getIdentityDirectories() const {
  return idDirs;
}

const vector<Directory>&
CAI::getCrlDirectories() const {
  return crlDirs;
}

void
CAI::writeObject(AkentiOutputStream& drain) const {
  int size = certs.size();

  if (!size) {
    throw Exception("CertificateAuthorityInfo:writeObject:must have at least one cert");
  }

  drain << ' ' << size;

  CertificateFactory fac = CertificateFactory::getInstance();
  
  for (int i = 0; i < size; ++i) {
    string der = fac->getCertificateEncoded("DER", certs[i]);
    int len = (int) der.size();

    drain << ' ' << len << ' ';
    drain.write((char *) der.data(), len);
  } 

  size = idDirs.size();
  drain << ' ' << size; 

  for (int i = 0; i < size; ++i) {
    drain << ' ' << idDirs[i];
  }

  size = crlDirs.size();
  drain << ' ' << size; 
  
  for (int i = 0; i < size; ++i) {
    drain << ' ' << crlDirs[i];
  }
}

void
CAI::readObject(AkentiInputStream& source) {
  int size;

  source >> size;

  if (size <= 0) {
    throw Exception("CertificateAuthorityInfo:readObject:certs:size<=0");
  }

  CertificateFactory fac = CertificateFactory::getInstance();

  for (int i = 0; i < size; ++i) {
    int len;

    source >> len;

    if (source.get() != ' ') {
      throw Exception("CertificateAuthorityInfo:readObject: expecting blank space");
    }

    string der = string(len, '\0');

    source.read((char*) der.data(), len);

    Certificate cert = fac->initCertificate("DER", der); 

    certs.push_back(cert);
  }

  ca = certs[size - 1]->getPrincipal();

  source >> size;

  if (size < 0) {
    throw Exception("CertificateAuthorityInfo:readObject:id dirs:size<=0");
  }

  for (int i = 0; i < size; ++i) {
    Directory dir;

    source >> dir;

    idDirs.push_back(dir);
  }

  source >> size;

  if (size < 0) {
    throw Exception("CertificateAuthorityInfo:readObject:crl dirs:size<=0");
  }

  for (int i = 0; i < size; ++i) {
    Directory dir;

    source >> dir;

    crlDirs.push_back(dir);
  }
}

void 
CAI::toXML(DOMElement* caInfo) const {
  DOMDocument *doc = caInfo->getOwnerDocument();
  DOMElement *temp = doc->createElement(XMLUtil::convert("CADN"));
  string caDN = ca.toString();
  
  temp->appendChild(doc->createTextNode(XMLUtil::convert(caDN.c_str())));
  caInfo->appendChild(temp);
  
  CertificateFactory fac = CertificateFactory::getInstance();
  
  for (int i = 0; i < certs.size(); i++) {
    string enc = Base64::encode(fac->getCertificateEncoded("DER", certs[i]));

    for (int i = 64; i < enc.size(); i += 65) {
      enc[i] = ' ';
    }

    temp = doc->createElement(XMLUtil::convert("X509Certificate"));
    temp->appendChild(doc->createTextNode(XMLUtil::convert(enc.c_str())));
    caInfo->appendChild(temp);
  }
  
  if (idDirs.size() > 0) {
    DOMElement *idDirsElement = doc->createElement(XMLUtil::convert("IdDirs"));

    for (int i = 0; i < idDirs.size(); i++) {
      temp = doc->createElement(XMLUtil::convert("URL"));
      idDirs[i].toXML(temp);
      idDirsElement->appendChild(temp);
    }

    caInfo->appendChild(idDirsElement);
  }

  if (crlDirs.size() > 0) {
    DOMElement *crlDirsElement = doc->createElement(XMLUtil::convert("CRLDirs"));

    for (int i = 0; i < crlDirs.size(); i++) {
      temp = doc->createElement(XMLUtil::convert("URL"));
      crlDirs[i].toXML(temp);
      crlDirsElement->appendChild(temp);
    }

    caInfo->appendChild(crlDirsElement);
  }
}

void 
CAI::fromXML(const DOMElement* caInfo) {
  certs.clear();
  idDirs.clear();
  crlDirs.clear();

  DOMNode *child = caInfo->getFirstChild();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("Certificate Authority Infos:fromXML:missing ca");
  }
  
  DOMNode *temp = child->getFirstChild();
  
  if (temp == NULL) {
    throw Exception("Certificate Authority Infos:fromXML:missing ca");
  }

  ca = XMLUtil::convert(temp->getNodeValue());

  DOMNodeList* list = caInfo->getElementsByTagName(XMLUtil::convert("X509Certificate"));
  int len = list->getLength();

  if (!len) {
    throw Exception("Certificate Authority Infos:fromXML:NO CA INFOS");
  }
  
  CertificateFactory fac = CertificateFactory::getInstance();

  for (int i = 0; i < len; i++) {
    child = list->item(i);

    if ((temp = child->getFirstChild()) == NULL) {
      throw Exception("CertificateAuthorityInfo:fromXML:missing x509 cert");
    }
    
    // OPENSL B64 REQUIRES LINES OF 64 CHARACTERS 
    // XXXX 
    string cert = XMLUtil::convert(temp->getNodeValue());
    
    if (cert[0] == '\n') {
      cert = cert.substr(1);
    }

    for (int i = 64; i < cert.size(); i += 65) {
      cert[i] = '\n';
    }

    cert = Base64::decode(cert);
    certs.push_back(fac->initCertificate("DER", cert));
  }

  child = child->getNextSibling();

  if (child == NULL) {
    return;
  }

  if (XMLString::compareString(child->getNodeName(), XMLUtil::convert("IdDirs")) == 0) {
    list = ((DOMElement*) child)->getElementsByTagName(XMLUtil::convert("URL"));

    if ((len = list->getLength()) == 0) {
      throw Exception("CertificateAuthorityInfo:fromXML:missing idDir");
    }
    
    for (int i = 0; i < len; i++) {
      temp = list->item(i);
      Directory dir;

      dir.fromXML((DOMElement*) temp);
      idDirs.push_back(dir);
    }

    if ((child = child->getNextSibling()) == NULL) {
      return;
    } else if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
      throw Exception("CertificateAuthorityInfo:fromXML:expecting CRL DIRS"); 
    }
  }

  list = ((DOMElement*) child)->getElementsByTagName(XMLUtil::convert("URL"));

  if ((len = list->getLength()) == 0) {
    throw Exception("CertificateAuthorityInfo:fromXML:missing CRL dir");
  }

  for (int i = 0; i < len; i++) {
    temp = list->item(i);
    Directory dir;

    dir.fromXML((DOMElement*) temp);
    crlDirs.push_back(dir);
  }
}

string
CAI::paramString() const {
  const vector<Certificate>& certs = getChain();

  int size = certs.size();

  if (!size) {
    return "[CACERT=NOCERT]";
  }

  ostringstream os;

  os << '['
     << getCa().paramString(),
  os << ','
     << '[';
  
  os << "SN=" << certs[size - 1]->getSerialNumber();

  for (int i = size - 2; i >= 0; --i) {
    os << ',' << "SN=" << certs[i]->getSerialNumber();
  }
  

  os << ']'
     << ','
     << '[';

  const vector<Directory>& idDirs = getIdentityDirectories();

  size = idDirs.size();

  if (size > 0) {
    os << idDirs[0].paramString();
  } 

  for (int i = 1; i < size; ++i) {
    os << ',' << idDirs[i].paramString();
  }

  os << ']'
     << ','
     << '[';

  const vector<Directory>& crlDirs = getCrlDirectories();

  size = crlDirs.size();

  if (size > 0) {
    os << crlDirs[0].paramString();
  } 

  for (int i = 1; i < crlDirs.size(); ++i) {
    os << crlDirs[i].paramString();
  }

  os << ']'
     << ']';

  return os.str();
}

}
