#include "CertificateRetrievalFactory.hpp"
#include "HttpCertificateRetrievalImpl.hpp"
#include "FileCertificateRetrievalImpl.hpp"
#include "LdapCertificateRetrievalImpl.hpp"
#include "DirectoryServiceException.hpp"
#include "Certificate.hpp"
#include "AkentiCertificate.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Directory.hpp"
#include "URL.hpp"

using namespace Akenti;

typedef CertificateRetrievalFactory CRF;

CertificateRetrievalFactory CRF::fac = CertificateRetrievalFactory();

CRF::CertificateRetrievalFactory() {
  addCertificateRetrieval("http", new HttpCertificateRetrievalImpl);
  addCertificateRetrieval("file", new FileCertificateRetrievalImpl);
  addCertificateRetrieval("ldap", new LdapCertificateRetrievalImpl);
}

CRF::CertificateRetrievalFactory(const CertificateRetrievalFactory &fac) {
  this->crs = fac.crs;
}

CRF::~CertificateRetrievalFactory() {
}

void 
CRF::addCertificateRetrieval(const string& protocol, CertificateRetrievalImpl* cri) {
  if (!cri || !protocol.length()) {
    throw Error( "CertificateRetrievalFactory:addCertificateRetrieval:illegal argument");
  }

  crs[protocol] = cri;
}

const CertificateRetrieval&
CRF::getCertificateRetrieval(const string& protocol) {
  if (!crs[protocol]) {
    throw DirectoryServiceException(
      "CertificateRetrievalFactory:getCertificateRetrieval:CR not found for :"
      + protocol
    );
  }
  
  return crs[protocol]; 
}

CertificateRetrievalFactory&
CRF::getCertificateRetrievalFactory() {
  return fac;
}

vector<Certificate>
CRF::getCertificates(const Directory& dir, map<string, string>& args) {
  vector<Certificate> certs;
  URL url;
  
  if (dir.parseDir(url)) {
    const CertificateRetrieval& cr = fac.getCertificateRetrieval(url.getProtocol()); 

    cr->getCertificates(url, args, certs);
  } else {
    // LOG THIS EXCEPTION
  }

  return certs;
}

vector<Certificate>
CRF::getCertificates(const URL& url, map<string, string>& args) {
  vector<Certificate> certs;

  try {
    const CertificateRetrieval& cr = fac.getCertificateRetrieval(url.getProtocol()); 
    
    cr->getCertificates(url, args, certs);
  } catch(Exception& ex) {
    // LOG THIS EXCEPTION
  }
  
  return certs;
}

vector<AkentiCertificate>
CRF::getAkentiCertificates(const Directory& dir, map<string, string>& args) {
  vector<AkentiCertificate> certs;
  URL url;

  if (dir.parseDir(url)) {
    const CertificateRetrieval& cr = fac.getCertificateRetrieval(url.getProtocol()); 

    cr->getAkentiCertificates(url, args, certs);
  } else {
    // LOG THIS EXCEPTION
  }

  return certs;
}

vector<AkentiCertificate>
CRF::getAkentiCertificates(const URL& url, map<string, string>& args) {
  vector<AkentiCertificate> certs;

  try {
    const CertificateRetrieval& cr = fac.getCertificateRetrieval(url.getProtocol()); 
    
    cr->getAkentiCertificates(url, args, certs);
  } catch(Exception& ex) {
    // LOG THIS EXCEPTION
  }

  return certs;
}

