#include "ExpressionParser.hpp"
#include "Expression.hpp"
#include "ParserException.hpp"
#include "TreeNode.hpp"
#include "Error.hpp"

using namespace Akenti;

/**
  * Grammar:
  *
  * EXPR      ->  AEXPR   { || AEXPR }
  * AEXPR     ->  NEXPER  { && NEXPR }
  * NEXPER    ->  { ! NEXPR } | PEXPER 
  * PEXPER    ->  TEXPER AOP TEXPER | ( EXPER ) 
  * TEXPER    ->  { string }
  * AOP       ->  "=" | "!=" | "<" | "<=" | ">" | ">="
  *
  */
ExpressionParser::ExpressionParser() {
}

ExpressionParser::~ExpressionParser() {
}

TreeNode*
ExpressionParser::parseExpression(const string& expression) {
  scanner.tokenize(expression);
  scanner.nextToken();
  TreeNode* n = orExpression();

  if (scanner.getTokenType() != INVALID) {
    throw ParserException(
      "ExpressionParser:parseExpression:did not expect a right paren"
    );
  }
  
  return n;
}

TreeNode*
ExpressionParser::orExpression() {
  TreeNode* node = andExpression();

  if (scanner.getTokenType() == OR) {
    scanner.nextToken();

    if (scanner.getTokenType() == INVALID) {
      throw ParserException(
        "ExpressionParser:orExpression:expecting an expression"
      );
    }

    node = new TreeNode(OR, "||", node, orExpression());
  }
  
  return node;
}

TreeNode*
ExpressionParser::andExpression() {
  TreeNode* node = notExpression();
  
  if (scanner.getTokenType() == AND) {
    scanner.nextToken();

    if (scanner.getTokenType() == INVALID) {
      throw ParserException(
        "ExpressionParser:andExpression:expecting an expression"
      );
    }

    node = new TreeNode(AND, "&&",  node, andExpression());
  }
  
  return node;
}

TreeNode*
ExpressionParser::notExpression() {
  if (scanner.getTokenType() == NOT) {
    scanner.nextToken();

    if (scanner.getTokenType() == INVALID) {
      throw ParserException(
        "ExpressionParser:notExpression:expecting an expression"
      );
    }

    return new TreeNode(NOT, "!", notExpression(), NULL);
  } else {
    return parenExpression();
  }
}

TreeNode*
ExpressionParser::parenExpression() {
  TreeNode* node;

  if (scanner.getTokenType() == LPAREN) {
    scanner.nextToken();
    
    if (scanner.getTokenType() == RPAREN) {
      node = new TreeNode(ENULL, "");
    } else {
      node = orExpression();
    
      if (scanner.getTokenType() != RPAREN) {
        throw ParserException("ExpressionParser:parenExpression:expecting right paren");
      } 
    }
  } else if (scanner.getTokenType() == CONST) {
    string attribute = scanner.getLexeme();

    scanner.nextToken();

    TokenType type = scanner.getTokenType();

    if (type < GR || type > NEQ) {
      throw ParserException(
        "ExpressionParser:parenExpression:expecting an equality operator"
      );
    }

    string op = scanner.getLexeme();

    scanner.nextToken();

    if (scanner.getTokenType() != CONST) {
      throw ParserException(
        "ExpressionParser:parenExpression:expecting a value for " + attribute
      );
    }

    string value = scanner.getLexeme();
    
    node = new TreeNode(
      type, 
      op ,
      new TreeNode(CONST, attribute),
      new TreeNode(CONST, value)
    );

  } else if (scanner.getTokenType() == INVALID) {
    node = new TreeNode(ENULL, "");
  } else {
    throw ParserException(
      "ExpressionParser:parenExpression:expecting a left paren or an attr"
    );
  }

  scanner.nextToken();
  return node;
}
