#include <sstream> 
#include <map> 
#include "AKRuntime.hpp"
#include "URL.hpp"
#include "Directory.hpp"
#include "Properties.hpp"
#include "CertificateAuthorityInfo.hpp"
#include "AkentiX509CertificateImpl.hpp"
#include "Exception.hpp"

typedef AKRuntime AR;
typedef CertificateAuthorityInfo CAInfo;
typedef AkentiX509CertificateImpl X509Impl;

using std::istringstream;
using std::map;

namespace Akenti {

class RuntimePolicy : public CertificateStatusListener {

public:

  /**
    * Constructs this object.
    */
  RuntimePolicy(const Policy& policy, RuntimePolicy* parent) 
    : policy(policy), parent(parent) {
    caHoldersSet = attrURlsSet = 0;
  }

  /*
   * Here we must not perform this first branch more than once.
   * We cannot store the parents in the local for they may change
   * and we will be considering certificates that failed more than once.
   *
   * We could be a bit more sneakier but the certificate verifier 
   * will get a bit complicated.
   *
   * So for simplicity we just dump the certs in the cache.
   * And this does not hurt anything except more certs to dump
   * in the external cache. We do however make an exception for
   * root policy certs. And why not?
   *
   * caHoldersSet = 0 when first called, 
   * caHoldersSet = 1 return this->caHolders
   * caHoldersSet = 2 return parent.getCaCerts(...)
   */
  const vector<AKCertHolder>& 
  getCaCerts(CertificateCache& certCache, int cacheTime, int& endLevel) {
    if (caHoldersSet == 0) {
      const vector<CAInfo>& caInfos = policy->getCertificateAuthorityInfos();

      caHoldersSet = int(caInfos.size() != 0);

      if (caHoldersSet == 1) {
        UtcTime start, end(cacheTime);

        for (int i = 0; i < caInfos.size(); ++i) {
          const vector<Certificate>& caChain = caInfos[i].getChain();
          int caIndex = caChain.size() - 1;
          AkentiCertificate cert = new X509Impl(caChain[caIndex]);
          AKCertHolder holder;

          if (parent == NULL) {
            holder = new AKCertHolderImpl(cert, 1, start, end);
          } else {
            holder = certCache.addCertificate(cert, 0, start, end);
            holder->addCertificateStatusListener(this);
            // Dump the rest (if any) in the cache.
            while (--caIndex >= 0) {
              cert = new X509Impl(caChain[caIndex]);
              certCache.addCertificate(cert, 0, start, end);
            }
          }
          
          // XXX CHECK FOR DUPS
          caHolders.push_back(holder);
          
          vector<URL>& temp = idUrlsTable[caInfos[i].getCa().toString()];
          const vector<Directory>& idDirs = caInfos[i].getIdentityDirectories();

          // DO NOT ALLOW DUPS  

          temp.clear();

          URL url;

          for (int i = 0; i < idDirs.size(); ++i) { 
            if (idDirs[i].parseDir(url)) {
              temp.push_back(url);
            }
          }
        }
      }
    }
     
    if (caHoldersSet != 1) {
      return parent->getCaCerts(certCache, cacheTime, --endLevel);
    }

    return caHolders;
  }
    
  /**
    *
    */
  vector<URL>& getIdentityUrls() {
    if (caHoldersSet == 0) {
      throw Error("RuntimePolicy:getIdentityUrls:SHOULD NOT HAPPEN");
    } else if (caHoldersSet == 2) {
      return parent->getIdentityUrls();
    } else if (!idUrls.size() && idUrlsTable.size()) {
      std::map<string, vector<URL> >::iterator iter;

      for (iter = idUrlsTable.begin(); iter != idUrlsTable.end(); ++iter) {
        vector<URL>& urls = iter->second;
        
        for (int i = 0; i < urls.size(); ++i) {
          idUrls.push_back(urls[i]);
        }
      } 
    }

    return idUrls;
  }

  /**
    */
  void
  verifyingCertFailed(const AkentiCertificate& cert) {
    int i = caHolders.size() - 1;

    for (;i >= 0 && caHolders[i]->getCertificate() != cert; --i) {
    }

    caHolders[i]->removeCertificateStatusListener(this);
    caHolders.erase(caHolders.begin() +i );

    // XXX This is one way, we can do better or can we?
    const X509Impl* impl = (const X509Impl*) cert.get(); 

    idUrlsTable.erase(impl->getIssuer().getName().toString());
    idUrls.clear();
  }

  /**
    * The cacheTime is inherited from the parent 
    * if and only if this policy.cacheTime < 0.
    *
    * Note that if the root policy also has a negative
    * cacheTime we take it from the config file.
    * if that fails we return 0.
    *
    * In other words We assume that a negative cacheTime 
    * at the root is a don't care and at other levels
    * it means to be inherited. 
    */ 
  long 
  getCacheInterval() {
    int cacheTime = policy->getCacheInterval();
   
    if (cacheTime >= 0) {
      return cacheTime;
    } else if (parent != NULL) { 
      return parent->getCacheInterval(); 
    } else {
      Properties& props = Properties::getSystemProperties();
      string temp = props.getValue("CacheTime", "0");
      istringstream is(temp);

      is >> cacheTime;
      return (cacheTime < 0) ? 0 : cacheTime;
    }
  }

  /* 
   * We want to be lazy, not parse more than once and not hold on
   * to the urls in more than one place.
   *
   * Thus we have to be a little sneaky.
   *
   * Note we only inherit the parent if the policy does
   * not contain any dirs. The fact that they parse or 
   * not is irrelevant.
   *
   * attrUrlSet = 0 when first called, 
   * attrUrlSet = 1 return this->attrUrls
   * attrUrlSet = 2 return parent.getAttributeUrls()
   *
   */
  vector<URL>& getAttributeUrls() {
    if (attrURlsSet == 0) {
      const vector<Directory>& attrDirs = policy->getAttributeDirectories();
      
      attrURlsSet = int(attrDirs.size() != 0);
      URL url;

      for (int i = 0; i < attrDirs.size(); ++i) { 
        if (attrDirs[i].parseDir(url)) {
          attrUrls.push_back(url);
        }
      }
    } 

    if (attrURlsSet != 1) {
      attrURlsSet = 2;
      
      if (parent) {
        return parent->getAttributeUrls();
      }
    }
   
    return attrUrls;
  }

  Policy& getPolicy() {
    return policy;
  }

  string paramString() const {
    return policy->paramString();
  }

private:

  Policy policy;

  RuntimePolicy* parent;
  
  vector<AKCertHolder> caHolders;
  int caHoldersSet;  

  vector<URL> attrUrls;
  int attrURlsSet;    

  map<string, vector<URL> > idUrlsTable;
  vector<URL> idUrls;

};  // end of class RuntimePolicy

}; // end of namespace	

AR::AKRuntime(bool stop) 
  : stop(stop), cacheCap(true), ctx(new PolicyContextImpl), certVerifier(this), certCache() {
}

AR::~AKRuntime() {
  for (int i = 0; i < runtimePolicies.size(); ++i) {
    delete(runtimePolicies[i]);
    runtimePolicies[i] = NULL;
  }
}

const string&
AR::getResource() {
  return ctx->getResource();
}

void
AR::setResource(const string& resource) {
  ctx->setResource(resource);
}

void
AR::setResourceAltName(const string& resAltName) {
  ctx->setResourceAltName(resAltName);
}

const AkentiPrincipal& 
AR::getAkentiPrincipal() {
  return ctx->getAkentiPrincipal();
}

void
AR::setAkentiPrincipal(const AkentiPrincipal& ap) {
  ctx->setAkentiPrincipal(ap);
}

void 
AR::setSubjectCert(const AkentiX509Certificate& cert) {
  ctx->setSubjectCert(cert);
}

void 
AR::setCapabilityCert(const CapabilityCertificate& cert) {
  ctx->setCapabilityCert(cert);
}

void 
AR::addUnconditionalAction(const string& uncondionalAction) {
  ctx->addUnconditionalAction(uncondionalAction);
}

void 
AR::addConditionalActions(const ConditionalActions& condActions) {
  ctx->addConditionalActions(condActions);
}

const vector<Policy>& 
AR::getPolicies() {
  return ctx->getPolicies();
}

const PolicyContext& 
AR::getPolicyContext() {
  return ctx;
}

CertInfoCollector& 
AR::getCertInfoCollector() {
  return cinfoCol;
}

const set<string>& 
AR::getDesiredActions() {
  return ctx->getDesiredActions();
}

void 
AR::setDesiredActions(const set<string>& desiredActions) {
  ctx->setDesiredActions(desiredActions);
}

const multimap<string, string>& 
AR::getRestrictedAttributes() {
  return ctx->getRestrictedAttributes();
}

void
AR::setRestrictedAttributes(const multimap<string, string>& desiredAttributes) {
  ctx->setRestrictedAttributes(desiredAttributes);
}

bool 
AR::isAttributeDesired(const string& attribute, const string& value) const {
  return ctx->isAttributeDesired(attribute, value);
}

bool 
AR::isActionDesired(const string& action) const {
  return ctx->isActionDesired(action);
}

void 
AR::setCacheCap(bool cacheCap) {
  this->cacheCap = cacheCap;
}
 
bool 
AR::getCacheCap() const {
  return this->cacheCap;
}

void 
AR::setExternalEvaluator(const ExternalEvaluator& externalEvaluator) {
  this->externalEvaluator = externalEvaluator;
}

void
AR::addAkentiX509Cert(const AkentiX509Certificate& x509Cert) {
  ctx->addAkentiX509Cert(x509Cert);
}

void
AR::addAttributeCert(const AttributeCertificate& attrCert) {
  ctx->addAttributeCert(attrCert);
}

ExternalEvaluator& 
AR::getExternalEvaluator() {
  return externalEvaluator;
}

const vector<AKCertHolder>& 
AR::getCACerts(int startLevel, int& endLevel) {
  // XXX NEED TO THINK ABOUT CACHE TIME
  RuntimePolicy* runtimePolicy = runtimePolicies[startLevel - 1];
  int cacheTime = runtimePolicy->getCacheInterval();

  endLevel = startLevel;
  return runtimePolicy->getCaCerts(certCache, cacheTime, endLevel);
}

int 
AR::getCALevel(int startLevel) {
  int caLevel;

  getCACerts(startLevel, caLevel);
  return caLevel;
}

const vector<URL>& 
AR::getIdentityURLS(int level) {
  return runtimePolicies[level - 1]->getIdentityUrls();
}

const vector<URL>& 
AR::getAttributeURLS() {
  return runtimePolicies[level - 1]->getAttributeUrls();
}
  
long 
AR::getCacheInterval() {
  return runtimePolicies[level - 1]->getCacheInterval();
}

void 
AR::addPolicy(const Policy& policy) {
  ctx->addPolicy(policy);

  int size = runtimePolicies.size();
  RuntimePolicy* runtimePolicy = (size == 0) ? NULL : runtimePolicies[size - 1];

  runtimePolicies.push_back(new RuntimePolicy(policy, runtimePolicy));
}

Policy& 
AR::getPolicy(int index) {
  return runtimePolicies[index]->getPolicy();
}

int 
AR::numOfPolicies() {
  return runtimePolicies.size();
}

void 
AR::setLevel(int level) {
  this->level = level;
}

int 
AR::getLevel() {
  return level;
}

CertificateCache& 
AR::getCertificateCache() {
  return certCache;
}

CertificateVerifier& 
AR::getCertificateVerifier() {
  return certVerifier;
}

bool
AR::shouldStop() const {
  return false;
}

string
AR::paramString() const {
  string ret;

  ret += "***** BEGIN RUNTIME *****";
  ret += '\n';

  for (int i = 0; i < runtimePolicies.size(); ++i) {
    ret += runtimePolicies[i]->paramString();
  }
  
  ret += '\n';
  ret += "***** END RUNTIME *****";
  return ret;
}

AKRuntime*
AR::createAKRuntime() {
  return new AKRuntime;
}

