#include "AkentiInit.hpp"
#include "AkentiEngine.hpp"
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "ExternalCache.hpp"
#include "FileExternalCacheImpl.hpp"
#include "ServerExternalCacheImpl.hpp"
#include "ServerContext.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "IOException.hpp"
#include "AkentiConstants.hpp"
#include "XMLUtil.hpp"

namespace Akenti {

bool
AkentiInit::init(const string& confFile, string& message) {
  try {
    message = "BEGIN ";
    Properties& props = Properties::getSystemProperties();
  
    message += "loading config file ";
    props.load(confFile);
    message += "OK";
    
    string temp;
    bool ret = init(temp);
    
    message += ' ' + temp;
    return ret;
  } catch(IOException& ex) {
    message += "FAIL";
    return false;
  }
}

bool
AkentiInit::init(string& message) {
  bool ret = true;

  try {
    Properties& props = Properties::getSystemProperties();

    if (!Provider::isProviderInstalled()) {
      message += "installing security provider ";
      Provider provider = Provider(new OpenSSLProviderImpl);

      provider->loadDigestAlgorithms();
      provider->loadSignatureAlgorithms();
      Provider::setProvider(provider);
      message += "OK";
    }

    ExternalCache::setExternalCache(NULL);
    string caching = props.getValue("Caching", "off");

    if (StringUtil::equalsIgnoreCase(caching, "on")) {
      message += " setting the cache ";

      string cacheType = props.getValue("CacheType", CacheType);
      
      if (StringUtil::equalsIgnoreCase(cacheType, "file")) {
        ExternalCache::setExternalCache(new FileExternalCacheImpl);
        message += "OK";
      } else if (StringUtil::equalsIgnoreCase(cacheType, "server")) {
        ExternalCache::setExternalCache(new ServerExternalCacheImpl);
        message += "OK";
      } else {
        message += "FAIL:cache type " + cacheType + " unsupported";
        ret = false;
      }
    } else {
      message += "WARNING:cache is not on ";
    }

    ServerContext& sc = ServerContext::getServerContext();
   
    sc.unloadCredentials();

    string signCerts = props.getValue("SignCerts", "off");

    if (StringUtil::equalsIgnoreCase(signCerts, "on")) {
      message += " setting server's cert ";
      sc.initCertificateFromFile("PEM", props.getValue("ServerCertFile", ServerCertFile));
      message += "OK";
      message += ' ';
      message += "setting server's key ";
      sc.initPrivateKeyFromFile("PEM",  props.getValue("ServerKeyFile", ServerKeyFile));
      message += "OK";
    } else {
      message += "WARNING:Certs will not be signed.";
    }

    // load the resource map if one exists
    string resourceMap =  props.getValue("ResourceMap");

    if (resourceMap != ""){
      message += " loading ResourceMap at " + resourceMap;
      AkentiEngine::loadResourceMap(resourceMap);
    }

    message += "initializing XML library ";
    XMLUtil::init();
    message += "OK";
  } catch(Exception& ex) {
    message += "FAIL(" + ex.getMessage() + ")";
    ret = false;
  }
  
  return ret;
}

}
