#include <sstream>
#include <iostream>
#include "PolicyAuthorizer.hpp" 
#include "AKRuntime.hpp"
#include "CertificateRetrievalFactory.hpp"
#include "CertificateVerifier.hpp"
#include "CertificateCache.hpp"
#include "AttributeCertificate.hpp"
#include "ConditionalActions.hpp"
#include "ExpressionEvaluator.hpp"
#include "Expression.hpp"
#include "MessageDigest.hpp"
#include "AKCertHolder.hpp"
#include "DistinguishedName.hpp"
#include "Directory.hpp"
#include "URL.hpp"
#include "Logger.hpp"
#include "Util.hpp"

using namespace Akenti;

typedef CertificateRetrievalFactory CRF;
typedef UseCondCertificateImpl UCImpl;

namespace Akenti {

class Evaluator : public ExpressionEvaluator , LogInterface {

public: 
  
  Evaluator(AKRuntime* akRuntime) : akRuntime(akRuntime) {
  }

  void 
  setAttributeInfos(const vector<AttributeInfo>& attributeInfos) {
    newAttributeInfos.clear();
    this->attributeInfos = attributeInfos;
  }
  
  const vector<AttributeInfo>&
  getAttributeInfos() {
    return newAttributeInfos;
  }

  int
  reduce(TokenType op, const string& attribute, const string& value, Expression& expr) {
    bLog(
      LOG_MON,
      "EVAL_ATTRIBUTE",
      "ATTRIBUTE=" + attribute + " VALUE=" + value
    );

    int index = -1;
    
    for (int i = 0; index == -1 && i < attributeInfos.size(); ++i) {
      if (attributeInfos[i].hasAttribute(attribute) && attributeInfos[i].hasValue(value)) {
        index = i;
      }
    }
    
    if (index == -1) {
      throw Error("PolicyAuthorizer:Evaluator:reduce:NO ATTR INFO (SHOULD NOT HAPPEN");
    } 

    const AkentiPrincipal& ap = akRuntime->getAkentiPrincipal();
    const DistinguishedName& ca = ap.getGuarantor();
    const DistinguishedName& user = ap.getName();

    switch(attributeInfos[index].getType()) {
      case AttributeInfo::X509:
        if (!akRuntime->isAttributeDesired(attribute, value)) {
          fLog(LOG_MON,"EVAL_ATTRIBUTE","ATTR_VALUE_DESIRED=NO");
          return 0;
        } else if (!attributeInfos[index].hasCA(ca)) {
          fLog(
            LOG_MON,
            "EVAL_ATTRIBUTE",
            "USER_CA_CHECK=FAIL USER_CA_"+ ca.toLogString()
          );
          return 0;
        } else if (!user.containsAttributeValue(attribute, value)) {
          fLog(
            LOG_MON,
            "EVAL_ATTRIBUTE",
            "USER_DN=DOES_NOT_CONTAIN_ATTRIBUTE " "USER_" + user.toLogString()
          );

          return 0;
        }

        sLog(LOG_MON, "EVAL_ATTRIBUTE");
        return 1;
      case AttributeInfo::AKENTI: {
        if (!akRuntime->isAttributeDesired(attribute, value)) {
          fLog(LOG_MON,"EVAL_ATTRIBUTE","ATTR_VALUE_DESIRED=NO");
          return 0;
        }

        MessageDigest md = MessageDigest::getInstance("MD5");

        md->update(Util::getHashDN(user));
        md->update(Util::getHashDN(ca));
        md->update(attribute);
        md->update(value);

        AKCertHolder holder = getAttributeInfoCertHolder(
          md->digest(), attributeInfos[index]
        );

        if (!holder) {
          fLog(LOG_MON,"EVAL_ATTRIBUTE","FECTH_ATTR_CERT=FAIL"); 
          return 0;
        }

        CertificateVerifier& certVerifier = akRuntime->getCertificateVerifier();

        if (!certVerifier.verify(holder)) {
          fLog(LOG_MON, "EVAL_ATTRIBUTE", "TRUST_CHECK=FAIL"); 
          return 0;
        } 

        AttributeCertificate cert = (AttributeCertificate&) holder->getCertificate();

        akRuntime->addAttributeCert(cert);

        if (!cert->containsCondition()) {
          sLog(LOG_MON, "EVAL_ATTRIBUTE");
          return 1;
        }

        bLog(LOG_MON, "EVAL_CONDITION");
        const Condition& condition = cert->getCondition();
        Expression temp;

        if (!condition.parseConstraint(temp)) {
          throw Error("PolicyAuthorizer:Evaluator:reduce:ATTR CERT BAD CONSTRAINT");
        }

        vector<AttributeInfo> attributeInfos = this->attributeInfos;

        this->attributeInfos = condition.getAttributeInfos();

        Expression newExpression;
        int ret = temp.reduce(*this, newExpression);
        
        this->attributeInfos = attributeInfos;
          
        switch(ret) {
          case -1:
            expr = newExpression;
            fLog(LOG_MON, "EVAL_ATTRIBUTE", "CANNOT HANDLE ATTR=YES");
            break;
          case 1:
            sLog(LOG_MON, "EVAL_ATTRIBUTE");
            break;
          case 0:
            fLog(LOG_MON, "EVAL_ATTRIBUTE");
            break;
          default:
            throw Error("PolicyAuthorizer:Evaluator:reduce:BAD RETURN VALUE");
        }

        return ret;
      } 

      case AttributeInfo::SYSTEM: {
        ExternalEvaluator& externalEvaluator = akRuntime->getExternalEvaluator();
        
        if (!externalEvaluator.canHandleSystemAttribute(attribute)) {
          newAttributeInfos.push_back(attributeInfos[index]);
          fLog(LOG_MON, "EVAL_ATTRIBUTE", "CANNOT_HANDLE_ATTR=YES");
          return -1;
        } 
        
        bool ret = externalEvaluator.evaluateSystemAttribute(op, attributeInfos[index]);

        if (ret) {
          sLog(LOG_MON, "EVAL_ATTRIBUTE");
          return 1;
        } else {
          fLog(LOG_MON, "EVAL_ATTRIBUTE");
          return 0;
        } 
      }

      case AttributeInfo::EXT_AUTH: {
        ExternalEvaluator& externalEvaluator = akRuntime->getExternalEvaluator();
        
        if (!externalEvaluator.canHandleExternalAttribute(attribute)) {
          newAttributeInfos.push_back(attributeInfos[index]);
          fLog(LOG_MON, "EVAL_ATTRIBUTE", "CANNOT_HANDLE_ATTR=YES");
          return -1;
        } 
        
        bool ret = externalEvaluator.evaluateExternalAttribute(op, attributeInfos[index]);

        if (ret) {
          sLog(LOG_MON, "EVAL_ATTRIBUTE");
          return 1;
        } else {
          fLog(LOG_MON, "EVAL_ATTRIBUTE");
          return 0;
        } 
      }
    } 
  } 
  
  AKCertHolder
  getAttributeInfoCertHolder(const string& hash, const AttributeInfo& attribute) {
    int cacheTime = akRuntime->getCacheInterval();
    CertificateCache& certCache = akRuntime->getCertificateCache();
   
    bLog(LOG_L3,"TRY_CACHED_ATTR_CERTS");

    vector<AKCertHolder> holders = certCache.getAKCertHolders(
      hash, "Attribute", UtcTime(-1 * cacheTime) 
    );
    
    for (int i = 0; i < holders.size(); ++i) {
      AkentiCertificate cert = holders[i]->getCertificate();

      if (attribute.hasIssuer(cert->getIssuer().getIssuer())) {
        sLog(LOG_L3,"TRY_CACHED_ATTR_CERTS");
        return holders[i];
      }
    }

    fLog(LOG_L3,"TRY_CACHED_ATTR_CERTS");

    map<string, string> args;

    args["type"] = "Attribute";
    args["base"] = akRuntime->getAkentiPrincipal().getName().toString(); 
    args["hash"] = Util::createFileName(hash);

    bLog(LOG_L3,"TRY_ATTR_DIRS", "FILE_HASH_NAME=" + args["hash"]);

    vector<URL> attrURLS;
    bool ok = attribute.parseDirs(attrURLS);

    if (!ok) {
      iLog(LOG_L3, "INFO", "PARSING_ISSUER_URLS=failed was not able to parse all attr urls");
    }

    const vector<URL>& 
      urls = ok && !attrURLS.size() ? akRuntime->getAttributeURLS() : attrURLS;
    
    UtcTime now, endCacheTime(cacheTime);
    AKCertHolder holder;
    std::ostringstream os;
    int size = urls.size();

    os << size;

    iLog(LOG_L3, "INFO", "NB_OF_ATTR_DIRS=" + os.str());

    for (int j = 0; !holder && j < urls.size(); ++j) {
      bLog(LOG_L3, "FETCH_ATTR_CERTS", "DIR=" + urls[j].toString());

      vector<AkentiCertificate> certs = CRF::getAkentiCertificates(urls[j], args);
      int size = certs.size();
      std::ostringstream os;
      
      os << size;

      if (!size) { 
        fLog(LOG_L3, "FETCH_ATTR_CERTS");
      } else {
        sLog(LOG_L3, "FETCH_ATTR_CERTS", "NB_OF_ATTR_CERTS_FOUND=" + os.str());
      }

      for (int i = 0; !holder && i < size; ++i) {
        bLog(LOG_L3, "CHECK_ATTR_CERT", certs[i]->toLogString());

        AttributeCertificate cert = (AttributeCertificate&) certs[i];

        if (certs[i]->hashCode() != hash) {
          fLog(LOG_L3,"CHECK_ATTR_CERT", "HASH_CHECK=FAIL");
        } else if (!cert->checkValidity(now)) {
          fLog(LOG_L3,"CHECK_ATTR_CERT", "TIME_CHECK=FAIL");
        } else if (!attribute.hasIssuer(certs[i]->getIssuer().getIssuer())) {
          fLog(LOG_L3,"CHECK_ATTR_CERT", "ISSUER_CHECK=FAIL");
        } else if (cert->containsCondition() && !cert->getCondition().isValid()) {
          fLog(LOG_L3,"CHECK_ATTR_CERT", "COND_CHECK_VALIDITY=FAIL");
        } else {
          holder = certCache.addCertificate(certs[i], 0, now, endCacheTime);
          sLog(LOG_L3,"CHECK_ATTR_CERT");
        }
      }
    }

    if (!holder) {
      fLog(LOG_L3, "TRY_ATTR_DIRS");
    } else {
      sLog(LOG_L3, "TRY_ATTR_DIRS");
    }

    return holder;
  }

private:

  AKRuntime* akRuntime;

  vector<AttributeInfo> attributeInfos;

  vector<AttributeInfo> newAttributeInfos;
  
}; // end of class Evaluator

}; // end of namespace

typedef PolicyAuthorizer PA;

PA::PolicyAuthorizer(AKRuntime* akRuntime)
 : akRuntime(akRuntime) {
}

PA::~PolicyAuthorizer() {
}

const AkentiMessage& 
PA::getAkentiMessage() const {
  return message;
}

namespace Akenti {

// to = from ^ filter (INTERSECTION)
void
filter1(const vector<string>& from, const set<string>& filter, vector<string>& to) {
  for (int j = 0; j < from.size(); ++j) {
    set<string>::iterator pos;

    pos = find(filter.begin(), filter.end(), from[j]);

    if (pos != filter.end()) {
      to.push_back(from[j]);
    }
  }
}

// to = from - filter  (DIFFERENCE)
bool
filter2(const vector<string>& from, const set<string>& filter, vector<string>& to) {
  for (int j = 0; j < from.size(); ++j) {
    set<string>::iterator pos;

    pos = find(filter.begin(), filter.end(), from[j]);

    if (pos == filter.end()) {
      to.push_back(from[j]);
    }
  } 

  return to != from;
}

}

int
PA::authorize(Evaluator& evaluator, 
              const vector<UseCondCertificate>& ucCerts) {
  int count = 0;

  for (int i = 0; i < ucCerts.size(); ++i) {
    bLog(LOG_MON, "SATISFY_USE_COND", ucCerts[i]->toLogString());

    const Condition& condition = ucCerts[i]->getCondition();
    bool critical = ucCerts[i]->isCritical();
    const set<string>& desiredActions = akRuntime->getDesiredActions(); 
    vector<string> actions;
    Expression expr;

    if (desiredActions.size()) {
      // THROW AWAY ANY ACTION THAT IS NOT IN THE DESIRED SET 
      filter1(ucCerts[i]->getActions(), desiredActions, actions);
      
      vector<string> temp;

      // THROW AWAY ANY ACTION THAT WAS GRANTED OR WOULD BE GRANTED IF A 
      // CRITICAL CONDITION EVALUATES TO TRUE
      filter2(actions, minimalActions, temp);
      actions = temp;
    } else { 
      // THROW AWAY ANY ACTION THAT WAS GRANTED OR WOULD BE GRANTED IF A 
      // CRITICAL CONDITION EVALUATES TO TRUE
      filter2(ucCerts[i]->getActions(), minimalActions, actions);
    }

    // WE NOW ONLY BOTHER EVALUATING IF WE HAVE TO (CRITICAL) OR IF WE CAN GAIN
    // NEW ACTIONS. I have my efficiency hat on. :)

    if (!critical && !actions.size()) { 
      sLog(
        LOG_MON, "SATISFY_USE_COND", 
        "ACTION_INFO= actions already in the bag"
      );
    } else if (!condition.parseConstraint(expr)) {
      fLog(LOG_MON, "SATISFY_USE_COND", "PARSE_EXPRESSION=FAIL");
    } else {
      evaluator.setAttributeInfos(ucCerts[i]->getAttributeInfos());

      Expression temp;
      int ret = expr.reduce(evaluator, temp);

      if (ret == 0) {
        fLog(LOG_MON, "SATISFY_USE_COND", "EVALUATION=0");
        
        if (critical) {
          // STOP CRITICAL UC NOT SATIFIED
          return count;
        }
      } else if (ret == 1) {
        for (int i = 0; i < actions.size(); ++i) {
          unconditionalActions.insert(actions[i]);
          minimalActions.insert(actions[i]);
          iLog(LOG_MON, "INFO", "ACTION=" + actions[i]);
        } 

        ++count;
        sLog(LOG_MON, "SATISFY_USE_COND");
      } else {
        Condition condition(temp.toString(), evaluator.getAttributeInfos());
        ConditionalActions condActions(condition, actions, critical);

        if (critical) {
          for (int i = 0; i < actions.size(); ++i) {
            minimalActions.insert(actions[i]);
          }
        }

        condActionsList.push_back(condActions);
        ++count;
        sLog(
          LOG_MON, "SATISFY_USE_COND", 
          temp.toLogString() + " must be checked further"
        );
      }
    }
  }

  return count;
}

bool
PA::authorize() {
  Evaluator evaluator(akRuntime);
  int size = akRuntime->numOfPolicies();

  for (int i = 0; i < size; ++i) {
    akRuntime->setLevel(i + 1);

    const Policy& policy = akRuntime->getPolicy(i);
    vector<UseCondCertificate> ucCerts = policy->getCriticalUCCerts();
    int size = ucCerts.size();

    if (size) {
      bLog(LOG_MON,"SATISFY_CRITICAL_POLICY", policy->toLogString());

      if (authorize(evaluator, ucCerts) != size) {
        message = AkentiMessage(
          301, " for " + policy->getResource()
        );

        fLog(LOG_MON,"SATISFY_CRITICAL_POLICY", "MUST_SATISFY_UC=FAIL");
        return false;
      } else { 
        sLog(LOG_MON,"SATISFY_CRITICAL_POLICY");
      }
    }
  }
   
  const set<string>& desiredActions = akRuntime->getDesiredActions(); 
  bool stop = desiredActions.size() && (desiredActions == unconditionalActions);

  for (int i = 0; !stop && i < size; ++i) {
    akRuntime->setLevel(i + 1);

    const Policy& policy = akRuntime->getPolicy(i);
    vector<UseCondCertificate> ucCerts = policy->getFlexibleUCCerts();

    if (ucCerts.size()) {
      bLog(LOG_MON,"SATISFY_NON_CRITICAL_POLICY", policy->toLogString());

      if (!authorize(evaluator, ucCerts)) {
        fLog(LOG_MON,"SATISFY_NON_CRITICAL_POLICY", "NB_OF_UC_SATISFIED=0");
      } else {
        sLog(LOG_MON,"SATISFY_NON_CRITICAL_POLICY");
      }

      stop = desiredActions.size() && (desiredActions == unconditionalActions);
    }
  }

  if (!unconditionalActions.size() && !condActionsList.size()) {
    message = AkentiMessage(302);
    return false;
  } else if (!condActionsList.size() 
              && 
              desiredActions.size() && desiredActions != unconditionalActions) {
    set<string>::iterator pos = unconditionalActions.begin();

    for (; pos != unconditionalActions.end(); ++pos) {
      akRuntime->addUnconditionalAction(*pos);
    }

    // ALLOW THE UPPER LEVEL TO MAKE THE FINAL DECISION
    message = AkentiMessage(106);
    return true;
  } else if (!condActionsList.size()) {
    set<string>::iterator pos = unconditionalActions.begin();

    for (; pos != unconditionalActions.end(); ++pos) {
      akRuntime->addUnconditionalAction(*pos);
    }

    message = AkentiMessage(104);
    return true;
  }

  set<string>::iterator pos = unconditionalActions.begin();

  for (; pos != unconditionalActions.end(); ++pos) {
    akRuntime->addUnconditionalAction(*pos);
  }

  minimalActions = unconditionalActions;

  int i;

  // COMPACT A LITTLE MORE
  for (i = 0; i < condActionsList.size() && condActionsList[i].isCritical(); i++) {
    vector<string> actions;

    if (filter2(condActionsList[i].getActions(), minimalActions, actions)) {
      ConditionalActions newCondActions(condActionsList[i].getCondition(), actions, true);

      akRuntime->addConditionalActions(newCondActions);
    } else {
      akRuntime->addConditionalActions(condActionsList[i]);
    }
    
    for (int j = 0; j < actions.size(); ++j) {
      minimalActions.insert(actions[j]);
    }
  }

  if (desiredActions.size() && desiredActions == minimalActions) {
    if (i != condActionsList.size()) {
      throw Error("FIRST SANITY CHECK IN POLICY AUTHORIZER FAILED");
    }
  }

  // THESE ARE ALREADY AS COMPACT AS THEY CAN BE
  for (; i < condActionsList.size(); i++) {
    const vector<string>& actions = condActionsList[i].getActions();

    if (!actions.size()) {
      throw Error("SECOND SANITY CHECK IN POLICY AUTHORIZER FAILED");
    }

    for (int j = 0; j < actions.size(); ++j) {
      minimalActions.insert(actions[j]);
    }

    akRuntime->addConditionalActions(condActionsList[i]);
  }

  if (desiredActions.size() && desiredActions != minimalActions) {
    message = AkentiMessage(107);
    return true;
  }

  message = AkentiMessage(105);
  return true;
}

