#include <sstream>
#include "OutputStream.hpp"
#include "InputStream.hpp"
#include "CacheRequest.hpp"
#include "UtcTime.hpp"
#include "StringUtil.hpp"
#include "Exception.hpp"
#include "StringTokenizer.hpp"
#include "Error.hpp"

typedef CacheRequest CR;

using std::ostringstream;
using std::istringstream;

CR::CacheRequest() 
  : method("NO_METHOD"), 
    rootResourceName("NO_ROOT_RESOURCE"),
    certHash("NO_CERT_HASH"),
    dirName("") {
}

CR::CacheRequest(const string& method) 
  : method(method), 
    rootResourceName("NO_ROOT_RESOURCE"),
    certHash("NO_CERT_HASH"),
    dirName("") {
  if (!method.length()) {
    throw Error("CacheRequest:constructor:illegal arguemt");
  }
}

CR::~CacheRequest() {
}

const string& 
CR::getMethod() const {
  return method;
}

bool 
CR::hasMethod(const string& method) const {
  return StringUtil::equalsIgnoreCase(this->method, method);
}

void 
CR::setRootResourceName(const string& rootResourceName) {
  if (!rootResourceName.length()) {
    throw Error("CacheRequest:setRootResourceName:illegal arguemt");
  }

  this->rootResourceName = rootResourceName;
}

const string& 
CR::getRootResourceName() const {
  return rootResourceName;
}
  
void 
CR::setCertHash(const string& certHash) {
  if (!certHash.length()) {
    throw Error("CacheRequest:setCertHash:illegal arguemt");
  }

  this->certHash = certHash;
}

const string& 
CR::getCertHash() const {
  return certHash;
}

void 
CR::setDirName(const string& dirName) {
  this->dirName = dirName;
}

const string& 
CR::getDirName() const {
  return dirName;
}

void 
CR::setCertificate(const CacheCertificate& cert) {
  if (!cert) {
    throw Error("CacheRequest:setCertificate:illegal arg");
  }

  this->cert = cert;
}
  
const CacheCertificate& 
CR::getCertificate() const {
  return cert;
}

const UtcTime& 
CR::getNotAfter() const {
  return cert->getNotAfter();
}

void 
CR::writeHeaderInfo(OutputStream& os) const {
  os << "METHOD: " << method << '\n';
}
  
void 
CR::writeResourceInfo(OutputStream& os) const {
  os << "ROOT_RESOURCE_NAME: " << rootResourceName << '\n';
  os << "CERT_HASH: " << certHash << '\n';

  if (dirName.length()) {
    os << "DIRNAME: " << dirName << '\n';
  }
}

void 
CR::writeCertInfo(OutputStream& os) const {
  if (cert) {
    ostringstream stream;

    stream << cert;

    string temp = stream.str(); 

    os << "CERT_EXP_TIME: " << cert->getNotAfter().toString() << '\n' 
       << "CERT_LENGTH: " << temp.length() - 1 << '\n'
       << temp;
  }
}

void 
CR::writeExternal(OutputStream& os) const {
  writeHeaderInfo(os);
  writeResourceInfo(os);
  writeCertInfo(os);
  os << '\n';
}

void 
CR::readHeaderInfo(InputStream& is) {
  method = parse("METHOD", readLine(is));
}
  
void 
CR::readResourceInfo(InputStream& is) {
  rootResourceName = parse("ROOT_RESOURCE_NAME", readLine(is));
  certHash = parse("CERT_HASH", readLine(is));
  
  string line = readLine(is);
  
  try {
    dirName = parse("DIRNAME", line);
  } catch(Exception& ex) {
    dirName = "";
    
    for (int i = 0; i <= line.length(); ++i) {
      is->unget();
    }
  }
}

void 
CR::readCertInfo(InputStream& is) {
  UtcTime notAfter(parse("CERT_EXP_TIME", readLine(is)));
  istringstream stream1(parse("CERT_LENGTH", readLine(is)));
  
  int len;

  if (!(stream1 >> len)) {
    throw Exception("CacheRequest:readCertInfo:failed to read cert length");
  }
  
  string temp(++len, '\0');
 

  if (!is->read((char*) temp.c_str(), len)) {
    throw Exception("CacheRequest:readCertInfo:failed to read cert length");
  }

  istringstream stream2(temp);
  
  AkentiCertificate akCert = AkentiCertificate::initFromStream("PEM", stream2);

  if (!akCert.instanceOf<CacheCertificate>()) {
     throw Exception("CacheRequest:readCertInfo:expected a cache cert");
  }

  cert = (CacheCertificate&) akCert;
}

void 
CR::readExternal(InputStream& is) {
  readHeaderInfo(is);
  readResourceInfo(is);
  cert = CacheCertificate();
  string line = readLine(is);
  
  if (line.length()) {
    for (int i = 0; i <= line.length(); ++i) {
      is->unget();
    }

    readCertInfo(is);
    
    if (readLine(is).length()) {
      throw Exception("CacheRequest:readExternal:failed to read eline");
    }
  }
}

string
CR::parse(const string& keyword, const string& line) const {
  StringTokenizer tokenizer(line, " =:");

  if (!tokenizer.hasMoreTokens()) {
    throw Exception("CacheRequest:parse:(1)" + line);
  }

  if (!StringUtil::equalsIgnoreCase(keyword, tokenizer.nextToken())) {
    throw Exception("CacheRequest:parse:(2)" + line);
  }

  if (!tokenizer.hasMoreTokens()) {
    throw Exception("CacheRequest:parse:(3)" + line);
  }

  return tokenizer.nextToken();
}

string 
CR::readLine(InputStream& is) const {
  string line = is.getline();

  if (!is->good()) {
    throw Exception("CacheRequest:readLine:error reading a line");
  }

  return line;
}


string 
CR::toString() const {
  ostringstream* delegate = new ostringstream();
  OutputStream os(delegate);

  writeExternal(os);
  return delegate->str();
}

string 
CR::paramString() const {
  return toString();
}
