#include "CacheCertificate.hpp"
#include "CacheCertGenerator.hpp"
#include "CertificateCache.hpp"
#include "Properties.hpp"
#include "UtcTime.hpp"
#include "DistinguishedName.hpp"
#include "PublicKey.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "ExternalCache.hpp"
#include "ServerContext.hpp"

using namespace Akenti;

typedef CertificateCache CC;

CC::Entry::Entry() : viewChanged(false) {
}

CC::Entry::~Entry() { 
}

void
CC::Entry::setViewChanged(bool viewChanged) {
  this->viewChanged = viewChanged;
}

bool
CC::Entry::getViewChanged() const {
  return viewChanged;
}

/*
 * We only call this method when we collect fresh certs from
 * the external cache.
 **/
void 
CC::Entry::setAKCertHolders(const vector<AKCertHolder>& certHolders) {
  if (this->certHolders.size() || viewChanged) {
    throw Error("Entry:setAKCertHolders:should never happen");
  }

  UtcTime now;

  for (int i = 0; i < certHolders.size(); ++i) {
    if (certHolders[i]->checkValidity(now)) {
      certHolders[i]->addCertificateStatusListener(this);
      this->certHolders.push_back(certHolders[i]);
    }
  }

  // I hope the comments helped :) 
  // The devil is in the details. 
  viewChanged = this->certHolders.size() != certHolders.size(); 
}

int 
CC::Entry::containsCertificate(const AkentiCertificate& cert) const {
  int i = certHolders.size() - 1;

  for (; i != -1 && certHolders[i]->getCertificate() != cert; --i) {
  }
      
  return i;
}

int 
CC::Entry::size() const {
  return certHolders.size();
}

AKCertHolder& 
CC::Entry::operator[](int i) {
  return certHolders[i];
}

AKCertHolder& 
CC::Entry::addAKCertHolder(const AKCertHolder& holder) {
  certHolders.push_back(holder);
  holder->addCertificateStatusListener(this);
  viewChanged = true;
  return certHolders.back(); 
}

void
CC::Entry::certStartCacheTimeChanged(const AkentiCertificate& cert) {
  viewChanged = true;
}

void
CC::Entry::certEndCacheTimeChanged(const AkentiCertificate& cert) {
  viewChanged = true;
}

void
CC::Entry::verifyingCertOK(const AkentiCertificate& cert) {
  viewChanged = true;
}

void
CC::Entry::verifyingCertFailed(const AkentiCertificate& cert) {
  int index = containsCertificate(cert);

  certHolders[index]->removeCertificateStatusListener(this);
  certHolders.erase(certHolders.begin() + index);
  viewChanged = true;
}

CC::CertificateCache() : useExtCache(false) {
  if (ExternalCache::isExternalCacheSet()) {
    useExtCache = ExternalCache::getExternalCache()->openConnection();
  }
}

CC::~CertificateCache() {
  if (useExtCache) {
    ExternalCache::getExternalCache()->closeConnection();
  }
}

void
CC::setDirectoryName(const string& rootResourceName) {
  this->rootResourceName = rootResourceName;
}

AKCertHolder&
CC::addCertificate(const AkentiCertificate& cert,
                   int level,
                   const UtcTime& startCacheTime,
                   const UtcTime& endCacheTime) {
  Entry& entry = table[cert->hashCode() + cert->getType()];
  int index = entry.containsCertificate(cert);

  if (index != -1) {
    entry[index]->setStartCacheTime(startCacheTime);
    entry[index]->setEndCacheTime(endCacheTime);
    entry[index]->setLevel(level);
    return entry[index];
  } else {
    return entry.addAKCertHolder(
      new AKCertHolderImpl(
        cert,
        level,
        startCacheTime,
        endCacheTime
      )
    );
  }
}

AKCertHolder&
CC::addAKCertHolder(const AKCertHolder& holder) {
  AkentiCertificate cert = holder->getCertificate();
  Entry& entry = table[cert->hashCode() + cert->getType()];
  int index = entry.containsCertificate(cert);

  if (index != -1) {
    entry[index]->setStartCacheTime(holder->getStartCacheTime());
    entry[index]->setEndCacheTime(holder->getEndCacheTime());
    entry[index]->setLevel(holder->getLevel());
    return entry[index];
  } else {
    return entry.addAKCertHolder(holder);
  }
}

/**
  * Remember that if the entry is empty and the the viewChanged flag is 
  * set to true, then we must have fetched the certificates from the external 
  * cache and they must have been removed at a later point. So what do we do?
  * Here is what we don't do. We don't go to the external cache. 
  * Are u smiling? I am. Peace.
  */
vector<AKCertHolder>
CC::getAKCertHolders(const string& key, const string& type) {
  Entry& entry = table[key + type];
  vector<AKCertHolder> certHolders;

  if (!entry.size() && !entry.getViewChanged() && useExtCache) {
    try {
      ExternalCache& extCache = ExternalCache::getExternalCache();
      CacheCertificate cert = extCache->retrieveCacheCertificate(
          rootResourceName, key, ""
      );

      if (cert && cert->checkValidity()) {
        Properties& props = Properties::getSystemProperties();

        if (props.getValue("SignCerts", "off") == "on") {
          ServerContext& context = ServerContext::getServerContext();
          const Certificate& serverCert = context.getCertificate();
          
          if (!cert->verify(serverCert->getSubjectPublicKey())) {
            // XXXX LOG THIS 
            // cout << "CertificateCache:cache cert verified FAILED" << endl;
            return certHolders;
          }
        } 

        entry.setAKCertHolders(cert->getAKCertHolders());
      }
    } catch(Exception& ex) {
      // XXX LOG THIS EXCEPTION
    }
  } 

  for (int i = 0; i < entry.size(); ++i) {
    certHolders.push_back(entry[i]);
  }
  
  return certHolders;
}

vector<AKCertHolder>
CC::getAKCertHolders(const string& key,
                     const string& type,
                     const UtcTime& notBefore) {
  vector<AKCertHolder> certHolders = getAKCertHolders(key, type);
  vector<AKCertHolder> retHolders;

  for (int i = 0; i < certHolders.size(); ++i) {
    if (certHolders[i]->getStartCacheTime() >= notBefore) {
      retHolders.push_back(certHolders[i]);
    }
  }

  return retHolders;
}

bool
CC::updateExternalCache() {
  if (!useExtCache) {
    return false;
  }

  vector<CacheCertificate> cacheCertificates;
  vector<Entry*> updatedEntries;
  map<string, Entry>::iterator pos; 
  UtcTime now;
  
  for (pos = table.begin(); pos != table.end(); ++pos) {
    Entry& entry = pos->second;

    if (entry.getViewChanged()) {
      vector<AKCertHolder> holders;

      for (int i = 0; i < entry.size(); ++i) {
        if (entry[i]->checkValidity(now)) {
          holders.push_back(entry[i]);
        }
      }

      // Should we call flush if the view changed and . 
      // Next version maybe. :)
    
      if (holders.size()) {
        cacheCertificates.push_back(
          generateCacheCertificate(holders)
        ); 
          
        updatedEntries.push_back(&entry);
      }
    }
  }

  int num = 0;
  ExternalCache& extCache = ExternalCache::getExternalCache();

  for (int i = 0; i < cacheCertificates.size(); ++i) {
    if (extCache->storeCacheCertificate(rootResourceName, cacheCertificates[i], "")) {
      updatedEntries[i]->setViewChanged(false);
      num++;
    }
  }

  return num != 0;
}

CacheCertificate
CC::generateCacheCertificate(const vector<AKCertHolder>& holders) {
  CacheCertGenerator gen;
  
  gen.setVersionNumber(1);
  gen.setUniqueID("UID");
  gen.setAKCertHolders(holders);

  UtcTime now;

  gen.setNotBefore(now);

  UtcTime max(holders[0]->getEndCacheTime());

  for (int i = 1; i < holders.size(); ++i) {
    UtcTime temp = holders[i]->getEndCacheTime();
    
    if (max < temp) {
      max = temp;
    }
  }

  gen.setNotAfter(max);

  Properties& props = Properties::getSystemProperties();

  if (props.getValue("SignCerts", "off") == "on") {
    // cout << "CERTIFICATE CACHE: SIGNING CERTIFICATE..." << endl;

    ServerContext& context = ServerContext::getServerContext();
    const Certificate& cert = context.getCertificate();

    AkentiIssuer issuer(
      AkentiPrincipal(
        cert->getPrincipal(), cert->getGuarantor()
      )
    );
    
    gen.setIssuer(issuer);

    const PrivateKey& key = context.getPrivateKey();
    string algorithm = key->getAlgorithm();
    
    if (algorithm == "RSA") {
      gen.sign(key, "RSA-MD5");
    } else if (algorithm == "DSA") {
      gen.sign(key, "DSA-SHA1");
    } else {
      // cout << "CERTIFICATE CACHE: SIGNING CERTIFICATE FAILED" << endl;
      throw Error(
       "CertificateCache:generateCacheCertificate:unknown algorithm:" + algorithm 
      ); 
    }

    // cout << "CERTIFICATE CACHE: SIGNED CERTIFICATE..." << endl;
  }

  return gen.generate();
}

