#include <sstream>
#include "CheckUCCertCollector.hpp"
#include "Policy.hpp"
#include "AKRuntime.hpp"
#include "CertificateRetrievalFactory.hpp"
#include "CertificateCache.hpp"
#include "UseCondIssuerGroup.hpp"
#include "CertificateVerifier.hpp"
#include "UtcTime.hpp"
#include "Policy.hpp"
#include "URL.hpp"
#include "Util.hpp"
#include "Logger.hpp"

#include <iostream> // SRS - for debugging only
using namespace std;

using std::ostringstream;

using namespace Akenti;

typedef CertificateRetrievalFactory CRF;

typedef CheckUCCertCollector UCCC;

UCCC::CheckUCCertCollector(AKRuntime* akRuntime) : akRuntime(akRuntime) {
}

UCCC::~CheckUCCertCollector() {
}
  
const AkentiMessage&
UCCC::getAkentiMessage() const {
  return message;
}

vector <AkentiCertificate>
UCCC::getUCCertHolders(Policy& policy, vector <string>& dirVector) {
  string hash = policy->hashCode();

  bLog(LOG_MON,"TRY_CACHED_UC_CERTS");

  vector <AkentiCertificate> ucCertVec;

  CertInfoCollector &cinfoCol = akRuntime->getCertInfoCollector();

  map<string, string> args;
  int numOfGroups = policy->numOfUseCondIssuerGroups();
  ostringstream os;

  args["type"] = "UseCondition";
  args["hash"] = Util::createFileName(hash);
  os << numOfGroups;

  bLog(
    LOG_MON, 
    "TRY_UC_DIRS", 
    "FILE_HASH_NAME=" + args["hash"] + " NB_OF_UC_GROUPS=" + os.str()
  );

  vector<AkentiCertificate> ucCerts, trustedUCCerts;
  vector<string> dirVec1, dirVec2;
  int groupIndex = 0;
  bool foundValidUC;
  UtcTime now;
  
  CertificateVerifier& certVerifier = akRuntime->getCertificateVerifier();

  do {
    const UseCondIssuerGroup& group = policy->getUseCondIssuerGroup(groupIndex);
    const vector<Directory>& dirs = group.getDirectories();

    foundValidUC = false;

    for (int j = 0; !foundValidUC && j < dirs.size(); ++j) {
      bLog(LOG_MON, "FETCH_UC_CERTS", "DIR=" + dirs[j].toString());

      vector<AkentiCertificate> certs = CRF::getAkentiCertificates(
        dirs[j], args
      );

      int size = certs.size();
      ostringstream os;

      os << size;
          
      if (!size) {
        fLog(LOG_MON, "FETCH_UC_CERTS");
      } else {
        sLog(LOG_MON, "FETCH_UC_CERTS", "NB_OF_UC_CERTS_FOUND=" + os.str());
      }

      int k;

      // We must find trusted uc cert for each group
      for (k = 0; !foundValidUC && k < size; ++k) {
        UseCondCertificate cert = (UseCondCertificate&) certs[k];
        const AkentiIssuer& issuer = cert->getIssuer();

        bLog( LOG_L3, "CHECK_UC_CERT", cert->toLogString());
        fLog(LOG_L3, "CHECK_UC_CERT", "HASH_CHECK=failed");

        /* first check for fundamental error - maybe policy or resource
         * doesnt exist, in which case level wouldnt have been set
         * How did it get to this cert then ??? XXX 
         */
        if (akRuntime->getLevel() <= 0) {
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                     (AkentiMessage (501)),
                     cert->hashCode(), "Use Condition", cert->getResource(),
                     true); 
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (501));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (cert->hashCode() != hash) { 
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                     (AkentiMessage (502, " Hash check failed")),
                     cert->hashCode(), "Use Condition", cert->getResource(),
                     true); 
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Hash check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!cert->checkValidity(now)) {
          fLog(LOG_L3, "CHECK_UC_CERT", "TIME_CHECK=failed");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Validation failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(),
                  true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Validation failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!group.containsUCIssuer(issuer.getIssuer())) {
          fLog(LOG_L3, "CHECK_UC_CERT", "GROUP_CHECK=failed");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Group check failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(),
                  true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Group check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!cert->getCondition().isValid()) {
          fLog(LOG_L3,"CHECK_UC_CERT", "COND_CHECK_VALIDITY=FAIL");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Conditions check failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(),
                   true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Conditions check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!certVerifier.verify(cert)) {
          fLog(LOG_L3, "CHECK_UC_CERT", "TRUST_CHECK=FAIL");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Trust check failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(), true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Trust check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else {

          foundValidUC = true;
          trustedUCCerts.push_back(cert);
          dirVec2.push_back (dirs[j].toString());
          sLog(LOG_L3, "CHECK_UC_CERT");
        }
      }

      // We will check the trust later 
      for (int l = k; l < certs.size(); ++l) {
        UseCondCertificate cert = (UseCondCertificate&) certs[l];
        CertInfo cinfo; 

        const AkentiIssuer& issuer = cert->getIssuer();

        bLog (LOG_L3, "CHECK_UC_CERT", cert->toLogString());

        if (cert->hashCode() != hash) {
          fLog(LOG_L3, "CHECK_UC_CERT", "HASH_CHECK=failed");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Hash check failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(), true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Hash check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!cert->checkValidity(now)) {
          fLog(LOG_L3, "CHECK_UC_CERT", "TIME_CHECK=failed");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Validation failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(), true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Validation failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!group.containsUCIssuer(issuer.getIssuer())) {
          fLog(LOG_L3, "CHECK_UC_CERT", "GROUP_CHECK=failed");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Group check failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(), true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Group check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else if (!cert->getCondition().isValid()) {
          fLog(LOG_L3,"CHECK_UC_CERT", "COND_CHECK_VALIDITY=FAIL");
           int index = cinfoCol.lookup (cert->hashCode(), "Use Condition",
                     cert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, cert, 
                  (AkentiMessage (502, " Conditions check failed")),
                  cert->hashCode(), "Use Condition", cert->getResource(), true);
               cinfo.setLocation (dirs[j].toString());
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (502, " Conditions check failed"));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
        } else {
          ucCerts.push_back(cert);
          dirVec1.push_back (dirs[j].toString());
          sLog(LOG_L3, "CHECK_UC_CERT");
        }
      }
    }
  } while (foundValidUC && ++groupIndex < numOfGroups);

  if (foundValidUC) {
    int level = akRuntime->getLevel();
    for (int i = 0; i < trustedUCCerts.size(); ++i) {
      ucCertVec.push_back(trustedUCCerts[i]);
      dirVector.push_back (dirVec2[i]);
    }
    for (int i = 0; i < ucCerts.size(); ++i) {
      ucCertVec.push_back(ucCerts[i]);
      dirVector.push_back (dirVec1[i]);
    }
  }

  if (ucCertVec.size()) {
    sLog(LOG_MON,"TRY_UC_DIRS");
  } else {
    ostringstream os;

    os << groupIndex;
    fLog(LOG_MON, "TRY_UC_DIRS", "NO_VALID_UC_FOR_GROUP=" + os.str());
  }

  return ucCertVec;
}

bool
UCCC::collectUseCondCertificates(Policy& policy) {
  bLog(LOG_MON,"COLLECT_UC_CERTS", "RESOURCE=" + policy->getResource());
  CertInfoCollector &cinfoCol = akRuntime->getCertInfoCollector();
  vector <string> dirVector;


  vector <AkentiCertificate> ucCertVec = getUCCertHolders (policy, dirVector);
  int size = ucCertVec.size();

  if (!size) {
    fLog(
      LOG_MON,"COLLECT_UC_CERTS", 
      "NOT_FOUND_UC=did not find any UCCerts"
    );

    message = AkentiMessage(
      501, 
      " for " + policy->getResource()
    );

    cinfoCol.setStatus (false);
  }

  ostringstream os1;
  
  os1 << size;

  iLog(LOG_MON, "INFO", "NB_OF_UC_CERTS_FOUND=" + os1.str());

  CertificateVerifier& certVerifier = akRuntime->getCertificateVerifier();
  int count = 0;

  for (int i = 0; i < ucCertVec.size(); ++i) {
    UseCondCertificate ucCert = (UseCondCertificate&) ucCertVec[i];

    if (!ucCert->isScopeLocal() || policy->isLeaf()) {
      if (certVerifier.verify(ucCertVec[i])) {
           policy->addUCCertificate(ucCert);
           int index = cinfoCol.lookup (ucCert->hashCode(), "Use Condition",
                     ucCert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (true, ucCert, 
                  (AkentiMessage (103, " for " + policy->getResource())),
                  ucCert->hashCode(), "Use Condition", ucCert->getResource());
               cinfo.setLocation (dirVector[i]);
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage(index, AkentiMessage(103, " for " + policy->getResource()));
           }
        ++count;
      }
      else {// certificate could not be verified
           int index = cinfoCol.lookup (ucCert->hashCode(), "Use Condition",
                     ucCert.get()->getUniqueID());
           if (index < 0) { // doesnt exist, add it
               CertInfo cinfo (false, ucCert, (AkentiMessage (503)),
                   ucCert->hashCode(), "Use Condition", ucCert->getResource(),
                   true);
               cinfo.setLocation (dirVector[i]);
               cinfoCol.addCertInfo (cinfo);
           } 
           else {
               cinfoCol.addMessage (index, AkentiMessage (503));
               cinfoCol.setUsed (index, false);
               cinfoCol.setError (index, true);
           }
           cinfoCol.setStatus (false);
      }
    }
    else {
       // comes here if verified but not used - like if there was
       // an error before this point.
       int index = cinfoCol.lookup (ucCert->hashCode(), "Use Condition",
                     ucCert.get()->getUniqueID());
       if (index < 0) { // doesnt exist, add it
          CertInfo cinfo (false, ucCert,
               (AkentiMessage (103, " for " + policy->getResource())),
               ucCert->hashCode(), "Use Condition", ucCert->getResource());
          cinfoCol.addCertInfo (cinfo);
          }
           else {
               cinfoCol.addMessage (index, AkentiMessage (103, " for " + policy->getResource()));
               cinfoCol.setUsed (index, false);
           }
    }

  }

  ostringstream os2;
  
  os2 << count;

  iLog(LOG_MON,"INFO", "NB_OF_ACCEPTABLE_UC_CERTS=" + os2.str());
  sLog(LOG_MON,"COLLECT_UC_CERTS");

  message = AkentiMessage(
    103, 
    " for " + policy->getResource()
  );

  return true; // Necessary for Test case - otherwise doesnt mean anything
}

bool
UCCC::collectUseCondCertificates() {
  bool ret = true;
  CertInfoCollector &cinfoCol = akRuntime->getCertInfoCollector();

  int size = cinfoCol.getCertInfoVec().size();

  for (int i = 0; i < size; ++i) {
      CertInfo &cinfo = cinfoCol.getCertInfo(i);
      if (cinfo.getType() == "Policy") {
         AkentiCertificate akcert = cinfo.getCertificate();
         PolicyCertificate pcert = (PolicyCertificate &)akcert;
         Policy policy  = new PolicyImpl(cinfo.isLeaf(), pcert);
         collectUseCondCertificates (policy);
      }
  }
  return true; // Necessary for Test case - otherwise doesnt mean anything
}
