#include <utility>
#include <memory>
#include <sstream>
#include <map>
#include <set>
#include "PolicyChecker.hpp"
#include "AKRuntime.hpp"
#include "CertificateVerifier.hpp"
#include "AkentiCertificate.hpp"
#include "CheckPolCertCollector.hpp"
#include "CheckUCCertCollector.hpp"
#include "Certificate.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "Logger.hpp"
#include "DistinguishedName.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "PolicyContext.hpp"
#include "MessageDigest.hpp"
#include "StringTokenizer.hpp"
#include "Util.hpp"
#include "CertInfoCollector.hpp"

#include <iostream> // srs for debugging
using namespace std; 

using namespace Akenti;

typedef PolicyChecker PC;

Properties PC::resourceMap = Properties();

PC::PolicyChecker() : akRuntime(AKRuntime::createAKRuntime()) {
}

PC::~PolicyChecker() {
  delete(akRuntime);
}

bool
PC::checkPolicy(const string& resource, 
                CertInfoCollector &ciCol,
                AkentiMessage& mesg) {
  string uid = "NO_UID";

  std::auto_ptr<Logger> logger(initLogger("PolicyChecker"));

  bLog(LOG_MON,"CHECK_POLICY","RESOURCE=" + resource);
  
  string dresource = resourceMap.getValue(resource, resource);

  if (dresource != dresource) {
    iLog(LOG_MON,"RESOURCE_MAPPED_TO","RESOURCE=" + dresource);
  }

  akRuntime->setResource(dresource);
  akRuntime->setResourceAltName(resource);

  
  CertificateCache& cache = akRuntime->getCertificateCache();
  StringTokenizer st(resource, "/ ", false);

  cache.setDirectoryName(st.nextToken());

  if (!collectAndVerifyPolicies(akRuntime, mesg)) { 
    fLog(LOG_MON,"CHECK_POLICY");
  } else if (!collectAndVerifyUseConds(akRuntime, mesg)) {
    fLog(LOG_MON,"CHECK_POLICY");
  } else if (!verifyCAS(akRuntime, mesg)) {
    fLog(LOG_MON,"CHECK_POLICY");
  } 

  sLog(LOG_MON,"CHECK_POLICY");
 
  ciCol = akRuntime->getCertInfoCollector();

  // Call this to make sure Policies and associated use cond vectors
  // are clubbed together

  ciCol.reorderCertInfoVec ();

  return true;
}

bool 
PC::collectAndVerifyPolicies(AKRuntime* akRuntime, AkentiMessage& mesg) const {
  bLog(LOG_MON,"POLICIES_COLLECTION_PHASE");

  CheckPolCertCollector policyCollector(akRuntime);

  if (!policyCollector.collectPolicies()) {
    fLog(LOG_MON,"POLICIES_COLLECTION_PHASE","COLLECTING_POLICIES=FAIL");
    mesg = policyCollector.getAkentiMessage();
  } else if (!policyCollector.verifyPolicies()) {
    fLog(LOG_MON,"POLICIES_COLLECTION_PHASE","CHECKING_POLICIES=FAIL");
    mesg = policyCollector.getAkentiMessage();
  }

  sLog(LOG_MON,"POLICIES_COLLECTION_PHASE");
  return true;
}

bool
PC::collectAndVerifyUseConds(AKRuntime* akRuntime, AkentiMessage& mesg) const {
  bLog(LOG_MON,"UC_COLLECTION_PHASE");

  CheckUCCertCollector ucCollector(akRuntime);

  if (!ucCollector.collectUseCondCertificates()) {
    fLog(LOG_MON,"UC_COLLECTION_PHASE");
    mesg = ucCollector.getAkentiMessage();
  }

  sLog(LOG_MON,"UC_COLLECTION_PHASE");
  return true;
}

namespace Akenti {

class Compare {

public:

  bool operator()(const AkentiPrincipal& ap1, const AkentiPrincipal& ap2) {
    return Util::compare(ap1.getName(), ap2.getName())  
           && 
           Util::compare(ap2.getGuarantor(), ap2.getGuarantor());
  }

};

}

bool
PC::verifyCAS(AKRuntime* akRuntime, AkentiMessage& mesg) const {
  bLog(LOG_MON,"CAS_VERIFICATION_PHASE");
  const PolicyContext& ctx = akRuntime->getPolicyContext();
  const vector<Policy>& policies = ctx->getPolicies();
  CertificateVerifier& verifier = akRuntime->getCertificateVerifier();
  bool ok = true;

  // SKIP ROOT SINCE IT HAS ALREADY BEEN CHECKED
  for (int i = 1; i < policies.size(); ++i) {
    akRuntime->setLevel(i + 1);

    int count = 0;
    const vector<CertificateAuthorityInfo>& 
       infos = policies[i]->getCertificateAuthorityInfos();
    
    for (int j = 0; j < infos.size(); j++) {
      const vector<Certificate>& caChain = infos[i].getChain();
      AkentiX509Certificate akCert(caChain[caChain.size() - 1]);

      bool ret = verifier.verify(akCert);
      
      if (!ret) {
        iLog(LOG_MON,"INVALID_CA_WARNING");
        ok = false;
      } else {
        const AkentiIssuer& ai = akCert->getIssuer();
        const AkentiPrincipal& ap = ai.getIssuer();
        iLog(LOG_MON,"VALID_CA", ap.toLogString());
      }

      count += int(ret); 
    }
     
    if (!count && infos.size()) {
      fLog(LOG_MON,"CAS_VERIFICATION_PHASE");
      mesg = AkentiMessage(460);
      return false;
    }
  }

  if (!ok) {
    mesg = AkentiMessage(151);
  } else {
    mesg = AkentiMessage(150);
  }

  sLog(LOG_MON,"CAS_VERIFICATION_PHASE");
  return true;
}

bool
PC::verifyUCIssuers(AKRuntime* akRuntime, AkentiMessage& mesg) const {
#if 0
  bLog(LOG_MON,"UC_ISSUERS_VERIFICATION_PHASE");
  const PolicyContext& ctx = akRuntime->getPolicyContext();
  const vector<Policy>& policies = ctx->getPolicies();
  CertificateVerifier& verifier = akRuntime->getCertificateVerifier();

  for (int i = 0; i < policies.size(); ++i) {
    akRuntime->setLevel(i + 1);

    set<AkentiPrincipal, Compare> aps;
    const AkentiIssuer& ai = policy[i]->getIssuer();
    const AkentiPrincipal& ap = ai.getIssuer();
    aps.insert(ap);  // ap is the issuer of the policy and thus has already been cheked
    
    int j = policies[i].numOfUseCondIssuerGroups() - 1;

    do {
      const UseCondIssuerGroup& ucGroup = policies[i].getUseCondIssuerGroup(j);
      const vector<AkentiPrincipal>& ucIssuers = ucGroup.getUCIssuers();
      int count = 0;
      
      for (int k = 0; k < ucIssuers.size(); ++k) {
        if (aps.find(ucIssuers[k]) == aps.end()) {
          if (verifier.verify(ucIssuers[k])) {
            count++;
            iLog(LOG_MON,"VALID_UC_ISSUER", ucIssuers[k].toLogString());
            aps.insert(ucIssuers[k]);
          } else {
            iLog(LOG_MON,"INVALID_UC_ISSUER_WARNING", ucIssuers[k].toLogString());
          }
        } else {
          count++;
        }
      }
      
      if (!count) {
        ostringstream os;

        //os << "ALL UC ISSUERS IN GROUP " << j << " ARE UNACCEPTABLE");
        //mesg = AkentiMessage(470, os.str());
        mesg = AkentiMessage (470, ": " + j);
        fLog(LOG_MON,"UC_ISSUERS_VERIFICATION_PHASE");
        return false;
      }
    } while (--j >= 0);
  }

  sLog(LOG_MON,"UC_ISSUERS_VERIFICATION_PHASE");
  return true;
#endif
}

Logger*
PC::initLogger(const string& id) const {
  Logger* logger = NULL;

  if (!Logger::hasBeenInitialized()) {
    string message;

    logger = Logger::init(id, message);
  } 
  
  return logger;
}

void 
PC::loadResourceMap(const string& fileName) {
  resourceMap.load(fileName);
}

