#ifndef AK_SOCKET_IMPL_HPP
#define AK_SOCKET_IMPL_HPP

#include "InputStream.hpp"
#include "OutputStream.hpp"
#include "IOHandler.hpp"

/**
  * A <i>SocketImpl</i> object provides the implementation for both server and
  * client socket functions such as set/get the host, set/get the port,
  * do the connect, bind, listen and accept for the socket. Contains the
  * host, port, inputstream, outputsream and IOhandler associated with the 
  * socket. The IOHandler in turn contains a file descriptor which refers
  * to either a plain Unix socket or a SSLsocket structure.
  * <p>
  * Normally a application will reference these methods thru either a Socket or
  * SSLSocket if it is the client-side of the connection or thru either a
  * ServerSocket or SSLServerSocket if it is the server-side of the connection.
  * Thus client sockets only call doConnect, and server sockets only call
  * doBind, doListen and doAccept.
  *
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class SocketImpl {

public:

  /**
    * Constructs this object. 
    */
  SocketImpl();
  
  /** 
    * Destroys this object.
    */
  virtual ~SocketImpl();

  /**
    * Sets the host name to host.
    * 
    * @param host the name of the host
    */
  void setHost(const string& host);

  /**
    * Returns the name of the host.
    * 
    * @return the name of the host
    */
  const string& getHost() const;

  /**
    * Sets the port number to port. 
    * 
    * @param port the port number 
    * @exception Error - if port is less than zero
    */
  void setPort(int port);
  
  /**
    * Returns the port number.
    * 
    * @return the port number
    */
  int getPort() const;

  /**
    * Provides the SocketImpl with either a plain Unix socket or a secure
    * socket that has been connected to the host and port associated with 
    * the SocketImpl. The IOHandler for the connected fd is
    * kept in the SocketImpl. Called by the client-side of a connection.
    *
    * @exception IOException if can't open stream socket
    * @exception IOException if can't connect to host
    * @exception IOException if can't connect to port
    * @see IOHandlerImpl()
    * @see openconnection()
    */
  virtual void doConnect();

  /*
    * Provides the SocketImpl with either a plain Unix socket or a secure
    * socket that has been bound to its port. The socket fd in kept in  
    * the SocketImpl. Called by the server-side of a connection.
    * 
    * @exception IOException if can't open stream socket
    * @exception IOException if setsockop failed
    * @see IOHandlerImpl#IOHandlerImpl(int)
    * @see bindToAddress()
    */
  virtual void doBind();

  /**
    * Does a listen call on the encapsulated socket file descriptor.
    * Done by the server-side of the connection, which is now ready to
    * accept a connect message from a client.
    *  
    * @param backlog the size of the request queue 
    * @exception Error the socketImpl's fd is not bound to a port.
    * @exception IOException listen call failed
    */
  virtual void doListen(int backlog);
  
  /**
    * Does an accept call on the encapsulated socket IOHandler which 
    * references a socket that a listen has been done on and
    * if successful, replaces the SocketImpl IOHandler with one referencing
    * the new socket that is connected to a client.
    *
    * @param impl the SocketImpl
    * @exception Error  if accept gets an invalid argument
    * @exception IOException accept call failed.
    * @see acceptConnection()
    */
  virtual void doAccept(SocketImpl* impl);

  /**
    * Closes the socket and sets the input and output streams to NULL.
    * 
    * @exception Error invalid operation if the socket has already been closed
    * @exception IOException error from close.
    */
  virtual void doClose();

  /**
    * If the socketImpl already has an input stream, returns it, otherwise 
    * it creates a new one and returns it.
    *
    * @exception Error if invalid operation if the SocketImpl fd is not set
    * @exception Error if bad alloc 
    * @exception IOException if bad input stream
    * @return the input stream
    * @see doGetInputStream()
    */
  InputStream& getInputStream();

  /**
    * If the socketImpl already has an output stream, returns it, otherwise 
    * it creates a new one and returns it.
    * 
    * @exception Error invalid operation if the SocketImpl fd is not set
    * exception Error if bad alloc
    * @exception IOException if bad output stream
    * @return the  output stream
    * @see doGetOutputStream()
    */
  OutputStream& getOutputStream();
  
  /**
    * Returns a new SocketImpl. The default host is null and the default port
    * is -1.
    *
    * @return a new SocketImpl
    */
  virtual SocketImpl* newInstance();

  /**
    * Returns a string representation for SocketImpl which is used for   
    * debugging purposes.
    *
    * @return debugging info
    */
  virtual string paramString() const;

protected:

  /**
    * Provides the SocketImpl with a new Unix socket which has
    * been connected to the host and the port associated with
    * this SocketImpl.
    *
    * @return the connected file descriptor
    * @exception IOException if can't open stream socket
    * @exception IOException if can't connect to host
    * @exception IOException if can't connect to port
    * @see IOHandlerImpl()
    */
  int openConnection();

  /**
    * Provides the SocketImpl with a new Unix socket which has been
    * bound to the port associated with the SocketImpl.
    *
    * @exception IOException if cannot open stream socket 
    * @exception IOException if setsockop failed 
    * @exception IOException if cannot bind local address
    * @return  the fd for the newly created socket
    */
  int bindToAddress();
  
  /**
    * Drills down in the SocketImpl to find a Unix socket file descriptor
    * on which it is assumed that listen call has been done, 
    * does a Unix accept call on it
    * and returns a new Unix socket file descriptor. This socket will 
    * be connected to a client who has made a connect call to the 
    * port that this server is listening on. 
    *
    * @exception Error if no impl is input.
    * @exception IOException if the new connection failed
    * @return the sd for the newly created socket
    */
  int acceptConnection(SocketImpl* impl);

  /**
    * Creates a new SocketInputStream initialized from a new SocketInBuffer
    * attached to the IOHandler associated with the socketImpl.
    * 
    * @return  a new istream connected to the SocketImpl's IOHandler.
    */
  virtual istream* doGetInputStream();

  /**
    * Creates a new SocketOutPutStream initialized from a new SocketOutBuffer
    * attached to the IOHandler associated with the socketImpl.
    * 
    * @return a new ostream connected to the SocketImpl's IOHandler.
    */
  virtual ostream* doGetOutputStream();

  string host;

  int port;

  InputStream is;

  OutputStream os;

  IOHandler fd;

};

} // end of namespace Akenti

#endif

