#ifndef AK_URL_HANDLER_FACTORY_IMPL_HPP
#define AK_URL_HANDLER_FACTORY_IMPL_HPP

#include <string>
#include <map>
#include "URLHandler.hpp"
#include "Error.hpp"

using std::string;
using std::map;

/**
  * Contains a map of URL protocols and their URLHandlers. Provides
  * methods to add a new protocol and handler and to return the handler
  * for a given protocol.
  * 
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class URLHandlerFactoryImpl {

public:
  
  /**
    * Constructor
    */
  URLHandlerFactoryImpl() {
  }

  /**
    * Destructor
    */
  virtual ~URLHandlerFactoryImpl() {
  }

  /**
    * Adds a new URLHandler/protocol entry
    *
    * @param protocol  name of protocol, e.g. http, file or ldap
    * @param handler A URLHandler which wraps a URLHandlerImpl for the prtocol
    * @exception Error if either the handler or protocol arguments are null
    */
  void 
  addURLHandler(const string& protocol, const URLHandler& handler) {
    if (!handler || !protocol.length()) {
      throw Error("URLHandlerFactoryImpl:addURLHandler:illegal argument");
    }

    handlers[protocol] = handler;
  }

  /**
    * Returns the URLHandler associated with a protocol
    *
    * @param name the name of the protocol, e.g. http, ldap or file
    * @return a URLHandler that wraps a URLHandlerImpl that provides the
    *   correct methods for this protocol.
    */  
  URLHandler& 
  getURLHandler(const string& name) {
    return handlers[name];
  }

private:

  map<string, URLHandler> handlers;

};

} // end namespace Akenti

#endif

