#ifndef AK_URL_HANDLER_IMPL_HPP
#define AK_URL_HANDLER_IMPL_HPP

#include <string>

using std::string;

namespace Akenti {

class URL;
class URLConnection;

/**
  * A <i>URLHandlerImpl<i> is base class for URL protocol-specific handlers. 
  * It provides methods such parsing the URL string representation,
  * initializing the URL from a given string, returning the external
  * format for the URL, and returning the  default port for its protocol.
  * In general this base class implments methods for the http protocol, which
  * thus need to be overridden by the subclasses for file and ldap.
  * URLHandlerImpls are created by the URLHandlerFactoryImpl which defines
  * one URLHandlerImpl for each of the supported protocols: http, file or ldap.
  *
  * @see URL
  * @see URLConnection
  *
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class URLHandlerImpl {

public:

  /**
    * Destroys this object.
    */
  virtual ~URLHandlerImpl();

  /**
    * Parses the string spec to set the fields in the URL.
    * 
    * @param url the URL
    * @param spec string to be parsed     
    * @exception MalformedURLException if the next token is not correct or
    * @exception MalformedURLException if reading port failed
    * @see URL#URL(const string& url)
    * @see StringTokenizer#StringTokenizer(const string& source, ... )
    */
  virtual void 
  parseURL(URL& url, const string& spec) const;

  /**
    * Initializes the URL with the host, port and name of file.  
    *
    * @param url URL to be initialized  
    * @param host name of the host
    * @param port port number 
    * @param file the name of the file 
    * @see setHost(URL& url, const string& host)  
    * @see setPort(URL &url, int port)
    * @see setFile(URL& url, const string& file)  
    */
  virtual void 
  initURL(URL& url, 
          const string& host, 
          int port, 
          const string& file) const;
  
  /**
    * Formats the URL to an external string.
    *
    * @param url the URL
    * @return string the external string format 
    */
  virtual string 
  toExternalForm(const URL& url) const;

  /**
    * Returns the URL connection given the URL. Pure virtual method,
    * must be supplied by protocol-specific handlers.
    *
    * @param url the URL
    * @return the URL connection
    */
  virtual URLConnection 
  getURLConnection(const URL& url) const = 0;

protected:

  /**
    * Constructs this object.
    */
  URLHandlerImpl();

  /**
    * Sets the host in the URL.
    *
    * @param url the URL
    * @param host the name of the host   
    * @exception MalformedURLException if there is no host name
    * @see setHost(URL& url, const string& host) 
    */
  void setHost(URL& url, const string& host) const;

  /**
    * Sets the port in the URL.
    *
    * @param url the URL
    * @param port the port number 
    * @exception MalformedURLException if port is less than zero 
    */
  void setPort(URL &url, int port) const;

  /**
    * Sets the file in the URL.
    *
    * @param url the URL
    * @param file string the name of the file
    * @exception MalformedURLException if port is less than zero 
    */
  void setFile(URL& url, const string& file) const;

};

} // end of namespace Akenti

#endif
