#ifndef OPEN_SSL_BASE_64_IMPL_HPP
#define OPEN_SSL_BASE_64_IMPL_HPP
#include <openssl/x509.h>
#include "SecurityException.hpp"
#include "Base64Impl.hpp"

namespace Akenti {

/**
  * This class is the subclass of <i>Base64Impl</i> which implements 
  * the functionality of Base64 data encoding and decoding. This class 
  * contains a data buffer and an openssl EVP-encode-context structure.
  * The methods are implemented by calls to the openssl library. The
  * use model for this class is to intialize the object for either encoding
  * or decoding which empties the buffer and sets  the context. Data is then
  * appended to the buffer by one or more update calls. Finally a doEncode or
  * doDecode method is called which en/decodes the accumlated data and returns
  * the en/decoded string. 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */ 

class OpenSSLBase64Impl : public Base64Impl {

public:

  /**
    * Constructs this object.
    * 
    * @see Base64Impl#Base64Impl() 
    */
  OpenSSLBase64Impl(); 

  /**
    * Destroys this object 
    */
  ~OpenSSLBase64Impl(); 
 
protected:
  
  /**
    * Initializes the <i>buffer</i> for encoding.
    */ 
  void doEncodeInit();

  /**
    * Updates engine for encoding from the input data string starting
    *  at the offset and taking the specified number of bytes
    *
    * @param data string used to update the engine
    * @param offset int begin index in data
    * @param len int number of bytes used to update the engine
    */
  void doEncodeUpdate(const string& data, int offset, int len);
 
   /**
     * Encodes all the data  accumlated in the buffer and returns
     * the encoded string. It does not reset the data buffer.
     *
     * @return a Base64 enoded string of all the accumlated data.
     */
  string doEncode();

  /**
    * Initializes the engine for decoding. 
    */
  void doDecodeInit();

  /**
    * Updates engine for decoding from the input data string starting
    * at the offset and taking the specified number of bytes
    *
    * Comment: will not work if the encoded string 'enc' is broken up and this 
    *          function is called several times to decode the string. 
    * 
    * @param data string used to update the engine
    * @param offset int begin index 
    * @param len int number of bytes used to update the engine
    */
  void doDecodeUpdate(const string& enc, int offset, int len);

   /**
     * Completes the decode procedure, returns 
     * decoded data in string.
     *
     * @return The Base64 decoded data as a string
     * @exception SecurityException if decoding failed
     */
  string doDecode();

private:

  EVP_ENCODE_CTX ctx;

  string buffer; 

};

} // end namespace Akenti

#endif

