#include <openssl/asn1.h>
#include "OpenSSLCRLImpl.hpp"
#include "DistinguishedName.hpp"
#include "UtcTime.hpp"
#include "Error.hpp"
#include "SSLConstants.hpp"
#include "OpenSSLX509RevokedCertificateImpl.hpp"

using Akenti::OpenSSLCRLImpl;

OpenSSLCRLImpl::OpenSSLCRLImpl(X509_CRL *x509crl) : crl(x509crl) {
  if (x509crl == NULL) {
    throw Error("OpenSSLCRLImpl:OpenSSLCRLImpl:illegal arg");
  }
}

OpenSSLCRLImpl::~OpenSSLCRLImpl() {
  X509_CRL_free(crl);
}

long
OpenSSLCRLImpl::getVersion() const {
  return X509_CRL_get_version(crl);
}

Akenti::DistinguishedName 
OpenSSLCRLImpl::getGuarantor() const {
  string dn(256, '\0');

  X509_NAME_oneline(X509_CRL_get_issuer(crl), (char *) dn.data(), dn.length());
  return DistinguishedName(dn);
}

string 
OpenSSLCRLImpl::getSignature() const {
  return string((char *) crl->signature->data, crl->signature->length);
}

Akenti::UtcTime 
OpenSSLCRLImpl::getLastUpdate() const {
  ASN1_UTCTIME *utctime = X509_CRL_get_lastUpdate(crl);

  return UtcTime((char *) utctime->data);
}

Akenti::UtcTime 
OpenSSLCRLImpl::getNextUpdate() const {
  ASN1_UTCTIME *utctime = X509_CRL_get_nextUpdate(crl);

  return UtcTime((char *) utctime->data);
}

vector<Akenti::RevokedCertificate> 
OpenSSLCRLImpl::getRevokedCertificates() const {
  vector<RevokedCertificate> revokedCertificates;

  for (int i=0; i < sk_X509_REVOKED_num(crl->crl->revoked); i++) {
    X509_REVOKED *revoked = sk_X509_REVOKED_value(crl->crl->revoked, 
						  i);
    revokedCertificates.push_back(
          RevokedCertificate (
	    new OpenSSLX509RevokedCertificateImpl (
	    ASN1_INTEGER_get(revoked->serialNumber),
	    UtcTime((char *) (revoked->revocationDate->data))
    )));
  }

  return revokedCertificates;
}

bool
OpenSSLCRLImpl::isRevoked(long serialnumber) const {
  vector<RevokedCertificate> revokedCerts = getRevokedCertificates();

  for (int i=0; i < revokedCerts.size(); i++) {
    if (serialnumber == revokedCerts[i]->getCertificateNumber()) {
      return true;
    }
  }

  return false;
}

int
OpenSSLCRLImpl::isRevoked(long serialnumber, const UtcTime& ctime) const {
  if ((getLastUpdate() < ctime) &&
       ctime < getNextUpdate()) {
    vector<RevokedCertificate> revokedCerts = 
      getRevokedCertificates();

    for (int i=0; i < revokedCerts.size(); i++) {
      if (serialnumber == revokedCerts[i]->getCertificateNumber()) {
        return SSLConstants::REVOKED;
      }
    }

    return SSLConstants::NOT_REVOKED;
  }

  return SSLConstants::UNKNOWN;
}
