#ifndef OPEN_SSL_CRL_IMPL_HPP
#define OPEN_SSL_CRL_IMPL_HPP

#include <openssl/x509.h>
#include "CertificateRevocationList.hpp"

namespace Akenti {

/**
  * An <i>OpenSSLCRLImpl<i> object is a subclass of
  * <i>CertificateRevocationListImpl</i>. It provides applications,
  * albeit indirectly, with the functionality of getting the version,
  * guarantor, signature, last update, next update of the CRL and the revoked
  * certificates. It contains a X509_CRL structure implements its methods
  * by wrapping OpenSSL library calls.
  *
  * @see CertificateRevocationListImpl
  * @see UtcTime 
  * @see RevokedCertificate 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  **/

class OpenSSLCRLImpl : public CertificateRevocationListImpl {

public:

  /**
    * Constructs an <i>OpenSSLCRLImpl</i> given a <i>X509_CRL</i>
    * object.
    *
    * @param crl X509_CRL* pointer to the X509 Certificate Revocation List (CRL)
    */
  OpenSSLCRLImpl(X509_CRL *crl);

  /**
    * Deletes this object
    */
  ~OpenSSLCRLImpl();

  /**
    * Returns the version of this CRL 
    * 
    * @return long the version number
    */
  long getVersion() const;

  /**
    * Returns the guarantor(issuer) of this CRL 
    * 
    * @return DistinguishedName the issuer of this crl 
    *
    */
  DistinguishedName getGuarantor() const;

  /**
    * Returns the signature of this CRL 
    * 
    * @return string the signature
    */
  string getSignature() const;

  /**
    * Returns the last update of this CRL
    * 
    * @return UtcTime the ime for last update
    */
  UtcTime getLastUpdate() const;

  /**
    * Returns the next update for this CRL
    * 
    * @return UtcTime the time for the next update  
    */
  UtcTime getNextUpdate() const;

  /**
    * Returns a <i>vector</i> which contains the
    * certificates have been revoked.
    *
    */
  vector<RevokedCertificate> getRevokedCertificates() const; 

 /**
   * Returns a <i>boolean</i> which specifies
   * the revocation status of the given serial number
   * of the certificate.
   *
   * @param long serialnumber that specifies the serial number 
   * 	         of the certificate to be checked for revocation
   * @return bool which indicates the revocation status of
   *	     the given serial number of a certificate
   */
  bool isRevoked(long serialnumber) const;

 /**
   * Returns a <i>boolean</i> which specifies
   * the revocation status of the given serial number
   * of the certificate.
   *
   * @param long serialnumber that specifies the serial number 
   * 	         of the certificate to be checked for revocation
   * @param UtcTime ctime that specifies the time at which the
   *	 	    revocation status is requested.
   * @return int which indicates the revocation status of
   *	     the given serial number of a certificate. In addition
   *	     to true or false, this also includes the UNKNOWN status.
   */
  int isRevoked(long serialnumber, const UtcTime& ctime) const;

private:

  X509_CRL *crl;

};

} // end namespace Akenti

#endif
