#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/asn1.h>
#include "OpenSSLCertificateFactoryImpl.hpp"
#include "Certificate.hpp"
#include "CertificateChain.hpp"
#include "CertificateRequest.hpp"
#include "CertificateGenerator.hpp"
#include "CertificateRequestGenerator.hpp"
#include "SecurityException.hpp"
#include "IOException.hpp"
#include "Error.hpp"
#include "OpenSSLCertificateGeneratorImpl.hpp"
#include "OpenSSLCertificateRequestGeneratorImpl.hpp"
#include "OpenSSLCertificateRequestImpl.hpp"
#include "OpenSSLX509CertificateImpl.hpp"
#include "OpenSSLX509CertificateChainImpl.hpp"
#include "CertificateRevocationList.hpp"
#include "OpenSSLCRLImpl.hpp"

typedef Akenti::OpenSSLCertificateFactoryImpl OSCFI;

OSCFI::OpenSSLCertificateFactoryImpl() {
}

OSCFI::~OpenSSLCertificateFactoryImpl() {
}

vector<Akenti::Certificate>
OSCFI::initCertificatesFromFile(const string& type, const string& fileName) const {
  if (type == "PEM") {
    BIO* bio = BIO_new_file((char *) fileName.c_str(), "r");

    if (!bio) {
      throw IOException(
        "OpenSSLCertificateFactoryImpl:initCertificatesFromFile:can't open file " + fileName
      );
    }

    vector<Certificate> certs;
    X509* delegate;

    while ((delegate = PEM_read_bio_X509(bio, NULL, NULL, NULL))) {
      Certificate cert = new OpenSSLX509CertificateImpl(delegate);
      
      certs.push_back(cert);
    }

    BIO_free(bio);
    return certs;
  } else {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificatesFromFile:unsupported type:" + type
    );
  }
}

vector<Akenti::Certificate>
OSCFI::initCertificates(const string& type, const string& encoding) const {
  if (type == "PEM") {
    BIO* bio = BIO_new_mem_buf((char*) encoding.data(), encoding.length());
    
    if (!bio) {
      throw Error(
        "OpenSSLCertificateFactoryImpl:initCertificates:out of memory 1"
      );
    }

    vector<Certificate> certs;
    X509* delegate;

    while ((delegate = PEM_read_bio_X509(bio, NULL, NULL, NULL))) {
      Certificate cert = new OpenSSLX509CertificateImpl(delegate);

      certs.push_back(cert);
    }

    BIO_free(bio);
    return certs;
  } else {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificates:unsupported type:" + type
    );
  }
}

Akenti::CertificateChain
OSCFI::initCertificateChainFromFile(
  const string& type, const string& fname) const {

  if (type == "PEM") {

    BIO* bio = BIO_new_file((char *) fname.c_str(), "r");

    if (!bio) {
      throw IOException(
        "OpenSSLCertificateFactoryImpl:initCertificatesFromFile:can't open file " + fname
      );
    }


    STACK_OF(X509) *chain = sk_X509_new_null();

    if (chain == NULL) {
      throw Error(
        "OpenSSLCertificateFactoryImpl:initCertificateChainFromFile:stack memory error"
      );
    }

    X509 *x;

    while ( (x=PEM_read_bio_X509(bio, NULL, NULL, NULL)) != NULL) {
      sk_X509_push(chain, x);
    }

    // check for proper certificates in the chain
    if (sk_X509_num(chain) > 1) {
      bool done = false;
      int index = 0;

      // take the first one in the chain which is always the subject
      // cert . (assumption)
      while (!done) {
        bool match = false;
        for (int i=1, n=sk_X509_num(chain); i < n; i++) {
          if (X509_NAME_cmp(
		X509_get_issuer_name(sk_X509_value(chain,0)), 
                X509_get_subject_name(sk_X509_value(chain, i))) == 0) 
          {
            match = true;
            break;
          }
        }

        if (!match) {
          throw SecurityException(
            "OpenSSLCertificateFactoryImpl:initCertificateChainFromFile:invalid chain");
        }

        index++;

        if (index == sk_X509_num(chain)) {
          done = true;
        }
      }
    }

    BIO_free(bio);

    return CertificateChain(new OpenSSLX509CertificateChainImpl(chain));

  } else {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificateChainFromFile:unsupported type:" + type
    );
  }
}

Akenti::CertificateChain
OSCFI::initCertificateChain(const string& type, const string& encoding) const {
  if (type == "PEM") {
    BIO* bio = BIO_new_mem_buf((char*) encoding.data(), encoding.length());
    
    if (!bio) {
      throw Error(
        "OpenSSLCertificateFactoryImpl:initCertificateChain:out of memory 2"
      );
    }

    STACK_OF(X509) *chain = sk_X509_new_null();

    if (chain == NULL) {
      throw Error(
        "OpenSSLCertificateFactoryImpl:initCertificateChain:stack memory error"
      );
    }

    X509 *x;

    while ( (x=PEM_read_bio_X509(bio, NULL, NULL, NULL)) != NULL) {
      sk_X509_push(chain, x);
    }

    // check for proper certificates in the chain
    if (sk_X509_num(chain) > 1) {
      bool done = false;
      int index = 0;

      // take the first one in the chain which is always the subject
      // cert . (assumption)
      while (!done) {
        bool match = false;
        for (int i=1, n=sk_X509_num(chain); i < n; i++) {
          if (X509_NAME_cmp(
		X509_get_issuer_name(sk_X509_value(chain,0)), 
                X509_get_subject_name(sk_X509_value(chain, i))) == 0) 
          {
            match = true;
            break;
          }
        }

        if (!match) {
          throw SecurityException(
            "OpenSSLCertificateFactoryImpl:initCertificateChain:invalid chain");
        }

        index++;

        if (index == sk_X509_num(chain)) {
          done = true;
        }
      }
    }

    BIO_free(bio);

    return CertificateChain(new OpenSSLX509CertificateChainImpl(chain));

  } else {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificateChain:unsupported type:" + type
    );
  }
}

Akenti::Certificate
OSCFI::initCertificateFromFile(const string& type, const string& fileName) const {
  BIO* bio = BIO_new_file((char *) fileName.c_str(), "r");
   
  if (!bio) {
    throw IOException(
      "OpenSSLCertificateFactoryImpl:initCertificateFromFile:can't open file " + fileName
    );
  }

  X509* delegate = NULL;

  if (type == "DER") {
    delegate = d2i_X509_bio(bio, NULL);
  } else if (type == "PEM") {
    delegate = PEM_read_bio_X509(bio, NULL, NULL, NULL);
  } else {
    BIO_free(bio);

    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificateFromFile:unsupported type:" + type
    );
  }

  BIO_free(bio);

  if (!delegate) {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificateFromFile:bad encoding" 
    );
  }

  return new OpenSSLX509CertificateImpl(delegate);
}

Akenti::Certificate 
OSCFI::initCertificate(const string& type, const string& encoding) const {
  X509* delegate = NULL;

  if (type == "DER") {
    unsigned char* temp = (unsigned char*) encoding.data();
    long len = encoding.length();

    delegate = d2i_X509(NULL, &temp, len);
  } else if (type == "PEM") {
    BIO* bio = BIO_new_mem_buf((char*) encoding.data(), encoding.length());

    if (!bio) {
      throw Error(
        "OpenSSLCertificateFactoryImpl:initCertificate:out of memory 3"
      );
    }

    delegate = PEM_read_bio_X509(bio, NULL, NULL, NULL);
    BIO_free(bio);
  } else {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificate:unsupported type:" + type
    );
  }

  if (!delegate) {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:initCertificate:bad encoding"
    );
  }

  return new OpenSSLX509CertificateImpl(delegate);
}

string
OSCFI::getCertificateEncoded(const string& type, const Certificate& cert) const {
  OpenSSLX509CertificateImpl*
    impl = dynamic_cast<OpenSSLX509CertificateImpl*> (cert.get());

  if (impl == NULL) {
    throw Error("OpenSSLCertificateFactoryImpl:getCertificateEncoded:bad cert");
  }
  
  return impl->getEncoded(type);
}

Akenti::CertificateGenerator 
OSCFI::getCertificateGenerator(const string& type) const {
  return CertificateGenerator(new OpenSSLCertificateGeneratorImpl);
}


Akenti::CertificateRequest
OSCFI::initCertificateRequest(const string& type, const string& encoding) const {
  return CertificateRequest(OpenSSLCertificateRequestImpl::init(type, encoding));
}

string
OSCFI::getCertificateRequestEncoded(const string& type, const CertificateRequest& request) const {
  OpenSSLCertificateRequestImpl*
    impl = dynamic_cast<OpenSSLCertificateRequestImpl*> (request.get());
  
  if (!impl) {
    throw Error("OpenSSLCertificateFactoryImpl:getCertificateRequestEncoded:bad request");
  }

  return impl->getEncoded(type);
}

Akenti::CertificateRequestGenerator
OSCFI::getCertificateRequestGenerator(const string& type) const {
  return CertificateRequestGenerator(new OpenSSLCertificateRequestGeneratorImpl);
}

Akenti::CertificateRevocationList
OSCFI::initCRL(const string& crl, const string& type) const {
  if (type == "PEM") {
    BIO *mbio = BIO_new(BIO_s_mem());
    BIO_puts(mbio, crl.c_str());
    X509_CRL *x509crl = PEM_read_bio_X509_CRL(mbio, NULL, NULL, NULL);
    BIO_free(mbio);
    return CertificateRevocationList(new OpenSSLCRLImpl(x509crl));
  } else {
    throw SecurityException("OpenSSLCertificateFactoryImpl::initCRL:Unsupported type " + type);
  }
}

Akenti::CertificateRevocationList
OSCFI::initCRLFromFile(const string& crlfile, const string& type) const {
  if (type == "PEM") {
    BIO *fbio = BIO_new(BIO_s_file());
    BIO_read_filename(fbio, crlfile.c_str());
    X509_CRL *x509crl = PEM_read_bio_X509_CRL(fbio, NULL, NULL, NULL);
    BIO_free(fbio);
    return CertificateRevocationList(new OpenSSLCRLImpl(x509crl));
  } else {
    throw SecurityException("OpenSSLCertificateFactoryImpl::initCRL:Unsupported type " + type);
  }
}
