#ifndef OPEN_SSL_CERTIFICATE_FACTORY_IMPL_HPP
#define OPEN_SSL_CERTIFICATE_FACTORY_IMPL_HPP

#include "CertificateFactoryImpl.hpp"

namespace Akenti {

/**
  * A subclass of CertficateFactoryImpl that is used to create 
  * OpenSSLCertificates or OpenSSLCertificateReqests from encoded 
  * representations of these objects, to create Generators
  * for OpenSSLCertificates or OpenSSLCertificateRequests, and
  * to encode OpenSSLCertficates or OpenSSLCertificateRequests 
  * for exporting. 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class OpenSSLCertificateFactoryImpl : public CertificateFactoryImpl {

public:

  /**
    * Constructs this object.
    */
  OpenSSLCertificateFactoryImpl();

  /**
    * Deletes this object.
    */
  ~OpenSSLCertificateFactoryImpl();

  vector<Certificate>
  initCertificatesFromFile(const string& type, const string& fileName) const;

  vector<Certificate>
  initCertificates(const string& type, const string& encoding) const;

  CertificateChain
  initCertificateChainFromFile(const string& type, const string& fname) const;

  CertificateChain
  initCertificateChain(const string& type, const string& encoding) const;

  Certificate
  initCertificateFromFile(const string& type, const string& fileName) const;

  /**
    * Returns a <i>Certificate</i> containing a 
    * <i>OpenSSLX509CertificateImpl</i> which has been initialized from the
    * encoded string. The <i>type</i> specifies the type of encoding 
    * the string is in, e.g.  DER or PEM.
    *
    * @param type the Certificate's encoding type
    * @param encoding a string of bytes
    * @return  a Certificate
    * @exception SecurityException if either type is unsupported
    *  or if decoding fails
    * @see CertificateFactory
    * @see CertificateFactoryImpl
    */
  Certificate 
  initCertificate(const string& type, const string& encoding) const;

  /**
    * Encodes the <i>Certificate</i> using an encoding  specified by 
    * <i>type</i>. It assumes that the Certificate holds a 
    * OpenSSLX509CertificateImpl and the type is one supported by the
    * X509 libraries (e.g. PEM or DER)
    *
    * @param type string specifies the type to be used
    * @param certificate Certificate to be encoded
    * @return encoding a string of bytes
    * @exception Error if certificate is unitialized or unrecognized
    * @exception SecurityException if type is unsupported or if the
    *  encoding fails
    * @see Certificate
    * @see CertificateImpl
    */
  string 
  getCertificateEncoded(const string& type, const Certificate& certificate) const;

  /**
    * Return an instance of <i>CertificateGenerator</i> which holds an
    * <i>OpenSSLCertificateGeneratorImpl</i>.
    *
    * @param type  not used currently
    *   but here to match the CertificateFactoryImpl interface.
    * @see CertificateGenerator
    * @see CertificateGeneratorImpl
    */
  CertificateGenerator 
  getCertificateGenerator(const string& type) const;

  /**
    * Returns a <i>CertificateRequest</i> which holds an 
    * <i>OpenSSLCertificateRequestImpl<i>.
    * The <i>type</i> specifies the type of encoding used by the string 
    * representation of the CertificateRequest, e.g. DER or PEM.
    *
    * @param type the CertificateRequest's encoding type
    * @param encoding a string of bytes
    * @return  a CertificateRequest
    * @exception SecurityException if type  is unsupported
    *  or if decoding fails
    * @see CertificateRequest
    * @see CertificateRequestImpl
    */
  CertificateRequest 
  initCertificateRequest(const string& type, const string& encoding) const;

  /**
    * Encodes the <i>CertificateRequst</i> which is assumed to contain
    * an OpenSSLCertificateRequestImpl, using an encoding type specified 
    * by <i>type</i>.
    *
    * @param type string specifies the type of encoding to be used
    * @param request CertificateRequest to be encoded
    * @return encoding a string of bytes
    * @exception Error if request is unitialized or unrecognized
    * @exception SecurityException if type is unsupported or if the
    *  encoding fails
    * @see CertificateRequest
    * @see CertificateRequestImpl
    */
  string 
  getCertificateRequestEncoded(const string& type, const CertificateRequest& request) const;

  /**
    * Return an instance of <i>CertificateRequestGenerator</i> wrapping an
    * OpenSSLCertificateGeneratorImpl.<i>type</i> is included as a parameter
    * to match the CertificateFactoryImpl method, but is not used here.
    *
    * @param type  not used.
    * @see CertificateRequestGenerator
    * @see CertificateRequestGeneratorImpl
    */
  CertificateRequestGenerator 
  getCertificateRequestGenerator(const string& type) const;

  /**
    * Return an instance of <i>CertificateRevocationList</i> wrapping
    * an OpenSSLCRLImpl. <i> type </i> is included as a parameter
    * to indicate which type of encoding has been used to create
    * the crl supplied
    *
    * @param crl string that refers to the CRL
    * @param type string that refers to the type of CRL encoding
    * @see CertificateRevocationList
    */
  CertificateRevocationList
  initCRL(const string& crl, const string& type) const;

  /**
    * Return an instance of <i>CertificateRevocationList</i> wrapping
    * an OpenSSLCRLImpl. <i> type </i> is included as a parameter
    * to indicate which type of encoding has been used to create
    * the crlfile that is supplied
    *
    * @param crl string that refers to the CRL filename
    * @param type string that refers to the type of CRL encoding
    * @see CertificateRevocationList
    */
  CertificateRevocationList
  initCRLFromFile(const string& crlfile, const string& type) const;

};

} // end namespace Akenti

#endif
