#ifndef OPEN_SSL_KEY_FACTORY_IMPL_HPP
#define OPEN_SSL_KEY_FACTORY_IMPL_HPP 

#include "KeyFactoryImpl.hpp"

namespace Akenti {

class PublicKey;
class PrivateKey;
class KeyPair;

/**
  * This class is a subclass of <i>KeyFactoryImpl</i>
  * which has methods for  key generation, encoding,
  * and decoding.
  *
  * a pair of RSA or DSA keyPairs will be generated by calling
  * generateRSAKeyPair()or generateDSAKeyPair().
  *
  * a privateKey or publicKey can be encoded by calling
  * getKeyEncoded() with the format specified as "DER".
  *
  * a privateKey or publicKey class can be initialized from a DER string
  * by calling initPrivateKey() or initPublicKey(). Both RSA and DSA keys
  * are currently handled.
  *
  * @see PublicKey
  * @see PrivateKey
  * @see KeyPair 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */    
class OpenSSLKeyFactoryImpl : public KeyFactoryImpl {

public:

  /**
    * Construct this object
    */
  OpenSSLKeyFactoryImpl();

  /**
    * Delete this object
    */
  ~OpenSSLKeyFactoryImpl();
 
  /**
    * Encodes the <i>key</i> using a format specified by <i>format</i>.
    *
    * @param format string specifies the format to be used. Only DER for now.
    * @param key PublicKey to be encoded
    *
    * @return string the encoding using the format specified
    *
    * @exception SecurityException if format is unsupported or if the
    *                              encoding fails
    * @exception Error if key is unitialized or unrecognized
    *
    * @see PublicKey
    */
  string getKeyEncoded(const string& format, const PublicKey& key) const;

  /**
    * Encodes the <i>key</i> using a format specified by <i>format</i>.
    *
    * @param format string specifies the format to be used. Only DER for now.
    * @param key PrivateKey to be encoded
    *
    * @return encoding a string of bytes
    *
    * @exception SecurityException if format is unsupported or if the
    *            encoding fails
    * @exception Error if key is uninitialized or unrecognized
    *
    * @see PrivateKey
    */
  string getKeyEncoded(const string& format, const PrivateKey& key) const;
 
  /** 
    * returns a <i>PublicKey</i>. The <i>format</i> specifies which format
    * was used to encode the key (must be DER for now). The <i>type</i> 
    * specifies the kind of key we are decoding. 
    * Examples of types are <b>DSA</b> and <b>RSA</b>.
    *
    * @param format string specifies the format used for encoding
    * @param type the key's type
    * @param encoding a string of bytes
    *
    * @return  a public key
    *
    * @exception SecurityException if either format or type is unsupported
    *                              or if decoding fails
    * @see PublicKey
    */
  PublicKey initPublicKey(const string& format, 
                          const string& encoding) const;

  /** 
    * returns a <i>PrivateKey</i>. The <i>format</i> specifies which format
    * was used to encode the key ( must be DER for now ) . The <i>type</i> 
    * specifies the kind of key we are decoding.
    * Examples of types are <b>DSA</b> and <b>RSA</b>.
    *
    * @param format string specifies the format used for encoding
    * @param type the key's type
    * @param encoding a string of bytes
    *
    * @return  a public key
    *
    * @exception SecurityException if either format or type is unsupported
    *            or if decoding fails
    *
    * @see PrivateKey
    */
  PrivateKey initPrivateKey(const string& format, 
                            const string& encoding) const;

  /**
    */
  PrivateKey initPrivateKeyFromFile(const string& format, 
                                    const string& fileName) const;
  /**
    * Generates an RSA key pair.
    *
    * @param size int modulo in bytes (power of 2)
    * @param size int exponent a prime number
    *
    * @return a KeyPair
    * @exception SecurityException if genaration fails
    *
    * @see KeyPair
    * @see PrivateKey
    * @see PublicKey
    */
  KeyPair generateRSAKeyPair(int size, unsigned long exponent) const;

  /**
    * Generates a DSA key pair.
    *
    * @param size int size of key in bytes
    *
    * @return a KeyPair
    *
    * @exception SecurityException if genaration fails
    *
    * @see KeyPair
    * @see PrivateKey
    * @see PublicKey
    */
  KeyPair generateDSAKeyPair(int size) const;

};

} // end namespace Akenti

#endif
