#ifndef OPEN_SSL_KEY_IMPL_HPP
#define OPEN_SSL_KEY_IMPL_HPP

#include <string>
#include <openssl/evp.h>

using std::string;

namespace Akenti {

/**
  * This class wraps an ssl ENV_PKEY structure which it assumes contains
  * a private and public RSA key pair. It implements methods to use
  * either the private or public key to encypt or decrypt an input string.
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class OpenSSLKeyImpl {

public:

  /**
    * Destructor
    */
  virtual ~OpenSSLKeyImpl();

  /**
    * Uses the contained private RSA key to decrypt the input string.
    *
    * @param encrypted the data to be decrypted
    * @returns the decyrpted string
    * @exception SecurityException OpenSSLRSAPrivateKeyImpl: decryption failure
    */
  string privateDecrypt(const string &encrypted) const;

  /**
    * Uses the contained private RSA key to encrypt the input string.
    *
    * @param data the data to be encrypted
    * @return the encyrpted string.
    * @exception SecurityException OpenSSLRSAPrivateKeyImpl: encryption failure
    */
  string privateEncrypt(const string &data) const;

  /**
    * Uses the contained public RSA key to decrypt the input string.
    *
    * @param encrypted the data to be decrypted
    * @return the decyrpted string
    * @exception SecurityException OpenSSLRSAPublicKeyImpl: decryption failure
    */
  string publicDecrypt(const string &encrypted) const;

  /**
    * Uses the contained public RSA key to encrypt the input string.
    *
    * @param data the data to be encrypted
    * @return the encyrpted string.
    * @exception SecurityException OpenSSLRSAPublicKeyImpl: encryption failure
    */
  string publicEncrypt(const string &data) const;

  /**
    * Returns ENV_PKey_size of the contained ENV_Pkey
    */
  int size() const;

  /**
    * Returns DER
    */
  string format() const;

  /**
    * Returns RSA
    */		
  string algorithm() const;

protected:

  /**
    * Constructor
    */
  OpenSSLKeyImpl(EVP_PKEY *evp);

  EVP_PKEY *evp;

};

} // end namespace Akenti

#endif
