#include <openssl/rand.h>
#include <openssl/evp.h>

#include "MessageDigest.hpp"
#include "Signature.hpp"
#include "CertificateFactory.hpp"
#include "KeyFactory.hpp"
#include "Cipher.hpp"
#include "Hmac.hpp"
#include "Base64.hpp"
#include "KeyStore.hpp"
#include "SSLContext.hpp"

#include "OpenSSLMessageDigestImpl.hpp"
#include "OpenSSLSignatureImpl.hpp"
#include "OpenSSLCipherImpl.hpp"
#include "OpenSSLHmacImpl.hpp"
#include "OpenSSLBase64Impl.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "OpenSSLKeyFactoryImpl.hpp"
#include "OpenSSLCertificateFactoryImpl.hpp"
#include "OpenSSLContextImpl.hpp"
#include "OpenSSLSecureSocketImpl.hpp"
#include "OpenSSLUtil.hpp"
#include "PKCS12KeyStoreImpl.hpp"
#include "DistinguishedName.hpp"

namespace Akenti {

typedef OpenSSLProviderImpl OSPI;

OSPI::OpenSSLProviderImpl() {
}

OSPI::~OpenSSLProviderImpl() {
}

void
OSPI::loadDigestAlgorithms() {
  OpenSSL_add_all_algorithms();
}

bool
OSPI::isDigestAlgorithmSupported(const string& alg) const {
  return EVP_get_digestbyname(alg.c_str());
}

MessageDigest 
OSPI::getMessageDigest(const string& alg) const {
  return new OpenSSLMessageDigestImpl(alg);
}

void
OSPI::loadSignatureAlgorithms() {
  OpenSSL_add_all_algorithms();
}

bool 
OSPI::isSignatureAlgorithmSupported(const string& alg)  const{
  return OpenSSLSignatureImpl::isSignatureAlgorithmSupported(alg);
}

Signature 
OSPI::getSignature(const string& alg) const {
  return new OpenSSLSignatureImpl(alg);
}

void
OSPI::loadCipherAlgorithms() {
  OpenSSL_add_all_ciphers();
}

bool
OSPI::isCipherAlgorithmSupported(const string& alg) const {
  return EVP_get_cipherbyname(alg.c_str());
}

Cipher
OSPI::getCipher(const string& alg) const {
  return new OpenSSLCipherImpl(alg);
}

Hmac 
OSPI::getHmac(const string& alg) const {
  return new OpenSSLHmacImpl(alg);
} 

KeyStore 
OSPI::getKeyStore(const string& type) const {
  if (type == "PKCS12" || type == "pkcs12") {
    return new PKCS12KeyStoreImpl(type);
  }
  
  throw SecurityException("OpenSSLProviderImpl:getKeyStore:type not supported" + type);
}

bool
OSPI::seedRandomGenerator(const string& seed) {
  RAND_seed(seed.data(), seed.length());
  return RAND_status();
}

Base64
OSPI::getBase64() const {
  return new OpenSSLBase64Impl;
}

KeyFactory 
OSPI::getKeyFactory() const {
  return new OpenSSLKeyFactoryImpl;
}

CertificateFactory 
OSPI::getCertificateFactory() const {
  return new OpenSSLCertificateFactoryImpl;
}

SSLContext 
OSPI::createSSLContext(SSLConstants::Version version, bool isClient) const {
  return SSLContext(new OpenSSLContextImpl(version, isClient));
}

SecureSocketImpl* 
OSPI::createSecureSocketImpl(const SSLContext& ctx) const {
  return new OpenSSLSecureSocketImpl(ctx);
}

string 
OSPI::toDerString(const DistinguishedName& dn) const {
  return OpenSSLUtil::toDer(dn);
}

}
