#ifndef OPEN_SSL_SECURE_SOCKET_IMPL_HPP
#define OPEN_SSL_SECURE_SOCKET_IMPL_HPP

#include <openssl/ssl.h>
#include "SecureSocketImpl.hpp"
#include "SSLContext.hpp"

namespace Akenti {

/**
  * This class implements the methods specific to OpenSSLsecureSockets and
  * inherits methods from SerureSocketImpl and SocketImpl. It contains
  * a pointer to an openssl SSL structure and  wraps openssl socket functions.
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class CertificateImpl;

class OpenSSLSecureSocketImpl : public SecureSocketImpl {

public:

  /**
   * Constructor - takes a security context, duplicates it and saves it
   *  in the new OpenSSLSecureSocketImpl.
   *
   * @param SSLContext& ctx - a security context
   * @exception Error don't recognize the SSLContext
   * @exception Error allocation failed
   */
  OpenSSLSecureSocketImpl(const SSLContext& ctx);

  /**
   * Destructor - free the SSL structure and destroys this object
   */
  ~OpenSSLSecureSocketImpl();

  /**
   * Creates a new SecureSocketImpl. This interface is also defined in the
   * superclass, SecureSocketImpl so that an application can create
   * a new SocketImpl without knowing what kind it has.
   *
   * @param SSLContext& ctx - a security context
   * @exception Error don't recognize the SSLContext
   * @exception Error allocation failed
   */
  SecureSocketImpl* newInstance(const SSLContext& ctx);

  CertificateImpl* doGetPeer() const;

  /**
   * Provides the SocketImpl with a secure socket (an SSL structure containing
   * a connected Unix socket)
   * that has been connected to the host and port associated with 
   * the SocketImpl. The IOHandler for the connected fd is
   * kept in the SocketImpl.
   * 
   * @exception SecurityException if the connection fails
   */
  void doConnect();

  /**
   *
   * @exception SecurityException if the handshake fails 
   */
  void doAccept(SecureSocketImpl *impl);

private:

  SSL* ssl;

};

} // end namespace Akenti

#endif

