#ifndef AK_CERTIFICATE_FACTORY_HPP
#define AK_CERTIFICATE_FACTORY_HPP

#include "ref_ptr.hpp"
#include "Provider.hpp"
#include "CertificateFactoryImpl.hpp"

/**
  * A <i>CertificateFactory</i> is a <i>reference-count</i> pointer
  * that wraps a <i>CertificateFactoryImpl</i>.  
  *
  * <pre>
  * 
  *  Usage:
  *  
  *  CertficateFactory fac = CertificateFactory::getInstance();
  *
  *  Certificate cert = fac->initCertificateFromFile("PEM", "/home/users/certs");
  *
  *  ...
  *  
  * </pre>
  *
  * @see CertificateFactoryImpl
  * @see ProviderImpl
  * @see Provider
  * @see ref_ptr
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class CertificateFactory : public ref_ptr<CertificateFactoryImpl> {

public:

  /**
    * Constructs this object given a <i>CertificateFactoryImpl</i>
    * object. Typically the <i>getInstance()</i> method is
    * used to get an instance.
    *
    * @see CertificateFactoryImpl
    * @see getInstance()
    */
  CertificateFactory(CertificateFactoryImpl *impl) : ref_ptr<CertificateFactoryImpl>(impl) {
  }

  /**
    * Destructor for CertificateFactory
    */
  ~CertificateFactory() {
  }

  /**
    * Returns the <i>CertificateFactory</i> from the installed <i>Provider</i>.
    *
    * @see Provider#getProvider
    * @see ProviderImpl#getCertificateFactory()
    */
  static CertificateFactory getInstance() {
    Provider& po = Provider::getProvider();

    return po->getCertificateFactory();
  }

};

} // end of namespace Akenti

#endif
