#ifndef AK_CERTIFICATE_REVOCATION_LIST_IMPL_HPP
#define AK_CERTIFICATE_REVOCATION_LIST_IMPL_HPP

#include <vector>
#include "RevokedCertificate.hpp"

using std::vector;

namespace Akenti {

class UtcTime;
class DistinguishedName;

/**
  * This abstract class provides the methods  of getting the
  * information associated with <i>CertificateRevocationList</i>
  * such as version, guarantor, signature,lastUpdate time, 
  * nextUpdate time, and also the list of the Certificates which
  * have been revoked. 
  *
  * @see DistinguishedName 
  * @see UtcTime  
  * @see RevokedCertificate 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class CertificateRevocationListImpl {

public:

  /**
    * Deletes the object.
    */ 
  virtual ~CertificateRevocationListImpl() {
  }
 
  /**
    * Returns the version of the CRL. 
    */
  virtual long getVersion() const = 0;

  /** 
    * Returns the guarantor of the CRL.
    */
  virtual DistinguishedName getGuarantor() const = 0;

  /**
    * Returns the signature of the CRL
    */
  virtual string getSignature() const = 0;

  /**
    * Returns the <i>UtcTime</i> of the last
    * <i>CertificateRevocationList</i> update time
    */ 
  virtual UtcTime getLastUpdate() const = 0;

  /**
    * Returns the <i>UtcTime</i> of the next
    * <i>CertificateRevocationList</i> update time.
    */
  virtual UtcTime getNextUpdate() const = 0;

  /**
    * Returns a <i>vector</i> which contains the
    * certificates have been revoked.
    */
  virtual vector<RevokedCertificate> getRevokedCertificates() const = 0;

  /**
    * Returns a <i>boolean</i> which specifies
    * the revocation status of the given serial number
    * of the certificate.
    * 
    * @param serialnumber serial number of the certificate
    */
  virtual bool isRevoked(long serialnumber) const = 0;

  /**
    * Returns a <i>int</i> which specifies the revocation
    * status (yes/no/unknown) of the given serial number
    * of the certificate.
    *
    * @param serialnumber the given serial number of the certificate
    * @param ctime specified length of time
    */
  virtual int isRevoked(long serialnumber,
		        const UtcTime& ctime) const = 0;

};

} // end of namespace Akenti

#endif
