#ifndef AK_KEY_FACTORY_HPP
#define AK_KEY_FACTORY_HPP 

#include "KeyFactoryImpl.hpp"
#include "Provider.hpp"
#include "ref_ptr.hpp"

namespace Akenti {

/**
  * A <i>KeyFactory</i> is a <b>reference-count</b> pointer that wraps
  * a <i>KeyFactoryImpl</i>. 
  *
  * <PRE>
  *
  * <b>Usage 1</b>
  *
  *    KeyFactory fac = KeyFactory::getInstance();
  *    KeyPair pair = fac->generateRSAKeyPair(); 
  *    string enc = fac->getKeyEncoded("DER", pair.getPrivateKey());
  *     ....
  *
  * </PRE>
  *
  * @see KeyFactoryImpl 
  * @see ProviderImpl
  * @see Provider
  * @see ref_ptr
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class KeyFactory : public ref_ptr<KeyFactoryImpl> {

public:

  /**
    * Constructs a <i>KeyFactory</i> given a <i>KeyFactoryImpl</i>.
    *
    * @see KeyFactoryImpl
    * @see ref_ptr
    */
  KeyFactory(KeyFactoryImpl *impl) : ref_ptr<KeyFactoryImpl>(impl) {
  }

  /**
    * Destroys this object.
    * 
    * @see ref_ptr
    */
  ~KeyFactory() {
  }

  /**
    * Returns the <i>KeyFactory</i> from the installed <i>Provider</i>.
    * 
    * @see SecurityManager#getProvider()
    * @see ProviderImpl#getKeyFactory()
    */
  static KeyFactory getInstance() {
    Provider& po = Provider::getProvider();

    return po->getKeyFactory();
  }

};

} // end of namespace Akenti

#endif
