#ifndef AK_KEY_FACTORY_IMPL_HPP
#define AK_KEY_FACTORY_IMPL_HPP 

#include <string>

using std::string;

namespace Akenti {

class PublicKey;
class PrivateKey;
class KeyPair;

/**
  * This abstract class provides applications with key generation, PEM or
  * DER encoding and decoding for key export and import. 
  *
  * Typically, an instance of this class is instantiated and wrapped in 
  * a <i>KeyFactory</i> object which is merely a <b>reference-count</b> pointer. 
  *
  * Applications should never have to instantiate a subclass of 
  * this class. Instead, the <i>KeyFactory#getKeyFactory()</i> should be used. 
  *
  * @see KeyFactory
  * @see Provider
  * @see PublicKey
  * @see PrivateKey
  * @see KeyPair
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  *
  */
class KeyFactoryImpl {

public:

  /**
    * Deletes this object.
    */
  virtual ~KeyFactoryImpl() {
  }
 
  /**
    * Encodes the public <i>key</i> using a format specified by <i>format</i>.
    *
    * @param format string specifies the format to be used 
    * @param key PublicKey to be encoded
    *
    * @return string encoding of the key
    *
    * @exception SecurityException if format is unsupported or if the 
    *                              encoding fails
    * @exception Error if key is unitialized or unrecognized 
    *
    * @see PublicKey
    */ 
  virtual string getKeyEncoded(const string& format, 
                               const PublicKey &key) const = 0;

  /**
    * Returns a <i>PublicKey</i>. 
    *
    * @param format the format used for encoding 
    * @param encoding a string of bytes 
    *
    * @exception SecurityException if either format or type is unsupported 
    *                              or if decoding fails 
    *
    * @see PublicKey
    */ 
  virtual PublicKey initPublicKey(const string& format, 
                                  const string& encoding) const = 0;

  /**
    * Encodes the private <i>key</i> using a format specified by <i>format</i>
    *
    * @param key PrivateKey to be encoded
    *
    * @param format the format used for encoding 
    * @return string the encoding of the key
    *
    * @exception SecurityException if format is unsupported or if the 
    *                              encoding fails
    * @exception Error if key is unitialized or unrecognized 
    *
    * @see PrivateKey
    */ 
  virtual string getKeyEncoded(const string& format, 
                               const PrivateKey &key) const = 0;

  /**
    * Returns a <i>PrivateKey</i>. The <i>format</i> specifies which format 
    * was used to encode the key, DER for nown. The <i>type</i> 
    * specifies the kind of key 
    * we are decoding. Examples of types are <b>DSA</b> and <b>RSA</b>.
    *
    * @param format string specifies the format used for encding 
    * @param type the key's type
    * @param encoding a string of bytes 
    *
    * @return PrivateKey 
    *
    * @exception SecurityException if format/type is unsupported 
    *                              or if decoding fails 
    * @see PrivateKey
    */ 
  virtual PrivateKey 
  initPrivateKey(const string& format, 
                  const string& encoding) const = 0;

  /**
    * Initiliazes the <i>PrivateKey</i> from a file.
    *
    * @param format specified format of the key, the format currently being 
    *   supported is <b>PEM</b>
    * @param fileName name of the file that contains the private key  
    * @return the Private Key that is initialized from the file
    */
  virtual PrivateKey 
  initPrivateKeyFromFile(const string& format, 
                         const string& fileName) const = 0;

  /**
    * Generates an RSA key pair. 
    *
    * @param size int modulo in bytes (power of 2)
    * @param size int exponent a prime number
    *
    * @return KeyPair  pair that holds the private and the public keys
    *
    * @exception SecurityException if generation fails
    *
    * @see KeyPair 
    * @see PrivateKey 
    * @see PublicKey 
    */
  virtual KeyPair generateRSAKeyPair(int size = 64, 
                                     unsigned long exponent = 3) const = 0;

  /**
    * Generates a DSA key pair. 
    *
    * @param size int size of key in bytes 
    *
    * @return KeyPair  pair that holds the private and the public keys
    *
    * @exception SecurityException if generation fails
    *
    * @see KeyPair 
    * @see PrivateKey 
    * @see PublicKey 
    */
  virtual KeyPair generateDSAKeyPair(int size = 64) const = 0;

protected:

  /**
    * Constructs this object.
    */
  KeyFactoryImpl() {
  }

};

} // end of namespace Akenti

#endif
