#ifndef AK_MESSAGE_DIGEST_IMPL_HPP
#define AK_MESSAGE_DIGEST_IMPL_HPP
#include "Error.hpp"

namespace Akenti {

/**
  * This abstract class provides applications with the interface
  * of a message digest algorithm, such as MD5 or SHA1. It also
  * provides some of the functionality. The rest of the functionality
  * is to be provided by the subclass implementors. Typically the clients
  * will not deal with an instance of this class directly. Instead they 
  * interact with the <i>MessageDigest</i> reference-count pointer.
  *
  * <p>
  * A MessageDigestImpl object starts out in an initial state containing
  * no data. Data is added to the digest object using the 
  * <i>update</i> methods. At any point  the <i>reset</i> method can
  * be called to reset the digest to the initial state of no data. 
  * Once all the data has been updated, one of the <i>digest</i> 
  * methods should be called to complete the hash computation.
  * After <i>digest</i> has been called, the <i>MessageDigestImpl</i>
  * object is reset to its initialized state.
  *
  * @see MessageDigest
  * @see Provider#getProvider()
  * @see ProviderImpl#getMessageDigest(const string&)
  * @see Error
  *  
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari 
  */

class MessageDigestImpl {

public:

  /**
    * Deletes this object.
    */
  virtual ~MessageDigestImpl() {
  }

  /**
    * Returns a string that identifies the algorithm, independent of
    * implementation details. The name should be a standard
    * name such as "MD5".
    * 
    * @return name of the algorithm
    */
  const string& getAlgorithm() const {
    return alg; 
  }

  /**
    * Updates the engine using the specified number of bytes, starting
    * at the specified offset in the data string.
    *
    * @param data input string of bytes.
    * @param offset the offset in data to start from
    * @param len the number of bytes to use starting at <i>offset</i>.
    * @exception Error if input buffer too short
    * @see doUpdate(const string&)
    */
  void update(const string& data, int offset, int len) {
    if (data.length() - offset < len) {
      throw Error("MessageDigestImpl:update:input buffer too short");
    }

    doUpdate(data, offset, len);
  }

  /**
    * Updates the engine using the specified string of bytes.
    *
    * @param data string to be updated. 
    * @see doUpdate(const string&)
    */
  void update(const string& data) {
    doUpdate(data, 0, data.length());
  }

  /**
    * Completes the hash computation.
    *
    * @return a string of bytes for the resulting hash value.
    */
  string digest() {
    string hash = doDigest();

    reset();
    return hash;
  }

  /**
    * Resets the digest to the initial state of no data.
    */
  virtual void reset() {
  } 

protected:

  friend class MessageDigest;

  /**
    * Constructs this object.  
    */
  MessageDigestImpl(const string& alg) : alg(alg) {
  }

  /**
    * Updates the digest using the specified number of bytes, starting
    * at the specified offset. Subclass implementors need not do any 
    * error checking since the update methods make sure that the arguments
    * are valid. 
    *
    * @param input string of bytes.
    * @param offset the offset to start from.
    * @param len the number of bytes to use starting at <i>offset</>.
    * @see update(const string&)
    * @see update(const string&, int, int)
    */
  virtual void doUpdate(const string& data, int offset, int len) {
  }
  
  /**
    * Returns the hash. 
    */
  virtual string doDigest() {
    return "";
  } 

  /**
    */
  string alg;

};

} // end of namespace Akenti

#endif
