#include "NetLogger.h"
#include <stdexcept>

#include <sys/time.h>

/*
 * NetLogger C++ API
 */
using std::string;
using std::runtime_error;

/* Constructor */
NetLogger::NetLogger()
{
   mHandle_p = NULL;
}

/* Open for output to URL 
 * see NetLoggerOpen() in C API.
 */
void
NetLogger::open (const string& progName, const string& url, 
		 int flags)
{
    mHandle_p = NetLoggerOpen( const_cast<char*>(progName.c_str()),
			       const_cast<char*>(url.c_str()),
			       flags );
    if (mHandle_p == NULL)
    {
        throw runtime_error ("Unable to open log file");
    }
}

/* Open for output to a file descriptor 
 */
void 
NetLogger::open(const std::string& progName, int fd, int flags )
{
    mHandle_p = NetLoggerOpenFd( const_cast<char*>(progName.c_str()),
			       fd,
			       flags );
    if (mHandle_p == NULL)
    {
        throw runtime_error ("Unable to open log file");
    }
}


/* Write a NetLogger message */   
void
NetLogger::write (const string& keyword, const string& mesg)
{
    struct timeval tv;
    gettimeofday (&tv, 0);
    if (NetLoggerGTWriteString (mHandle_p, tv.tv_sec, tv.tv_usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str())) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}

/* Write a NetLogger message with user-provided timestamp */
void
NetLogger::write (const string& keyword, 
                  const string& mesg, 
                  long sec, long usec)
{
    if (NetLoggerGTWriteString (mHandle_p, sec, usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str()) ) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}

/* Write a NetLogger message of a certain 'level' */   
void 
NetLogger::write(int lvl,
		 const std::string& keyword,
		 const std::string& mesg)
{
    struct timeval tv;
    gettimeofday (&tv, 0);
    if (NetLoggerGTWriteStringLvl (mHandle_p, lvl, tv.tv_sec, tv.tv_usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str())) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}
    
/* Write a NetLogger message of a certain 'level' 
 * with user-provided timestamp */
void 
NetLogger::write(int lvl, const std::string& keyword, 
		 const std::string& mesg, 
		 long sec, long usec)
{
    if (NetLoggerGTWriteStringLvl (mHandle_p, lvl, sec, usec,
				const_cast<char*>(keyword.c_str()), 
				const_cast<char*>(mesg.c_str()) ) == -1)
    {
	throw runtime_error ("Unable to write message");
    }
}

/* Flush buffered messages 
 * Return: 1=something was flushed; 0=no msgs flushed
 */
int
NetLogger::flush()
{
   return NetLoggerFlush( mHandle_p );
}

/* Set a 'tag' that is appended to all msgs --
 * for all instances of NetLogger!! */
void NetLogger::setTag( const string& tag )
{
   NetLoggerSetTag( const_cast<char *>(tag.c_str()) );
}

/* Flush buffer, close socket/file/syslog */
void
NetLogger::close ()
{
    if (NetLoggerClose (mHandle_p) == -1)
	throw runtime_error ("Unable to close the log");
}

/* Get the host name */
string NetLogger::getHostName()
{
    string s;
    if ( mHandle_p->hostname == NULL )
	s = "";
    else
	s = mHandle_p->hostname;
    return s;
}

/* Get the file name */
string NetLogger::getFileName() 
{
    string s;
    if (mHandle_p->filename == NULL)
	s = "";
    else
	s = mHandle_p->filename;
    return s;
}

/* Get the file descriptor */
int NetLogger::getFd()
{
    return mHandle_p->log_fd;
}

/* Get the current logging level */
int NetLogger::getLevel()
{
    return mHandle_p->loglvl;
}

/* Set the logging level. Level 0, the default, means "write all".
 * If !=0, write()s with a level that is greater than this level 
 * will return immediately.
 */
void 
NetLogger::setLevel(int lvl)
{
    NetLoggerSetLevel( mHandle_p, lvl );
}
