#include <sstream>
#include <iostream>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h> 
#include "AkentiServer.hpp"
#include "Provider.hpp"
#include "AkentiInit.hpp"
#include "SSLContext.hpp"
#include "SSLConstants.hpp"
#include "HttpConstants.hpp"
#include "Properties.hpp"
#include "SoapHttpRequest.hpp"
#include "SoapHttpResponse.hpp"
#include "PolicyCertificate.hpp"
#include "Certificate.hpp"
#include "CertificateAuthorityInfo.hpp"
#include "AkentiConstants.hpp"
#include "AkentiEngine.hpp"
#include "Policy.hpp"
#include "AkentiMessage.hpp"
#include "ProtocolHandler.hpp"
#include "SoapHttpRequest.hpp"
#include "SoapHttpResponse.hpp"
#include "Certificate.hpp"
#include "AkentiRequests.hpp"
#include "AkentiResponses.hpp"
#include "StringUtil.hpp"
#include "Exception.hpp"
#include "Error.hpp"

using namespace Akenti;
using namespace std;

#define DEBUG 1

AkentiServer::AkentiServer() {
  string status;

  if (!AkentiInit::init(status)) {
    throw Exception("Initializing system failed:" + status);
  } 
  
  Properties& props = Properties::getSystemProperties();
  string policyfilename = props.getValue("PolicyFileName", PolicyFileName);
  vector<pair<string,string> > resourceNamePairs = props.getKeyPairs();
  string sslMode = props.getValue("SSLMode");
  SSLContext serverContext = (secure = StringUtil::equalsIgnoreCase(sslMode, "on"))? 
                               Provider::getProvider()->createSSLContext(SSLConstants::TLS1, 
                                                                         false)
                               : NULL;

  bool found = false;

  for (int i=0; i < resourceNamePairs.size(); i++) {
    if (StringUtil::equalsIgnoreCase(resourceNamePairs[i].first, "RootResourceName")) {
      if (DEBUG) {
        cerr << "CHECKING POLICY FOR ROOT RESOURCE:" << resourceNamePairs[i].second << endl;
      }

      AkentiEngine eng;
      AkentiMessage mesg;
      PolicyContext ctx;
      bool ret = eng.checkPolicy(resourceNamePairs[i].second, ctx, mesg);

      if (DEBUG) {
	cerr << "ROOT RESOURCE:" << mesg.toString() << endl;
      }

      if (ret) {
        found = true;
      }

      if (ret && secure) {
        const vector<Policy>& policies = ctx->getPolicies();
        const vector<CertificateAuthorityInfo>& authorities = 
          policies[0]->getCertificateAuthorityInfos();

        for (int i=0; i < authorities.size(); i++) {
          const vector<Certificate>& chain = authorities[i].getChain();

          for (int j=0; j < chain.size(); j++) {
            serverContext->addTrustedCA(chain[j]);
          }
        }
      }
    }
  }

  if (!found) {
    throw Exception("No valid root resources");
  }


  string portString = props.getValue("AkentiPort", secure? "8443" : "8444");

  istringstream is(portString);
  int port;

  if (!(is >> port)) {
    throw Exception("Invalid port " + portString);
  }

  if (!secure) {
    serverSocket = ServerSocket(port);
    serverSocket.bind();
    serverSocket.listen(5);

    #if DEBUG
      cerr << "SERVER RUNNING AT PORT " + portString << endl;
    #endif
    return;
  }


  serverContext->setVerify(SSLConstants::VERIFY_PEER);
  serverContext->setPrivateKeyFile(props.getValue("ServerKeyFile"), SSLConstants::PEM);
  serverContext->setCertificateFile(props.getValue("ServerCertFile"), SSLConstants::PEM);
  sslServerSocket = SSLServerSocket(serverContext, port);
  sslServerSocket.bind();
  sslServerSocket.listen(5);
  
  #if DEBUG
    cerr << "SSL SERVER RUNNING AT PORT " + portString << endl;
  #endif
}

void
AkentiServer::service() {
  SoapHttpRequest req;
  int pid;

  req.setProtocol(HttpConstants::PROTOCOL);
  req.setMethod("POST");
  req.setContentType("text/xml");

  SoapHttpResponse resp;

  resp.setProtocol(HttpConstants::PROTOCOL);
  resp.setContentType("text/xml");

  InputStream in;
  OutputStream os;
  Certificate cert;
 
  if (secure) {
    SSLSocket sock = sslServerSocket.accept();

    in = sock.getInputStream();
    os = sock.getOutputStream();
    cert = sock.getPeerCertificate();
  } else {
    Socket sock = serverSocket.accept();

    in = sock.getInputStream();
    os = sock.getOutputStream();
  }

  signal(SIGCLD, SIG_IGN);  /* anti-zombie measure as children die */
  pid = fork();

  if (pid == 0) {
    try {
      req.fromStream(in);
    } catch(Exception& ex) {
      resp.setStatusCode(HttpConstants::HTTP_BAD);
      resp.setStatusMessage(HttpConstants::HTTP_BAD_MESSAGE);
      resp.toStream(os);
      throw ex;
    } catch(Error& err) { 
      resp.setStatusCode(HttpConstants::HTTP_ERROR);
      resp.setStatusMessage(HttpConstants::HTTP_ERROR_MESSAGE);
      resp.toStream(os);
      throw err;
    }
  
    try {
      AkentiRequests requests; 
   
      requests.fromXML(req.getParams()); 
   
      AkentiResponses responses; 
  
      ProtocolHandler::getInstance(ProtocolHandler::SERVER).handle(requests, responses, cert); 
      resp.setReturnValue(responses.toXML()); 
    } catch(Exception& ex) {
      resp.setStatusCode(HttpConstants::HTTP_ERROR);
      resp.setStatusMessage(HttpConstants::HTTP_ERROR_MESSAGE);
      resp.toStream(os);
      throw ex;
    } catch(Error& err) {
      resp.setStatusCode(HttpConstants::HTTP_ERROR);
      resp.setStatusMessage(HttpConstants::HTTP_ERROR_MESSAGE);
      resp.toStream(os);
      throw err;
    }

    resp.setStatusCode(HttpConstants::HTTP_OK);
    resp.setStatusMessage(HttpConstants::HTTP_OK_MESSAGE);
    resp.toStream(os);  // WE ARE DONE
    exit(0); // exit the child process - shouldnt have zombies
  }
}


void
AkentiServer::run() {
  while (true) {
    try {
      service();
    } catch(Exception& ex) {
      cerr << "SERVER:run:EXCEPTION" << ex.getMessage() << endl;
    } catch(Error& err) {
      cerr << "SERVER:run:FATAL ERROR(1):" << err.getMessage() << endl;
    } catch(...) {
      cerr << "SERVER:run:UNKNOWN FATAL ERROR(2):" << endl;
    }
  }
}

static void printUsage() {
  cerr << "Usage: akentiserver args" << endl
       << "       -f      - akenti server (engine) config file" << endl
       << "       -h      - prints this message" << endl;
}

int
main(int argc, char** argv) {
  try {
    string confile;
    int opt;

    while ((opt = getopt(argc,argv,"f:h")) != EOF) {
      switch(opt){
      case 'f':
        confile = optarg;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
      }
    }
    
    if (confile.length() == 0) {
      cerr << "Must supply config file" << endl;
    }
   
    Properties& props = Properties::getSystemProperties();
    string confFile(confile);
    
    #if DEBUG
    cerr << "SERVER LOADING CONFIG FILE " << confFile << " ..." << endl;
    #endif 

    props.load(confFile);

    AkentiServer server;

    server.run();
  } catch(Exception& ex) {
    cerr << "SERVER:" << ex << endl;
    exit(2);
  } catch(Error& err) {
    cerr << "SERVER:FATAL: report to akenti team:" << err << endl;
    exit(3);
  }
}
