#include <fstream>
#include <iostream>
#include <sstream>
#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"

#include "PolicyCertGeneratorTestCase.hpp"
#include "PolicyCertGenerator.hpp"
#include "KeyFactory.hpp"
#include "PrivateKey.hpp"
#include "UtcTime.hpp"
#include "URL.hpp"
#include "File.hpp"
#include "SecurityConstants.hpp" 
#include "CertificateFactory.hpp"
#include "Certificate.hpp"
#include "../shared/TestCaseConstants.hpp"
#include "XMLUtil.hpp"

using namespace Akenti;
using namespace std;

typedef PolicyCertGeneratorTestCase PCGTC;

void 
PCGTC::generatePolicyCerts() {
  string method;

  try {
    method = "PolicyCertGeneratorTestCase:generateROOT1PolicyCert:";
    generateROOT1PolicyCert();
    method = "PolicyCertGeneratorTestCase:generateROOT2PolicyCert:";
    generateROOT2PolicyCert();
    method = "PolicyCertGeneratorTestCase:generateSUB1PolicyCert:";
    generateSUB1PolicyCert();
    method = "PolicyCertGeneratorTestCase:generateSUB3PolicyCert:";
    generateSUB3PolicyCert();
    method = "PolicyCertGeneratorTestCase:generateSUB4PolicyCert:";
    generateSUB4PolicyCert();
    method = "PolicyCertGeneratorTestCase:generateROOT3PolicyCert:";
    generateROOT3PolicyCert();
    method = "PolicyCertGeneratorTestCase:generateROOT4PolicyCert:";
    generateROOT4PolicyCert();
  } catch (Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch (Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } 
}

void
PCGTC::generateROOT1PolicyCert() {
#if DEBUG
  cout << "\nGenerating ROOT1 Policy Certificate ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  Directory idDirectory("file:" + File(ID_DIR).getAbsolutePath());
  vector<Directory> idDirectories(1, idDirectory);

  Directory attrDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> attrDirectories(1, attrDirectory);

  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  vector<AkentiPrincipal> ucIssuers(1, guarantor);
  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  CertificateFactory cfac = CertificateFactory::getInstance();
  vector<Certificate> trustedCAs = cfac->initCertificatesFromFile(
    SecurityConstants::PEM, ID_DIR + CAI_CERT 
  );
  CertificateAuthorityInfo info(trustedCAs[0], idDirectories);
  vector<CertificateAuthorityInfo> infos(1, info);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setCacheTime(30 * 60);
  gen.setResource(ROOT1);
  gen.setIssuer(guarantor);
  gen.setAttributeDirectories(attrDirectories);
  gen.setUseCondIssuerGroups(groups);
  gen.setCertificateAuthorityInfos(infos);

  KeyFactory kfac = KeyFactory::getInstance();

  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = ROOT1_DIR + POLICY_FILE_NAME;
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());
   

{
  ostringstream os;
  os << cert->toText() << endl;
  AkentiCertificate::writeToStream("PEM", os, cert);
  istringstream is(os.str());
  AkentiCertificate certFromStream = AkentiCertificate::initFromStream("PEM", is);
 
  CPPUNIT_ASSERT(cert == certFromStream);
}

{
  ostringstream os;
  AkentiCertificate::writeToStream("XML", os, cert);
  istringstream is(os.str());
  AkentiCertificate certFromStream = AkentiCertificate::initFromStream("XML", is);

  CPPUNIT_ASSERT(cert == certFromStream);
}

#if DEBUG
  cout << "\nSaved ROOT1 Policy Certificate in " << fileName << endl;
#endif
 
}

void
PCGTC::generateROOT2PolicyCert() {
#if DEBUG
  cout << "\nGenerating ROOT2 Policy Certificate ... " << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  // Issuer/Guarantor of the root policy
  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  // id certs Directories
  Directory idDirectory("file:" + File(ID_DIR).getAbsolutePath());
  vector<Directory> idDirectories(1, idDirectory);

  // attr certs Directories
  Directory attrDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> attrDirectories(1, attrDirectory);

  // UC certs directories
  vector<AkentiPrincipal> ucIssuers;
  ucIssuers.push_back(guarantor);
  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  // groups for this resource
  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  // getting the trusted CA certs from a file  
  CertificateFactory cfac = CertificateFactory::getInstance();

  // path + filename to CAIA certificate 
  vector<Certificate> trustedCAs = cfac->initCertificatesFromFile(
    SecurityConstants::PEM, ID_DIR + CAI_CERT 
  );

  CertificateAuthorityInfo info(trustedCAs[0], idDirectories);
  vector<CertificateAuthorityInfo> infos(1, info);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setCacheTime(30 * 60);
  gen.setResource(ROOT2);
  gen.setIssuer(guarantor);
  gen.setAttributeDirectories(attrDirectories);
  gen.setUseCondIssuerGroups(groups);
  gen.setCertificateAuthorityInfos(infos);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = ROOT2_DIR + POLICY_FILE_NAME;

  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "Saved ROOT2 Policy Certificate in " << fileName << endl;
#endif
}

void
PCGTC::generateSUB1PolicyCert() {
#if DEBUG
  cout << "\nGenerating SUB1 Policy Certificate ... " << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  Directory attrDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> attrDirectories(1, attrDirectory);

  vector<AkentiPrincipal> ucIssuers(1, guarantor);
  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(SUB1);
  gen.setIssuer(guarantor);
  gen.setAttributeDirectories(attrDirectories);
  gen.setUseCondIssuerGroups(groups);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();

  string fileName = SUB1_DIR + POLICY_FILE_NAME;

  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());
    
#if DEBUG
  cout << "\nGenerated SUB1 Policy Certificate" << endl;
#endif
}

void
PCGTC::generateSUB3PolicyCert() {
#if DEBUG
  cout << "\nGenerating SUB3 Policy Certificate ... " << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<AkentiPrincipal> ucIssuers(1, guarantor);
  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(SUB3);
  gen.setIssuer(guarantor);
  gen.setUseCondIssuerGroups(groups);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = SUB3_DIR + POLICY_FILE_NAME;
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "Saved SUB1 Policy Certificate in " << fileName << endl;
#endif
}


void
PCGTC::generateSUB4PolicyCert() {
#if DEBUG
  cout << "\nGenerating SUB4 Policy Certificate ... " << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<AkentiPrincipal> ucIssuers(1, guarantor);
  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setCacheTime(60 * 60);
  gen.setResource(SUB4);
  gen.setIssuer(guarantor);
  gen.setUseCondIssuerGroups(groups);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();

  string fileName = SUB4_DIR + POLICY_FILE_NAME;

  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "\nSaved SUB4 Policy Certificate" << endl;
#endif
}

void
PCGTC::generateROOT3PolicyCert() {
#if DEBUG
  cout << "\nGenerating ROOT3 Policy Certificate ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  Directory idDirectory("file:" + File(ID_DIR).getAbsolutePath());
  vector<Directory> idDirectories(1, idDirectory);

  Directory attrDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> attrDirectories(1, attrDirectory);

  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  vector<AkentiPrincipal> ucIssuers(1, guarantor);
  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  CertificateFactory cfac = CertificateFactory::getInstance();
  vector<Certificate> trustedCAs = cfac->initCertificatesFromFile(
    SecurityConstants::PEM, ID_DIR + CAI_CERT 
  );
  CertificateAuthorityInfo info(trustedCAs[0], idDirectories);
  vector<CertificateAuthorityInfo> infos(1, info);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setCacheTime(30 * 60);
  gen.setResource(ROOT3);
  gen.setIssuer(guarantor);
  gen.setAttributeDirectories(attrDirectories);
  gen.setUseCondIssuerGroups(groups);
  gen.setCertificateAuthorityInfos(infos);

  KeyFactory kfac = KeyFactory::getInstance();

  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = ROOT3_DIR + POLICY_FILE_NAME;
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());
   
#if DEBUG
  cout << "\nSaved ROOT3 Policy Certificate in " << fileName << endl;
#endif
}

void
PCGTC::generateROOT4PolicyCert() {
#if DEBUG
  cout << "\nGenerating ROOT4 Policy Certificate ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  Directory idDirectory("file:" + File(ID_DIR).getAbsolutePath());
  vector<Directory> idDirectories(1, idDirectory);

  Directory attrDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> attrDirectories(1, attrDirectory);

  Directory ucDirectory("file:" + File(AKCERT_DIR).getAbsolutePath());
  vector<Directory> ucDirectories(1, ucDirectory);

  vector<AkentiPrincipal> ucIssuers(1, guarantor);
  UseCondIssuerGroup group(ucIssuers, ucDirectories);
  vector<UseCondIssuerGroup> groups(1, group);

  CertificateFactory cfac = CertificateFactory::getInstance();
  vector<Certificate> trustedCAs = cfac->initCertificatesFromFile(
    SecurityConstants::PEM, ID_DIR + CAI_CERT 
  );
  CertificateAuthorityInfo info(trustedCAs[0], idDirectories);
  vector<CertificateAuthorityInfo> infos(1, info);

  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setCacheTime(30 * 60);
  gen.setResource(ROOT4);
  gen.setIssuer(guarantor);
  gen.setAttributeDirectories(attrDirectories);
  gen.setUseCondIssuerGroups(groups);
  gen.setCertificateAuthorityInfos(infos);

  KeyFactory kfac = KeyFactory::getInstance();

  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = ROOT4_DIR + POLICY_FILE_NAME;
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());
   
#if DEBUG
  cout << "\nSaved ROOT4 Policy Certificate in " << fileName << endl;
#endif
}

CppUnit::TestSuite*
PCGTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("PolicyCertGeneratorTestCase");

  testSuite->addTest (
    new CppUnit::TestCaller<PolicyCertGeneratorTestCase> ( 
      "generatePolicyCertificates", 
      &PolicyCertGeneratorTestCase::generatePolicyCerts
    )
  );

  return testSuite;
}

