#include <iostream>
#include <fstream>

#include <cppunit/Exception.h>
#include <cppunit/TestCaller.h>
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"
#include "X509CertTestCase.hpp"
#include "CertificateFactory.hpp"
#include "CertificateGenerator.hpp"
#include "KeyFactory.hpp"
#include "KeyPair.hpp"
#include "MessageDigest.hpp"
#include "SecurityConstants.hpp"
#include "Util.hpp"
#include "File.hpp"
#include "TestCaseConstants.hpp"

using namespace std;
using namespace Akenti;

typedef X509CertTestCase XCTC;

void
XCTC::generateX509Certificates() {
  string method;

  try {
    method = "X509CertTestCase:generateCAICert:";
    generateCAICert();
    method = "X509CertTestCase:generateCAIACert:";
    generateCAIACert();
    method = "X509CertTestCase:generateUserICert:";
    generateUserICert();    
    method = "X509CertTestCase:generateSHICert:";
    generateSHICert();
    method = "X509CertTestCase:generateServerCert:";
    generateServerCert();
    method = "X509CertTestCase:generateClientCert:";
    generateClientCert();
    method = "X509CertTestCase:generateProxyCert:";  
    generateProxyCert();
  } catch (Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());
 
    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch (Error& err) {
    CppUnit::Message m(method + err.getMessage());
 
    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  }  
}

void 
XCTC::generateCAICert() {
#if DEBUG
  cout << "\nGenerating CA self-signed X509 Certificate ... " << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif 

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Generated key keyPair" << endl;
#endif 

  string keyFileName = ID_DIR + CAI_KEY;

#if DEBUG
  cout << "Saving CA's private key in " << keyFileName << " ..." << endl;
#endif 

  ofstream os1(keyFileName.c_str());

  os1 << kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());

#if DEBUG
  cout << "Saved CA's private key in " << keyFileName << endl;
#endif 

  Certificate cert = generateCertificate(
    keyPair.getPrivateKey(),
    CAI_DN,
    keyPair.getPublicKey(),
    CAI_DN 
  );

#if DEBUG
  cout << "Generated CA's x509 cert" << endl;
#endif 

  string certFileName = ID_DIR + CAI_CERT;

#if DEBUG
  cout << "Saving CA's x509 cert in " << certFileName << " ..." << endl;
#endif 

  CertificateFactory cfac = CertificateFactory::getInstance();
  ofstream os2(certFileName.c_str());

  os2 << cert->toText() << endl;
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());

#if DEBUG
  cout << "Saved CA's x509 cert in " << certFileName << endl;
#endif 

#if DEBUG
  cout << "TESTING READING X509 CERT AND VERIFICATION ..." << endl;
  Certificate tempCert = cfac->initCertificateFromFile(
    SecurityConstants::PEM,
    certFileName
  );

  CPPUNIT_ASSERT(tempCert->verify(keyPair.getPublicKey()));
  cout << "READING X509 CERT AND VERIFICATION OK" << endl;
#endif
}

void 
XCTC::generateCAIACert() {
#if DEBUG
  cout << "\nGenerating CAIA X509 Certificate ... " << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif 

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Generated key keyPair" << endl;
#endif 

  string keyFileName = ID_DIR + CAIA_KEY;

#if DEBUG
  cout << "Saving CAIA's private key in " << keyFileName << " ..." << endl;
#endif 

  ofstream os1(keyFileName.c_str());

  os1 << kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());

#if DEBUG
  cout << "Saved CAIA's private key in " << keyFileName << endl;
#endif 

  Certificate cert = generateCertificate(
    kfac->initPrivateKeyFromFile(SecurityConstants::PEM, ID_DIR + CAI_KEY),
    CAI_DN,
    keyPair.getPublicKey(),
    CAIA_DN 
  );

#if DEBUG
  cout << "Generated CAIA's x509 cert" << endl;
#endif 

  string hash = Util::createFileName(MessageDigest::digest("MD5", cert->getHashString()));
  string certFileName = ID_DIR + CAIA_CERT + "." + hash + "." + "0";


#if DEBUG
  cout << "Saving CAIA's x509 cert in " << certFileName << " ..." << endl;
#endif 

  CertificateFactory cfac = CertificateFactory::getInstance();
  ofstream os2(certFileName.c_str());

  os2 << cert->toText() << endl;
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());

#if DEBUG
  cout << "Saved CAIA's x509 cert in " << certFileName << endl;
#endif 
}

void
XCTC::generateUserICert() {
#if DEBUG
  cout << "\nGenerating UserI's X509 Certificate ..." << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Generated key keyPair" << endl;
#endif 

  string keyFileName = ID_DIR + USERI_KEY;

#if DEBUG
  cout << "Saving User's private key in " << keyFileName << " ..." << endl;
#endif 

  ofstream os1(keyFileName.c_str());

  os1 <<  kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());

#if DEBUG
  cout << "Saved User's private key in " << keyFileName << endl;
#endif 

  Certificate cert = generateCertificate(
    kfac->initPrivateKeyFromFile(SecurityConstants::PEM, ID_DIR + CAIA_KEY),
    CAIA_DN,
    keyPair.getPublicKey(),
    USERI_DN 
  );

#if DEBUG
  cout << "Generated User1's x509 certificate" << endl;
#endif

  string hash = Util::createFileName(MessageDigest::digest("MD5", cert->getHashString()));
  string certFileName = ID_DIR + USERI_CERT + "." + hash + "." + "0";

#if DEBUG
  cout << "Saving UserI 's x509 cert in " << certFileName << " ..." << endl;
#endif

  CertificateFactory cfac = CertificateFactory::getInstance();
  ofstream os2(certFileName.c_str());

  os2 << cert->toText() << endl;
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());

#if DEBUG
  cout << "Saved User1 x509 cert in " << certFileName << endl;
#endif
}

void
XCTC::generateSHICert() {
#if DEBUG
  cout << "\nGenerating StakeHolderI's x509 cert ..." << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Generated key keyPair" << endl;
#endif

  string keyFileName = ID_DIR + SHI_KEY;

#if DEBUG
  cout << "Saving StakeHolderI's private key in " << keyFileName << " ..." << endl;
#endif

  ofstream os1(keyFileName.c_str());

  os1 << kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());

#if DEBUG
  cout << "Saved StakeHolderI's private key in " << keyFileName << endl;
#endif

  Certificate cert = generateCertificate(
    kfac->initPrivateKeyFromFile(SecurityConstants::PEM, ID_DIR + CAIA_KEY),
    CAIA_DN,
    keyPair.getPublicKey(),
    SHI_DN 
  );

#if DEBUG
  cout << "Generated StakeHolderI's x509 cert" << endl;
#endif

  string hash = Util::createFileName(MessageDigest::digest("MD5", cert->getHashString()));
  string certFileName = ID_DIR + SHI_CERT + "." + hash + "." + "0";

#if DEBUG
  cout << "Saving StakeHolderI's cert in " << certFileName << " ..." << endl;
#endif

  CertificateFactory cfac = CertificateFactory::getInstance();
  ofstream os2(certFileName.c_str());

  os2 << cert->toText() << endl; 
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());
  
#if DEBUG
  cout << "Saved StakeHolderI's cert in " << certFileName << endl;
#endif
}

void
XCTC::generateServerCert() {
#if DEBUG
  cout << "\nGenerating server X509 Certificate ... " << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Generated RSA key keyPair ..." << endl;
#endif

  string keyFileName = ID_DIR + SERVER_KEY; 

#if DEBUG
  cout << "Saving server's private key in " << keyFileName << " ..." << endl;
#endif

  ofstream os1(keyFileName.c_str());

  os1 <<  kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());

#if DEBUG
  cout << "Saved server's private key in " << keyFileName << endl;
#endif

  Certificate cert = generateCertificate(
    kfac->initPrivateKeyFromFile(SecurityConstants::PEM, ID_DIR + CAI_KEY),
    CAI_DN,
    keyPair.getPublicKey(),
    SERVER_DN 
  );

#if DEBUG
  cout << "Generated server's X509 Certificate ... " << endl;
#endif

  string certFileName = ID_DIR + SERVER_CERT;

#if DEBUG
  cout << "Saving server's x509 cert in " << certFileName << " ..." << endl;
#endif

  CertificateFactory cfac = CertificateFactory::getInstance();
  ofstream os2(certFileName.c_str());

  os2 << cert->toText() << endl; 
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());

#if DEBUG
  cout << "Saved server's x509 cert in " << certFileName << endl;
#endif
}
 
void
XCTC::generateClientCert() {
#if DEBUG
  cout << "\nGenerating client X509 Certificate ... " << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Generated RSA key keyPair ..." << endl;
#endif

  string keyFileName = ID_DIR + CLIENT_KEY; 

#if DEBUG
  cout << "Saving client's private key in " << keyFileName << " ..." << endl;
#endif

  ofstream os1(keyFileName.c_str());

  os1 <<  kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());

#if DEBUG
  cout << "Saved client's private key in " << keyFileName << endl;
#endif

  Certificate cert = generateCertificate(
    kfac->initPrivateKeyFromFile(SecurityConstants::PEM, ID_DIR + CAI_KEY),
    CAI_DN,
    keyPair.getPublicKey(),
    CLIENT_DN 
  );

#if DEBUG
  cout << "Generated client's X509 Certificate ... " << endl;
#endif

  string certFileName = ID_DIR + CLIENT_CERT;

#if DEBUG
  cout << "Saving client's x509 cert in " << certFileName << " ..." << endl;
#endif

  CertificateFactory cfac = CertificateFactory::getInstance();
  ofstream os2(certFileName.c_str());

  os2 << cert->toText() << endl; 
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());

#if DEBUG
  cout << "Saved client's x509 cert in " << certFileName << endl;
#endif
}

void
XCTC::generateProxyCert() {
#if DEBUG
  cout << "\nGenerating proxy's X509 Certificate" << endl;
  cout << "Generating RSA key keyPair ..." << endl;
#endif

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair keyPair = kfac->generateRSAKeyPair();

#if DEBUG
  cout << "Genered key keyPair ..." << endl;
#endif

  string keyFileName = ID_DIR + PROXY_KEY;

#if DEBUG
  cout << "Saving proxy's key in " << keyFileName << " ..." << endl;
#endif

  ofstream os1(keyFileName.c_str());

  os1 <<  kfac->getKeyEncoded(SecurityConstants::PEM, keyPair.getPrivateKey());
  os1.close();
  CPPUNIT_ASSERT(!os1.fail());
#if DEBUG
  cout << "Saved proxy's key in " << keyFileName << endl;
#endif

  /**
    * Here we are looking for USERI's certificate
    * We only need the User's dn which obviously
    * stored in <i>USERI_DN</i> but we do it the
    * the long way so that we can test  der encoding
    * of the dn.
    **/
  
  string hash = Util::createFileName(
    MessageDigest::digest(
      "MD5",
      Provider::getProvider()->toDerString(USERI_DN)
    )
  );

  string certFileName = ID_DIR + USERI_CERT + "." + hash + "." + "0";

#if DEBUG
  cout << "LOADING USERI CERT FROM " << certFileName << endl;
#endif

  CertificateFactory cfac = CertificateFactory::getInstance();
  Certificate uICert = cfac->initCertificateFromFile(SecurityConstants::PEM, certFileName);

#if DEBUG
  cout << "FOUND USERI CERT IN " << certFileName << endl;
#endif

  Certificate cert = generateCertificate(
    kfac->initPrivateKeyFromFile(SecurityConstants::PEM, ID_DIR + USERI_KEY),
    uICert->getPrincipal(),
    keyPair.getPublicKey(),
    PROXY_DN 
  );

#if DEBUG
  cout << "Generated Proxy x509 certificate" << endl;
#endif

  hash = Util::createFileName(MessageDigest::digest("MD5", cert->getHashString()));
  certFileName = ID_DIR + PROXY_CERT + "." + hash + "." + "0";

#if DEBUG
  cout << "Saving proxy X509 Certificate in " << certFileName << " ..." << endl;
#endif

  ofstream os2(certFileName.c_str());

  os2 << cert->toText();
  os2 << cfac->getCertificateEncoded(SecurityConstants::PEM, cert);
  os2.close();
  CPPUNIT_ASSERT(!os2.fail());

#if DEBUG
  cout << "Saved proxy X509 Certificate in " << certFileName << endl;
#endif
}

Certificate 
XCTC::generateCertificate(const PrivateKey& privKey, 
                          const DistinguishedName& issuer, 
                          const PublicKey& pubKey,
                          const DistinguishedName& subject) {

  static int i = 0;

  CertificateFactory cfac = CertificateFactory::getInstance();
  CertificateGenerator cgen = cfac->getCertificateGenerator("X509");

  cgen->setSubjectPublicKey(pubKey);
  cgen->setSubject(subject);
  cgen->setValidity(300);
  cgen->setIssuer(issuer);
  cgen->setSerialNumber(++i);
  cgen->setVersionNumber(2);
  return cgen->generate(SecurityConstants::RSA_SHA1, privKey);
}
  
CppUnit::TestSuite*
XCTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("X509CertTestCase");

  testSuite->addTest(
    new CppUnit::TestCaller<X509CertTestCase>(
      "generateX509Certificates", 
      &X509CertTestCase::generateX509Certificates
    )
  );

  return testSuite;
}
