#include <iostream>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "Certificate.hpp"
#include "CertificateFactory.hpp"
#include "MessageDigest.hpp"
#include "DistinguishedName.hpp"
#include "Util.hpp"
#include "File.hpp"
#include "Exception.hpp"
#include "Error.hpp"

using namespace Akenti;
using namespace std;

/**
  * Program to generate the hash of an X509 certificate
  *
  * Usage: HashX509 -x X509Cert (in PEM format) [-v] [-h]
  */
namespace HashX509 {

void printPurpose() {
  cerr << "Purpose: Use this program to generate the hash of an X509 cert." << endl
       << endl;
}

void printUsage() {
  cerr << "Usage: HashX509 args" << endl
       << "  -x arg    - X509 certificate (in PEM format)" << endl
       << "  [-v]      - verbose" << endl
       << "  [-h]      - prints this message" << endl;
}

int
main(int argc, char **argv) {
   if (argc < 2 ) {
     printPurpose();
     printUsage();
     exit(1);
   }

   string x509CertFile;
   bool verbose = false;
   int opt;

   while ((opt = getopt(argc,argv,"x:vh")) != EOF) {
     switch (opt) {
       case 'x':
         x509CertFile = optarg;
         break;
       case 'v':
         verbose = true;
         break;
       case 'h':
       default:
         printUsage();
         exit(2);
     }
   }

   try {
     Provider po = Provider(new OpenSSLProviderImpl);

     po->loadDigestAlgorithms();

     if (verbose) {
       cout << "Installing security provider ..." << endl;
     }

     Provider::setProvider(po);

     if (verbose) {
       cout << "Opening x509 cert file " << x509CertFile << endl;
     }

     File file(x509CertFile);

     if (!file.exists() || !file.isFile()) {
       cerr << "Can't open x509 cert file at " << file.getAbsolutePath() << endl;
       exit(2);
     }

     CertificateFactory certFac = CertificateFactory::getInstance();
     Certificate x509Cert = certFac->initCertificateFromFile("PEM", x509CertFile);
     const DistinguishedName& issuerDN = x509Cert->getPrincipal();
     string dn = Provider::getProvider()->toDerString(issuerDN);
     string akentiDigest = MessageDigest::digest("MD5", dn);
     string openSSLDigest = MessageDigest::digest("MD5", x509Cert->getHashString());

     if (akentiDigest != openSSLDigest) {
       cerr << "Akenti hash is: " << Util::createFileName(akentiDigest) << endl
            << "OpenSSL hash is: " << Util::createFileName(openSSLDigest) << endl;
       throw Error("Open ssl hash differs");
     }

     cout << "Akenti hash is: " << Util::createFileName(akentiDigest) << endl;
     exit(0);
   } catch(Exception& ex) {
     cerr << ex.getMessage() << endl;
   } catch(Error& err) {
     cerr << "ERROR:Rrport to akenti team:" + err.getMessage() << endl;
   }
}

}
