#include <iostream>
#include <fstream>
#include <string>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "Certificate.hpp"
#include "CertificateFactory.hpp"
#include "CertificateGenerator.hpp"
#include "PolicyCertGenerator.hpp"
#include "UseCondCertGenerator.hpp"
#include "AttributeCertGenerator.hpp"
#include "KeyFactory.hpp"
#include "KeyPair.hpp"
#include "AkentiCertificate.hpp"
#include "AttributeInfo.hpp"
#include "Condition.hpp"
#include "File.hpp"
#include "URL.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "XMLUtil.hpp"

using namespace std;
using namespace Akenti;

/**
  * <pre>
  * <b>PrintAkentiCertTemplate</b>
  *   Use this program to print a certificate template.
  * </pre>
  */

namespace PrintAkentiCertTemplate {

void 
printPurpose() {
  cerr << "Use this program to print an akenti cert template." << endl
       << endl;
}

void 
printUsage() {
  cerr << "Usage: PrintAkentiCertTemplate args" << endl
       << "  [-R]        - Root Policy certificate" << endl
       << "  [-P]        - Policy certificate" << endl
       << "  [-U]        - Use-Condition certificate " << endl
       << "  [-A]        - Attribute certificate" << endl
       << "  [-o]        - output file name" << endl
       << "  [-v]        - verbose"<< endl 
       << "  [-h]        - prints this message "<< endl; 
}

int
main(int argc, char *argv[]) {
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  string outputCertFileName;
  string type;
  bool verbose = false;
  bool root;
  int opt;

  while ((opt = getopt(argc, argv, "o:PRUAvh")) != EOF) {
    switch(opt){
      case 'R':
        type = "Policy";
        root = true;
        break;
      case 'P':
        type = "Policy";
        root = false;
        break;
      case 'U':
        type = "UseCondition";
        break;
      case 'A':
        type = "Attribute";
        break;
      case 'o':
        outputCertFileName = optarg;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
    }
  }

  try {
    if (verbose) {
      cout << "Initializing XML library ..." << endl;
    }

    XMLUtil::init();

    if (verbose) {
      cout << "Setting security provider ..." << endl;
    }

    Provider po = Provider(new OpenSSLProviderImpl);
    
    po->loadDigestAlgorithms();
    po->loadSignatureAlgorithms();
 
    bool seeded = po->seedRandomGenerator(
      "WE need to seed the random generator"
    );
 
    if (!seeded) {
      throw Exception("Unable to seed generator");
    }

    Provider::setProvider(po);

    AkentiCertificate cert;

    if (type == "UseCondition") {
      string constraint = "GROUP = NO_GROUP_NAME  || CN = NO_CN && TIME > NO_TIME";
      AttributeInfo at1(true, "GROUP", "NO_GROUP_NAME", vector<AkentiPrincipal>(1));
      AttributeInfo at2("CN", "NO_CN", vector<DistinguishedName>(1));
      AttributeInfo at3("TIME", "NO_TIME");
      vector<AttributeInfo> attrInfos;
 
      attrInfos.push_back(at1);
      attrInfos.push_back(at2);
      attrInfos.push_back(at3);
      Condition condition(constraint, attrInfos);
      UseCondCertGenerator gen;
      
      gen.setCondition(condition);

      vector<string> actions;

      actions.push_back("read");
      actions.push_back("write");
      gen.setActions(actions);
      cert = gen.generate();
    } else if (type == "Attribute") {
      cert = AttributeCertGenerator().generate();
    } else if (type == "Policy" && ! root) {
      PolicyCertGenerator gen;

      gen.setResource("NO_ROOT_RESOURCE/NO_SUB_RESOURCE");
      cert = gen.generate();
    } else if (type == "Policy" && root) {
      PolicyCertGenerator gen;

      gen.setResource("NO_ROOT_RESOURCE");
      
      CertificateFactory cfac = CertificateFactory::getInstance();
      CertificateGenerator x509Gen = cfac->getCertificateGenerator("X509");
      KeyPair pair = KeyFactory::getInstance()->generateRSAKeyPair();
      
      x509Gen->setSubjectPublicKey(pair.getPublicKey());
      Certificate caCert = x509Gen->generate("RSA-MD5", pair.getPrivateKey());

      vector<Directory> dirs(1, URL().toString());
      CertificateAuthorityInfo cai(caCert, dirs);
      vector<CertificateAuthorityInfo> cais(1, cai);

      gen.setCertificateAuthorityInfos(cais);
      cert = gen.generate();
    } else {
      cerr << "Must select a certificate type" << endl;
      exit(2);
    }

    if (!outputCertFileName.length()) {
      AkentiCertificate::writeToStream("XML", cout, cert);
    } else {
      if (verbose) {
        cout << "Saving template to " << outputCertFileName << endl;
      }

      AkentiCertificate::writeToFile("XML", outputCertFileName, cert);
    }
   
    exit(0);
  } catch (Exception& ex) {
    cerr << "Exception:" << ex.getMessage() << endl;
    exit(2);
  } catch (Error& err) {
    cerr << "Error:Report to Akenti team:" << err.getMessage() <<endl;
    exit(3);
  }
}

}
