#include <sstream>
#include <iostream>
#include <string>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "CertificateChain.hpp"
#include "CheckAccessRequest.hpp"
#include "CheckAccessResponse.hpp"
#include "AkentiCertificate.hpp"
#include "CapabilityCertificate.hpp"
#include "ProtocolHandler.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "SSLContext.hpp"
#include "XMLUtil.hpp"

using namespace Akenti;
using namespace std;

/**
  * <pre>
  * <b>RemoteCheckAccess</b>
  *   Use this program to check a user's access to a remote resource.
  * </pre>
  */
namespace RemoteCheckAccess {

void printPurpose() {
  cerr << "Use this program to check a user's access to a remote resource." << endl 
       << endl;
}

void printUsage () {
  cerr << "Usage: RemoteCheckAccess args" << endl
       << "  -F arg          - akenti client config file" << endl
       << "  -r arg          - resource" << endl
       << "  [-u arg -c arg] - user DN and user's CADN" << endl 
       << "  [-x arg]        - user's X509 chain" << endl
       << "  [-q]            - concise output" << endl
       << "  [-v]            - verbose" << endl
       << "  [-h]            - prints this message" << endl;
}

int 
main(int argc, char** argv) {
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  string confFileName;
  string userDN;
  string caDN;
  string x509CertFileName;
  string resource;
  bool verbose = false;
  bool concise = false;
  int opt;

  while ((opt = getopt(argc,argv,"u:c:x:F:r:qvh")) != EOF) {
    switch(opt){
      case 'u':
        userDN = optarg;
        break;
      case 'c':
        caDN = optarg;
        break;
      case 'x':
        x509CertFileName = optarg;
        break;
      case 'F':
        confFileName = optarg;
        break;
      case 'r':
        resource = optarg;
        break;
      case 'q':
        concise=true;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
    }
  }

  if ((!userDN.length() || !caDN.length()) && (!x509CertFileName.length())) {
    cerr << "Must provide either a userDN and userCaDN or x509CertFileName" << endl;
    exit(2);
  } else if (!resource.length()) {
    cerr << "Must provide a resource" << endl;
    exit(2);
  }

  try {
    if (verbose) {
      cout  << "Initiizing XML library ... " << endl;
    }

    XMLUtil::init();

    if (verbose) {
      cout  << "Installing security provider ..." << endl;
    }
    
    Provider po = Provider(new OpenSSLProviderImpl);
 
    po->loadSignatureAlgorithms();
    po->seedRandomGenerator("WE GOT TO SEED THE RANDOM GENERATOR");
    Provider::setProvider(po);

    if (verbose) {
      cout  << "Loading client configuration file ..." << endl;
    }

    Properties& props = Properties::getSystemProperties();
 
    props.load(confFileName);

    if (StringUtil::equalsIgnoreCase(props.getValue("SSLMode"), "on")) {
      if (verbose) {
        cout  << "Setting SSL context ..." << endl;
      }

      SSLContext ctx = po->createSSLContext(SSLConstants::TLS1, true);
 
      ctx->setVerify(SSLConstants::VERIFY_PEER);
      ctx->setPrivateKeyFile(props.getValue("ClientKeyFile"), SSLConstants::PEM);
      ctx->setCertificateFile(props.getValue("ClientCertFile"), SSLConstants::PEM);
 
      string path = props.getValue("CertificateAuthorityPath");
 
      if (path.length() != 0) {
        path += "/";
      }
 
      path += props.getValue("CertificateAuthorityFile");
      ctx->setTrustedCAS(path);
      SSLContext::setDefaultSSLClientContext(ctx);
    }

    AkentiPrincipal user;

    if (x509CertFileName != "") {
      if (verbose) {
        cout << "Loading user's cert chain from " << x509CertFileName << endl;
      }

      CertificateFactory certFac = CertificateFactory::getInstance();
      CertificateChain chain = certFac->initCertificateChainFromFile(
        "PEM", x509CertFileName
      );

      Certificate cert = chain->getSubjectCertificate();

      user = AkentiPrincipal(cert->getPrincipal(), cert->getGuarantor());
    } else {
      user = AkentiPrincipal(userDN, caDN);
    }

    CheckAccessRequest req;
    CheckAccessResponse resp;

    req.setID(1);
    req.setResource(resource);
    req.setAkentiPrincipal(user);
    ProtocolHandler::getInstance(ProtocolHandler::CLIENT).handle(req, resp);

    AkentiMessage mesg = resp.getAkentiMessage();
    CapabilityCertificate capCert = resp.getCapabilityCertificate();

    cout << resp.getAkentiMessage().toString() << endl;

    if (!capCert) {
      exit(1);
    }

    if (!concise) {
      AkentiCertificate::writeToStream("XML", cout, capCert);
    }

    exit(0);
  } catch(Exception& ex) {
    cerr << "Exception:" << ex << endl;
    exit(2);
  } catch(Error& err) {
    cerr << "ERROR:(Report to akenti team)" << err << endl;
    exit(3);
  } 
}

}
