#include <sstream>
#include "CertificateFactory.hpp"
#include "AkentiX509CertificateImpl.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "PrivateKey.hpp"
#include "PublicKey.hpp"
#include "Util.hpp"
#include "Error.hpp"

namespace Akenti {

typedef AkentiX509CertificateImpl AXCI;

AXCI::AkentiX509CertificateImpl() 
  : AkentiCertificateImpl("X509"),
    signable("NO_SIGNABLE"),
    principal(),
    serialNumber(-1),
    revocationURL(),
    subjectPublicKey(),
    cert() {
}

AXCI::AkentiX509CertificateImpl(const Certificate& cert) 
  : AkentiCertificateImpl("X509"), cert(cert) {
  init();
}

void
AXCI::init() {
  // XXXXXXXXX
  version = cert->getVersionNumber();

  std::ostringstream os;

  os << cert->hashCode();

  id = os.str();

  DistinguishedName ca = cert->getGuarantor();

  principal = AkentiPrincipal(cert->getPrincipal(), ca);

  issuer = AkentiPrincipal(ca, DistinguishedName());
  notBefore = cert->getNotBefore();
  notAfter = cert->getNotAfter();
  subjectPublicKey = cert->getSubjectPublicKey();
  signable = cert->getSignable();
  signatureAlg = cert->getSignatureAlgorithm();
  signature = cert->getSignature();
  serialNumber = (int) cert->getSerialNumber();
  // XXX
  //  revocationURL = URL(cert->getRevocationURL());
}

AXCI::~AkentiX509CertificateImpl() {
}

AkentiX509CertificateImpl*
AXCI::newInstance() const {
  return new AkentiX509CertificateImpl;
}

string
AXCI::getHashString() const {
  return cert->getHashString();
}

const AkentiPrincipal&
AXCI::getPrincipal() const {
  return principal;
}

const PublicKey&
AXCI::getSubjectPublicKey() const {
  return subjectPublicKey;
}

const URL& 
AXCI::getRevocationURL() const {
  return revocationURL;
}

string
AXCI::getSignable() const {
  return signable;
}

int 
AXCI::getSerialNumber() const {
  return serialNumber;
}

void 
AXCI::sign(const string& signatureAlg, const PrivateKey &privateKey) {
  throw Error("AkentiX509CertificateImpl:sign:invalid operation");
}

void 
AXCI::writeObject(AkentiOutputStream& drain) const {
  CertificateFactory fac = CertificateFactory::getInstance();

  drain << fac->getCertificateEncoded("DER", cert);
}
 
void
AXCI::readObject(AkentiInputStream& source) {
  version = serialNumber = -1;
  id = "NO_ID";
  principal = AkentiPrincipal();
  issuer = AkentiIssuer();
  notBefore = notAfter = UtcTime(); 
  signatureAlg = "NO_SIG_ALG";
  signature = "NO_SIG";
  revocationURL = URL();
  subjectPublicKey = NULL;
  cert = NULL;

  string encoding;

  source >> encoding;

  CertificateFactory fac = CertificateFactory::getInstance();
  
  cert = fac->initCertificate("DER", encoding);
  init();
}

void 
AXCI::doParamString(ostream& os) const {
  os << "principal" << ':' <<  getPrincipal().paramString()
     << ','
     << "serialNumber" << ':' <<  getSerialNumber()
     << ','
     << "revocationURL" << ':' <<  getRevocationURL()
     << ','
     << "publicKey" << ':' << getSubjectPublicKey()->paramString();
}

void 
AXCI::doToText(int indentAmt, ostream& os) const {
  // XXX
  // to be implemented
}

void 
AXCI::doToLogString(ostream& os) const {
  os << getPrincipal().toLogString();
}

}
