#ifndef POLICY_CONTEXT_HPP
#define POLICY_CONTEXT_HPP

#include <string>
#include <vector>
#include <set>
#include <map>

#include "ref_ptr.hpp"
#include "AkentiPrincipal.hpp"
#include "CapabilityCertificate.hpp"
#include "AttributeCertificate.hpp"
#include "AkentiX509Certificate.hpp"
#include "ConditionalActions.hpp"
#include "Policy.hpp"

using std::string;
using std::vector;
using std::set;
using std::multimap;

using namespace Akenti;

namespace Akenti {

/**
  * The policy context contains all the relevent information 
  * that was used by the engine to satisfy a request.
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */
class PolicyContextImpl {

public:

  /**
    * Destroys this object. 
    */
  ~PolicyContextImpl();

  /**
    * Returns the name of the resource being accesed.
    */
  const string& getResource() const;

  /**
    * Returns the friendly resource name.
    *
    */
  const string& getResourceAltName() const;

  /**
    * Returns the dn/caDN of the user.
    */
  const AkentiPrincipal& getAkentiPrincipal() const;

  /**
    * Returns the user's x509Certificate.
    *
    * @see AkentiX509Certificate
    */
  const AkentiX509Certificate& getSubjectCert() const;
   
  /**
    * Returns the user's capability.
    *
    * @see CapabilityCertificate 
    */
  const CapabilityCertificate& getCapabilityCert() const;

  /**
    * Returns a vector of all policies that apply to this resource.
    *
    * @see PolicyImpl
    */
  const vector<Policy>& getPolicies() const;

  /**
    * Returns a vector of all users' attribute certificates
    * to evaluate the Use Conditions.
    * 
    * @see AttributeCertificate
    */
  const vector<AttributeCertificate>& getAttributeCerts() const;

  /**
    * Returns the desired attributes. 
    * 
    * <b>Note:</b> empty map implies all attributes are desired.
    *
    */
  const multimap<string, string>& getRestrictedAttributes() const;

  /**
    * Returns true if attribute and value is desired to be used.
    * 
    * <b>Note:</b> empty ==> all attribute/values are desired.
    */
  bool isAttributeDesired(const string& attribute, const string& value) const;

  /**
    * Returns the desired actions. 
    * 
    * <b>Note</b>: empty set implies all actions are desired.
    */
  const set<string>& getDesiredActions() const;

  /**
    * Returns true if action is desired.
    * 
    * <b>Note:</b> empty ==> all actions are desired.
    */
  bool isActionDesired(const string& action) const;

  /**
    * Returns the unconditional actions.
    */
  const vector<string>& getUnconditionalActions() const;

  /**
    * Returns the conditional actions.
    *
    * @see ConditionalActions
    */
  const vector<ConditionalActions>& getConditionalActionsList() const;
  
  /**
    * For debugging.
    */  
  string paramString() const;

private:

  friend class AKRuntime;
 
  /**
    */
  PolicyContextImpl();

  /**
    */
  void setResource(const string& resource);

  /**
    */
  void setResourceAltName(const string& resAltName);

  /**
    */
  void setAkentiPrincipal(const AkentiPrincipal& ap);

  /**
    */
  void setSubjectCert(const AkentiX509Certificate& cert);

  /**
    */
  void addPolicy(const Policy& policy);

  /**
    */
  void addAkentiX509Cert(const AkentiX509Certificate& cert); 

  /**
    */
  void addAttributeCert(const AttributeCertificate& attrCert);

  /**
    */
  void addUnconditionalAction(const string& unconditionalAction);

  /**
    */
  void addConditionalActions(const ConditionalActions& conditionalActions);

  /**
    */
  void setCapabilityCert(const CapabilityCertificate& cert);

  /**
    */
  void setDesiredActions(const set<string>& desiredActions);

  /**
    */
  void setRestrictedAttributes(const multimap<string, string>& restrictedAttributes);

  string resource;

  string resAltName;

  AkentiPrincipal ap;

  AkentiX509Certificate subjectCert;

  vector<AkentiX509Certificate> x509Certs;

  vector<AttributeCertificate> attrCerts;

  CapabilityCertificate capCert;

  vector<Policy> policies;
  
  vector<string> unconditionalActions;

  vector<ConditionalActions> conditionalActionsList;

  set<string> desiredActions;

  multimap<string, string> restrictedAttributes;
  
};

/**
  * Safe wrapper around code>PolicyContextImpl</code>.
  *
  * @see ref_ptr
  * @see PolicyContextImpl
  */ 
typedef ref_ptr<PolicyContextImpl> PolicyContext; 

} // end of name space

#endif
