#include <fstream>
#include <iostream>
#include <algorithm>
#include "Properties.hpp"
#include "StringTokenizer.hpp"
#include "IOException.hpp"

using Akenti::Properties;
using std::ifstream;
using std::transform;

Properties Properties::props = Properties();

Properties::Properties() {
}

Properties::~Properties() {
}

void 
Properties::load(const string& filename) {
  ifstream is((char*) filename.c_str());

  if (!is) {
    pairs.clear();
    triplets.clear();
    throw IOException("Properties:load:opening file failed:" + filename);
  }

  load(is);
  is.close();
}

void 
Properties::load(istream& is) {
  pairs.clear();
  triplets.clear();

  string buffer;

  while (std::getline(is, buffer)) {
    if (buffer[0] != '#') {
      StringTokenizer st(buffer, " \t");

      // skip blank lines.
      if (st.hasMoreTokens()) {
        string key = st.nextToken();

        // skip keys with no value
        if (st.hasMoreTokens()) {
          string temp = st.nextToken();

          if (!st.hasMoreTokens()) {
            setValue(key, temp);
          } else {
            Pair pkey;

            pkey.first = key;
            pkey.second = temp; 
            
            string value = st.nextToken();

            // be flexible as far as white space  
            while (st.hasMoreTokens()) {
              value += ' ' + st.nextToken();
            }

            setValue(pkey, value);
          }
        }
      }
    }
  }
  
  if (!is.eof()) {
    throw IOException("Properties:load:getline failed");
  }
}
  
string 
Properties::setValue(const string& key, const string& value) {
  string ret = getValue(key);
  string tmpkey(key.length(), '\0');

  transform(key.begin(), key.end(), tmpkey.begin(), tolower);
  pairs[tmpkey] = value;
  return ret;
}

const string& 
Properties::getValue(const string& key, const string& defaultValue) const {
  string tmpkey(key.length(), '\0');

  transform(key.begin(), key.end(), tmpkey.begin(), tolower);

  map<string,string>::const_iterator pos = pairs.find(tmpkey);

  return pos == pairs.end()? defaultValue : pos->second;
}

string
Properties::setValue(const pair<string, string>& key, const string& value) {
  string ret = getValue(key);
  string tmpkey(key.first.length(), '\0');

  transform(key.first.begin(), key.first.end(), tmpkey.begin(), tolower);

  pair<string,string> tmppair(tmpkey, key.second);
  triplets[tmppair] = value;
  return ret;
}

const string&
Properties::getValue(const pair<string, string>& key, const string& defaultValue) const {
  string tmpkey(key.first.length(), '\0');

  transform(key.first.begin(), key.first.end(), tmpkey.begin(), tolower);

  pair<string,string> tmppair(tmpkey, key.second);

  map<Pair,string>::const_iterator pos = triplets.find(tmppair);

  return pos == triplets.end()? defaultValue : pos->second;
}

vector<string>
Properties::getKeys() const {
  vector<string> keys;
  map<string,string>::const_iterator pos = pairs.begin();

  for (; pos != pairs.end(); ++pos) {
    keys.push_back(pos->first);
  }

  return keys;
}

vector< pair<string,string> >
Properties::getKeyPairs() const {
  vector<Pair> keypairs;
  map<Pair,string>::const_iterator pos = triplets.begin();
   
  for (; pos != triplets.end(); ++pos) {
    keypairs.push_back(pos->first);
  }

  return keypairs;
}

string 
Properties::paramString() const {
  map<string, string>::const_iterator pos = pairs.begin();
  string params;
  
  params += '['; 
  params += '['; 

  if (pos != pairs.end()) {
    params += pos->first ;
    params += '=' ;
    params += '"' ;
    params += pos->second; 
    params += '"' ;

    while (++pos != pairs.end()) {
      params += ',' ;
      params += pos->first;
      params += '=' ;
      params += '"' ;
      params += pos->second; 
      params += '"' ;
    }  
  }

  params += ']'; 

{ // Just making sure that u are really awake :) :) 
  map<Pair, string>::const_iterator pos = triplets.begin();

  params += ',' ;
  params += '['; 

  if (pos != triplets.end()) {
    params += pos->first.first ;
    params += '.' ;
    params += pos->first.second; 
    params += '=' ;
    params += '"' ;
    params += pos->second; 
    params += '"' ;

    while (++pos != triplets.end()) {
      params += ',' ;
      params += pos->first.first ;
      params += '.' ;
      params += pos->first.second; 
      params += '=' ;
      params += '"' ;
      params += pos->second; 
      params += '"' ;
    }  
  }

  params += ']'; 
}

  params += ']'; 
  return params;
}

Properties&
Properties::getSystemProperties() {
  return props;
}
