#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h> 
#include "File.hpp" 
#include "Filter.hpp" 
#include "IOException.hpp" 
#include "Error.hpp" 

using Akenti::File;

extern "C" {
  int mkdir(const char*, mode_t);
  int unlink(const char*);
  int rmdir(const char*);
} 

File::File(const string& filename) : filename(filename) {
}

string
File::getName() const {
  if (filename == "/") {
    return filename;
  }

  string::size_type index = filename.rfind('/');

  if (index == string::npos) {
    return filename;
  }

  return filename.substr(index + 1);
}

string 
File::getParent() const {
  if (filename == "/") {
    throw Error("File:getParent:file is root, has no parent");
  }

  string::size_type index = filename.rfind('/');

  if (index == string::npos) {
    throw Error("File:getParent:relative name, can't compute parent");
  } else if (!index) {
    return "/";
  } else {
    return filename.substr(0, index);
  }
}

bool 
File::hasParent() const {
  return !(filename == "/" || filename.rfind('/') == string::npos);
}

string 
File::getAbsolutePath() const {
  if (isAbsolute()) {
    return filename;
  }

  char cwd[256];

  if (getcwd(cwd, 255) == NULL) {
    throw Error("File:getAbsolutePath:coold not get working directory");
  }
   
  string ret(cwd);

  ret += '/';
  return ret += filename;
}

bool 
File::isAbsolute() const {
  return filename.length() && filename[0] == '/';
}

bool 
File::exists() const {
  struct stat buf;

  return stat(filename.c_str(), &buf) != -1;
}

bool 
File::isDirectory() const {
  struct stat buf;
  
  return stat(filename.c_str(), &buf) != -1 && S_ISDIR(buf.st_mode);
}

bool 
File::isFile() const {
  // TODO
  return exists() && !isDirectory();
}

vector<string> 
File::list() const {
  static Filter filter;

  return list(filter);
}

vector<string> 
File::list(const Filter& filter) const {
  DIR* dir = opendir(filename.c_str());

  if (!dir) {
    if (!exists()) {
      throw Error("File:list:file does not exists");
    } else if (!isDirectory()) {
      throw Error("File:list:file is not a directory");
    } else {
      throw IOException("File:list:unable to open directory");
    }
  }

  // skip "." and ".." directories
  readdir(dir);
  readdir(dir);

  struct dirent *entry;
  vector<string> files;

  while (entry = readdir(dir)) {
    string name(entry->d_name);

    if (filter.accept(name)) {
      files.push_back(name);
    }
  }

  closedir(dir);
  return files;
}

void
File::mkdir() const {
  if (::mkdir(filename.c_str(), 0755) == -1) {
    throw IOException("File:mkdir:mkdir failed");
  }
}

bool 
File::remove() const {
  return (
    isDirectory() ? rmdir(filename.c_str()) : unlink(filename.c_str())
  ) == 0;
}

string
File::paramString() const {
  string params;

  params += '[';
  params += "filename";
  params += '=';
  params += '"';
  params += filename;
  params += '"';
  params += ']';
  return params;
}

string
File::toLogString() const {
  string log;
  
  log += '"';  
  log += "FILE_NAME=";
  log += '"';
  log += filename;
  log += '"';
  log += '"';
  return log;
}
