#ifndef AK_SERVER_SOCKET_HPP
#define AK_SERVER_SOCKET_HPP

#include "Socket.hpp"
#include "SocketImplFactory.hpp"

/**
  * The <i>ServerSocket</i> provides the services needed for socket for 
  * the server.
  * 
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class ServerSocket {

public:

  /**
    * Creates this object 
    *
    * @param impl the SocketImpl object
    */
  ServerSocket(SocketImpl* impl = NULL) : impl(impl) {
  }

  /**
    * Prepares the server's socket and binds it to port
    * 
    * @param port the port number for the server
    * @see SocketImplFactory#getSocketImplFactory()
    * @see SocketImplFactory#createSocketImpl()
    */
  ServerSocket(int port) { 
    const SocketImplFactory& 
      fac = SocketImplFactory::getSocketImplFactory();

    impl = fac.createSocketImpl();
    impl->setPort(port);
  }

  /**
    * Destroys this object
    */   
  virtual ~ServerSocket() {
  }

  /**
    * Returns the host name for the server
    * 
    * @return the name of the host
    * @see SocketImpl#getHost()
    */
  const string& getHost() const {
    return impl->getHost();
  }

  /**
    * Returns the port number for the server
    * 
    * @return the port number 
    * @see SocketImpl#getPort() 
    */
  int getPort() const {
    return impl->getPort();
  }

  /**
    * Does the binding
    *
    * @see SocketImpl#doBind()
    */ 
  void bind()  {
    impl->doBind();
  }

  /**
    * Does the listen 
    * 
    * @param backlog the number of listener/backlog for this server
    * @see SocketImpl#doListen(int backlog)
    */
  void listen(int backlog)  {
    impl->doListen(backlog);
  }

  /**
    * Does the accept
    * 
    * @see SocketImpl#newInstance()
    * @see Socket#ret(SocketImpl):
    */
  Socket accept() {
    SocketImpl* newImpl = impl->newInstance();
    Socket ret(newImpl);  

    impl->doAccept(newImpl);
    return ret;
  }

  /**
    * Closes the socket
    *
    * @see SocketImpl#doClose()
    */
  void close() {
    impl->doClose();
  }

  /**
    * Returns a string representation for ServerSocket, which is used  
    * debugging purposes.
    *
    * @return debugging info
    */
  virtual string paramString() const {
    return impl->paramString(); 
  }
    
  /**
    */
  template<class T> bool instanceOf() {
     return dynamic_cast<typename T::ImplType> (impl.get());
  }

protected:

  typedef SocketImpl* ImplType;

  ref_ptr<SocketImpl> impl;

};

} // end of namespace Akenti

#endif

