#ifndef AK_SOCKET_HPP
#define AK_SOCKET_HPP

#include "ref_ptr.hpp"
#include "SocketImpl.hpp"
#include "SocketImplFactory.hpp"

/**
  * A <i>Socket<i> contains a reference count that wraps a <i>SocketImpl</i> 
  * object. It is used to indirectly access the SockImpl methods.
  *
  * @see ref_ptr
  * @see SocketImpl
  *
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class Socket {

public:

  /**
    * Creates the socket given the host and the port.
    *
    * @param host name of the host
    * @param port the port number
    */
  Socket(const string& host, int port) { 
    const SocketImplFactory& 
      fac = SocketImplFactory::getSocketImplFactory();

    impl = fac.createSocketImpl();
    impl->setHost(host);
    impl->setPort(port);
  }
  
  /**
    * Creates the socket given the SocketImpl.
    *
    * @param impl the socket implementation object
    */
  Socket(SocketImpl* impl = NULL) : impl(impl) {
  }

  /**
    * Decrements the ref_cnt by 1 and when it goes to 0 destroys this object.
    * 
    */
  virtual ~Socket() {
  }

  /**
    * Returns the host of the remote server.
    *
    * @return host string
    * @see SocketImpl#getHost()
    */
  const string& getHost() const {
    return impl->getHost();
  }

  /**
    * Returns the port number of the remote server.
    *
    * @return port int 
    * @see SocketImpl#getPort()
    */
  int getPort() const {
    return impl->getPort();
  }

  /**
    * Returns an <i>InputStream</i> associated with this <i>Socket</i>. 
    *
    * @see SocketImpl#getInputStream()
    * @see InputStream
    */
  InputStream& getInputStream() const {
    return impl->getInputStream();
  }

  /**
    * Returns an <i>OutputStream</i> associated with this <i>Socket</i>. 
    *
    * @see SocketImpl#getOutputStream()
    * @see OutputStream
    */
  OutputStream& getOutputStream() const {
    return impl->getOutputStream();
  }

  /**
    * Opens a connection to the host.
    *
    * @see SocketImpl#doConnect()
    */
  void connect() {
    impl->doConnect();
  }

  /**
    * Closes the connection and the streams associated with the socket.
    *
    * @see SocketImpl#doClose
    */
  void close() {
    impl->doClose();
  }
 
  /**
    * Returns a string representation for Socket which is used for  
    * debugging purposes.
    *
    * @return debugging info
    */
  string paramString() const {
    return impl->paramString();
  }   

  /**
    */
  template<class T> bool instanceOf() {
     return dynamic_cast<typename T::ImplType> (impl.get());
  }

protected:

  typedef SocketImpl* ImplType;

  ref_ptr<SocketImpl> impl;

};

} // end of namespace Akenti

#endif

