#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sstream>
#include "SocketImpl.hpp"
#include "SocketInputStream.hpp"
#include "SocketOutputStream.hpp"
#include "SocketInBuffer.hpp"
#include "SocketOutBuffer.hpp"
#include "IOException.hpp"
#include "Error.hpp"

using std::ostringstream;

typedef Akenti::SocketImpl SI;

SI::SocketImpl() : host(""), port(-1) {
}
  
Akenti::SocketImpl*
SI::newInstance() {
  return new SocketImpl;
}

SI::~SocketImpl() {
}

int
SI::openConnection() {
  int sock_fd;

  if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    throw IOException("SocketImpl:initLocalPort:can't open stream socket");
  }

  struct hostent *hp = gethostbyname(host.c_str());

  if (!hp) {
    throw IOException("SocketImpl:connect:unkown host:" + host);
  }

  struct sockaddr_in address;

  address.sin_family = AF_INET;
  address.sin_port = htons((unsigned short) getPort());
  memcpy(&address.sin_addr, *hp->h_addr_list, hp->h_length);

  if (connect(sock_fd, (struct sockaddr *) &address, sizeof(address))  < 0) {
    ostringstream oss;

    oss << port;
    throw IOException("Can't connect to " + host + " at " + oss.str());
  }

  return sock_fd;
}

void 
SI::doConnect() {
  fd = new IOHandlerImpl(openConnection());
}

int 
SI::bindToAddress() {
  int sock_fd;

  if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    throw IOException("SocketImpl:initLocalPort:can't open stream socket");
  }

  int on = 1;

  if (setsockopt(sock_fd, SOL_SOCKET, SO_REUSEADDR, (char*) &on, sizeof(on)) < 0) {
    throw IOException("SocketImpl:initLocalPort:setsockop failed");
  }

  struct sockaddr_in address;

  address.sin_family = AF_INET;
  address.sin_port = htons(getPort());
  address.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(sock_fd, (struct sockaddr*) &address, sizeof(address)) < 0) {
    throw IOException("SocketImpl:doBind:can't bind local address");
  }
  
  return sock_fd;
}

void 
SI::doBind() {
  fd = new IOHandlerImpl(bindToAddress());
}

void
SI::doListen(int backlog) {
  int sd = fd->getDescriptor();

  if (sd < 0) {
    throw Error("SocketImpl:doListen:invalid soperation");
  }

  if (listen(sd, backlog) < 0) {
    throw IOException("SocketImpl:doListen:listening failed");
  }
}

int
SI::acceptConnection(SocketImpl* impl) {
  if (!impl) {
    throw Error("SocketImpl:acceptConnection:invalid arg");
  }

  struct sockaddr_in address;
  socklen_t len = sizeof(address);

  int sd  = accept(
    fd->getDescriptor(), 
    (struct sockaddr*) &address,
    &len 
  );
  
  if (sd < 0) {
    throw IOException("SocketImpl:acceptConnection:new connection failed");
  }

  struct hostent* hp = gethostbyaddr(
     (char *)&address.sin_addr, sizeof(struct in_addr), AF_INET
  );

  // hosts using dhcp do not have hostnames assigned
  if (!hp) {
    impl->host = "unknown host";
  } else {
    impl->host = hp->h_name; 
  }
  impl->port = ntohs(address.sin_port); 

  return sd;
}

void
SI::doAccept(SocketImpl* impl) {
  impl->fd = new IOHandlerImpl(acceptConnection(impl));
}

void 
SI::doClose() {
  if (fd->isClosed()) {
    throw Error("SocketImpl:doClose:invalid operation");
  }

  if (fd->close() == -1) {
    throw IOException("SocketImpl:doClose:error closing socket");
  }
  
  is = NULL;
  os = NULL;
}

Akenti::InputStream&
SI::getInputStream() {
  if (!fd) {
    throw Error("SocketImpl:getInputStream:invalid operation");
  } 

  if (!is) {
    is = doGetInputStream();
  }

  if (!is->good()) {
    throw IOException("SocketImpl:getInputStream:bad input stream");
  }

  return is;
}

istream*
SI::doGetInputStream() {
  return new SocketInputStream(new SocketInBuffer(fd));
}

Akenti::OutputStream&
SI::getOutputStream() {
  if (!fd) {
    throw Error("SocketImpl:getOutputStream:invalid operation");
  } 

  if (!os) {
    os = doGetOutputStream();
  }

  if (!os->good()) {
    throw IOException("SocketImpl:getOutputStream:bad output stream");
  }

  return os;
}

string 
SI::paramString() const {
  ostringstream oss;

  oss << '['
     << "host"
     << '='
     << '"'
     << getHost() 
     << '"'
     << ','
     << "port" 
     << '='
     << getPort()
     << ','
     << "fd"
     << '='
     << fd->paramString()
     << ']';

  return oss.str();
}

ostream*
SI::doGetOutputStream() {
  return new SocketOutputStream(new SocketOutBuffer(fd));
}

int 
SI::getPort() const {
  return port;
}

const string& 
SI::getHost() const {
  return host;
}

void 
SI::setHost(const string& host_arg) {
  this->host = host_arg;
}

void 
SI::setPort(int port_arg) {
  if (port_arg < 0) {
    throw Error("SocketImpl:setPort:port < 0");
  }

  this->port = port_arg;
}

