#include <sstream>

#include "URL.hpp"
#include "StringTokenizer.hpp"
#include "URLHandlerFactory.hpp"
#include "URLConnection.hpp"
#include "FileURLHandlerImpl.hpp"
#include "HttpURLHandlerImpl.hpp"
#include "LdapURLHandlerImpl.hpp"
#include "URLHandler.hpp"
#include "Error.hpp"
#include "MalformedURLException.hpp"

using std::ostream;
using std::ostringstream;

namespace Akenti {

class NullURLHandlerImpl : public URLHandlerImpl {

public:

  URLConnection 
  getURLConnection(const URL& url) const {
    throw Error("uninitialized url");
  }
}* nullURLHandlerImpl = new NullURLHandlerImpl;

static URLHandler NULL_HANDLER = nullURLHandlerImpl; 
static map<string, URLHandler> DEFAULT_HANDLERS;

static class URLSInitializer {
  
public:

  URLSInitializer() {
    DEFAULT_HANDLERS["noprot"] = NULL_HANDLER; 
    DEFAULT_HANDLERS["file"]   = new FileURLHandlerImpl;
    DEFAULT_HANDLERS["http"]   = new HttpURLHandlerImpl;
    DEFAULT_HANDLERS["ldap"]   = new LdapURLHandlerImpl;
  }

} urlInitializer = URLSInitializer();

}

using Akenti::URL;

Akenti::URLHandler& 
URL::getURLHandler(const string& protocol) {
  const URLHandlerFactory& fac = URLHandlerFactory::getURLHandlerFactory();
  URLHandler& handler = fac->getURLHandler(protocol);

  if (!handler) {
    handler = DEFAULT_HANDLERS[protocol]; 

    if (!handler) {
      throw MalformedURLException("getURLHandler:No URL Handler" + protocol);
    }
  }
  
  return handler;
}


URL::URL() : 
  protocol("noprot"), 
  host("nohost"), 
  port(-1), 
  file("/nofile") {
}

URL::URL(const string& url) 
  : host(""), port(-1), file("/") {
  StringTokenizer tokenizer(url, ":", true);

  if (!tokenizer.hasMoreTokens()) {
    throw MalformedURLException("URL:URL:no protocol:" + url);
  }

  protocol = tokenizer.nextToken();
   
  if (!tokenizer.hasMoreTokens() || tokenizer.nextToken() != ":") {
    throw MalformedURLException("URL:URL:bad url:" + url);
  }

  URLHandler& handler = getURLHandler(protocol);

  if (tokenizer.hasMoreTokens()) {
    handler->parseURL(*this, tokenizer.nextToken(""));
  } else {
    handler->parseURL(*this, "");
  }
}

URL::URL(const string& protocol_arg,
         const string& host_arg,
         int port_arg,
         const string& file_arg)
  : protocol(protocol_arg), port(-1), host(""), file("/") {
  URLHandler& handler = getURLHandler(protocol_arg);

  handler->initURL(*this, host_arg, port_arg, file_arg);
}

URL::~URL() {
}

const string&
URL::getProtocol() const {
  return protocol;
}

void
URL::setHost(const string& host_arg) {
  this->host = host_arg;
}

const string&
URL::getHost() const {
  return host;
}

void
URL::setPort(int port_arg) {
  this->port = port_arg;
}

int
URL::getPort() const {
  return port;
}

void
URL::setFile(const string& file_arg) {
  this->file = file_arg;
}

const string&
URL::getFile() const {
  return file;
}

string
URL::toString() const {
  URLHandler& handler = getURLHandler(protocol);

  return handler->toExternalForm(*this);
}

bool
URL::operator==(const URL& url) const {
  return toString() == url.toString();
}

ostream& 
operator<<(ostream& os, const Akenti::URL& url) {
  return os << url.toString();
}

Akenti::URLConnection 
URL::getURLConnection() const {
  URLHandler& handler = getURLHandler(protocol);

  return handler->getURLConnection(*this);
}

URL::operator void*() const {
  URLHandler& handler = getURLHandler(protocol);

  return handler == NULL_HANDLER? NULL : (void *) this;
}

string
URL::paramString() const {
  ostringstream os;
 
  os << '[' 
     << "protocol"
     << '='
     << '"' 
     << getProtocol()
     << '"' 
     << ',' 
     << "host"
     << '='
     << '"' 
     << getHost()
     << '"' 
     << ',' 
     << "port"
     << '='
     << getPort()
     << ',' 
     << "file"
     << '='
     << '"' 
     << getFile()
     << '"' 
     << ']';
  return os.str();
}
