#include <openssl/x509.h>
#include <openssl/asn1.h>
#include <sstream>
#include "OpenSSLCertificateGeneratorImpl.hpp" 
#include "Certificate.hpp"
#include "CertificateImpl.hpp"
#include "PrivateKey.hpp"
#include "UtcTime.hpp"
#include "SecurityException.hpp"
#include "OpenSSLSignatureImpl.hpp"
#include "OpenSSLPrivateKeyImpl.hpp" 
#include "OpenSSLPublicKeyImpl.hpp"
#include "OpenSSLX509CertificateImpl.hpp"
#include "OpenSSLUtil.hpp"

using std::ostringstream;

typedef Akenti::OpenSSLCertificateGeneratorImpl OSCGI;

OSCGI::OpenSSLCertificateGeneratorImpl() {
}

OSCGI::~OpenSSLCertificateGeneratorImpl() {
}
  
void 
OSCGI::setSubject(const DistinguishedName& subj) {
  this->subject = subj;
}

void 
OSCGI::setSubjectPublicKey(const PublicKey& subjKey) {
  this->subjectKey = subjKey;
}

void 
OSCGI::setIssuer(const DistinguishedName& issuer_arg) {
  this->issuer = issuer_arg;
}

void 
OSCGI::setValidity(int days_arg) {
  this->days = days_arg;
}

void 
OSCGI::setSerialNumber(long sn) {
  this->serialNumber = sn;
}

void 
OSCGI::setVersionNumber(long vn) {
  this->versionNumber = vn;
}

Akenti::Certificate
OSCGI::generate(const string& sigalg, const PrivateKey& privateKey) {
  X509* delegate = X509_new();
  Certificate cert(new OpenSSLX509CertificateImpl(delegate));
  
  X509_set_version(delegate, versionNumber);

  X509_NAME *name = X509_get_subject_name(delegate);

  OpenSSLUtil::populateX509Name(name, subject);
  
  OpenSSLPublicKeyImpl* 
    subjectKeyImpl = dynamic_cast<OpenSSLPublicKeyImpl*> (subjectKey.get());

  if (!subjectKeyImpl) {
    throw SecurityException(
      "OpenSSLCertificateGeneratorImpl:generate:unrecognized public key"
    );
  }

  X509_set_pubkey(delegate, subjectKeyImpl->getEVP_PKEY());
  name = X509_get_issuer_name(delegate);
  OpenSSLUtil::populateX509Name(name, issuer);
  X509_set_issuer_name(delegate, name);
  ASN1_INTEGER_set(X509_get_serialNumber(delegate), this->serialNumber);

  UtcTime temp;
  char buffer[13];

  temp.toCharArray(buffer);
  ASN1_UTCTIME_set_string(X509_get_notBefore(delegate), buffer);
  temp += days * 24 * 60 * 60;
  temp.toCharArray(buffer);
  ASN1_UTCTIME_set_string(X509_get_notAfter(delegate), buffer);

  OpenSSLPrivateKeyImpl* 
    issuerKeyImpl = dynamic_cast<OpenSSLPrivateKeyImpl*> (privateKey.get());

  if (!issuerKeyImpl) {
    throw SecurityException(
      "OpenSSLCertificateGeneratorImpl:generate:unrecognized private key"
    );
  }
  
  const EVP_MD *md = OpenSSLSignatureImpl::getEVP_MD(sigalg);
  
  if (X509_sign(delegate, issuerKeyImpl->getEVP_PKEY(), md) <= 0) {
    throw SecurityException(
      "OpenSSLCertificateFactoryImpl:generate: signature failed"
    );
  }
  
  return cert;
}

// XXX We need to add the integral fields

string 
OSCGI::paramString() const {
  ostringstream os;

  os << '['
     << "subject"
     << '='
     << '"'
     << subject.paramString()
     << '"'
     << ','
     << "subjectKey"
     << '='
     << '"'
     << subjectKey->paramString()
     << '"'
     << ','
     << "issuer"
     << '='
     << '"'
     << issuer.paramString()
     << '"'
     << ','
     << "days" 
     << '='
     << days
     << ','
     << "serialNumber"
     << '='
     << serialNumber
     << ','
     << "versionNumber" 
     << '='
     << versionNumber 
     << ']'; 
  return os.str();
}

