#ifndef OPEN_SSL_CERTIFICATE_REQUEST_GENERATOR_IMPL_HPP
#define OPEN_SSL_CERTIFICATE_REQUEST_GENERATOR_IMPL_HPP

#include <string>
#include "CertificateRequestGeneratorImpl.hpp"
#include "PublicKey.hpp"
#include "DistinguishedName.hpp"

using std::string;

namespace Akenti {

/**
  * This class is a subclass of <i>CertificateRequestGeneratorImpl</i>,
  * and is used to create an OpenSSLCertificateRequestImpl and the
  * CertfificateRequest that points to it. A certficateRequest holds all
  * the information needed to create an X509 identity certificate. It can be 
  * sent to a CA who generates and signs the identity certficate.
  * It holds the subject name and publicKey and the certificate versionNumber.
  * It generates and signs the OpenSSLCertificateRequestImpl with a
  * specified algorithm and privateKey. It will display the Certificate
  * as a  string. Once a CertificateRequestImpl has been generated and signed
  * it can no longer be modified.
  *
  * @see DistinguishedName
  * @see PublicKey
  * @see PrivateKey
  * @see CertificateRequest
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class OpenSSLCertificateRequestGeneratorImpl : public CertificateRequestGeneratorImpl {

public:

  /**
    * Construct this object.
    */
  OpenSSLCertificateRequestGeneratorImpl();

  /**
    * Deletes this object.
    */
  ~OpenSSLCertificateRequestGeneratorImpl();
  
  /**
    * Sets param subject
    *
    * @param subject DN of the subject of the certificate that is to be 
    *   requested
    */
  void 
  setSubject(const DistinguishedName& subject);

  /**
    * Sets param publicKey
    *
    * @param publicKey the public key of the subject of the certificate. The
    *   publicKey should refer to an OpenSSLPublicKeyImpl.
    */
  void 
  setPublicKey(const PublicKey& subjectKey);

  /**
    * Sets param versionNumber
    *
    * @param versionNumber
    */
  void 
  setVersionNumber(long versionNumber);

  /**
    * Generates CertificateRequest using specified algorithm
    * such as "RSA-SHA1" or "DSA-MD5" and privateKey
    *
    * @param algorithm Signature algorithm for signing the request
    * @param privateKey of the requestor. Should contain a reference to
    *   an OpenSSLPrivateKeyImpl.
    */
  CertificateRequest 
  generate(const string& algorithm, const PrivateKey& privateKey);

  /**
    * Display the CertificateRequest in string
    */
  virtual string paramString() const;

private:

  DistinguishedName subject;

  PublicKey subjectKey;

  long versionNumber;
};

} // end namespace Akenti

#endif
