#include <openssl/pem.h>
#include "OpenSSLCertificateRequestImpl.hpp"
#include "OpenSSLPublicKeyImpl.hpp"
#include "PublicKey.hpp"
#include "DistinguishedName.hpp"
#include "SecurityException.hpp"

typedef Akenti::OpenSSLCertificateRequestImpl OSCRI;

namespace Akenti {

static const string CLASSNAME = "OpenSSLCertificateRequestImpl:";

}

OSCRI::OpenSSLCertificateRequestImpl(X509_REQ *request) 
  : request(request) {
}

OSCRI::~OpenSSLCertificateRequestImpl() {
  X509_REQ_free(request);
}

Akenti::PublicKey 
OSCRI::getPublicKey() const {
  return PublicKey(
    new OpenSSLPublicKeyImpl(X509_REQ_get_pubkey(request))
  );
}

Akenti::DistinguishedName 
OSCRI::getSubject() const {
  string name(200, '\0');

  X509_NAME_oneline(
    X509_REQ_get_subject_name(request),
    (char *) name.data(),
    name.length()
  );

  return DistinguishedName(name.c_str());
}

long 
OSCRI::getVersionNumber() const {
  return X509_REQ_get_version(request);
}

string 
OSCRI::getSignatureAlgorithm() const {
  return string(OBJ_nid2sn(OBJ_obj2nid(request->sig_alg->algorithm)));
}

string 
OSCRI::getSignable() const {
  int len = i2d_X509_REQ_INFO(request->req_info, NULL);
  string signable(len, '\0');
  unsigned char *temp = (unsigned char *) signable.data();

  if (i2d_X509_REQ_INFO(request->req_info, &temp) != len) {
    throw SecurityException(CLASSNAME + "getSignable:failed");
  }

  return signable;
}

string 
OSCRI::getSignature() const {
  return string(
    (char*) request->signature->data, 
    request->signature->length
  );
}

bool
OSCRI::verify() const {
  PublicKey publicKey = getPublicKey();

  OpenSSLPublicKeyImpl* 
    publicKeyImpl = dynamic_cast<OpenSSLPublicKeyImpl*> (publicKey.get());
  
  return X509_REQ_verify(
    request, 
    publicKeyImpl->getEVP_PKEY()
  ) == 1;
}

string
OSCRI::getEncoded(const string& type) const {
  if (type == "DER") {
    int len = i2d_X509_REQ(request, NULL);
    
    if (len <= 0) {
      throw Error(CLASSNAME + "getEncoded:should never happen");
    }

    string encoding(len, '\0');
    unsigned char *temp = (unsigned char *) encoding.data();

    i2d_X509_REQ(request, &temp);
    return encoding;
  } else if (type == "PEM") { 
    BIO* bio = BIO_new(BIO_s_mem());
    
    PEM_write_bio_X509_REQ(bio, request);

    char *pem = NULL;
    long len = BIO_get_mem_data(bio, &pem);
    string encoding(pem, len); 

    // only bio needs to be freed :)
    BIO_free(bio);
    return encoding;
  }
  
  throw SecurityException(CLASSNAME + "getEncoded:unsupported type:" + type);
}

Akenti::OpenSSLCertificateRequestImpl*
OSCRI::init(const string& type, const string& encoding) {
  if (type == "DER") {
    unsigned char * temp = (unsigned char *) encoding.data();
    X509_REQ* delegate = d2i_X509_REQ(NULL, &temp, encoding.length());

    if (!delegate) {
      throw SecurityException(CLASSNAME + "init:bad DER encoding");
    }

    return new OpenSSLCertificateRequestImpl(delegate);
  } else if (type == "PEM") {
    BIO* bio = BIO_new_mem_buf((char*) encoding.data(), encoding.length());
     
    X509_REQ* delegate = PEM_read_bio_X509_REQ(bio, NULL, NULL, NULL);
    
    BIO_free(bio);

    if (!delegate) {
      throw SecurityException(CLASSNAME + "init:bad PEM encoding");
    }

    return new OpenSSLCertificateRequestImpl(delegate);
  }

  throw SecurityException(CLASSNAME + "getEncoded:unsupported type:" + type);
}

string 
OSCRI::paramString() const {
  string param;

  param = "[";
  param += "subject=";
  param += getSubject().paramString(); 
  param += ",";
  param += "publicKey=";
  param += getPublicKey()->paramString();
  param += ",";
  param += "SigAlg="; 
  param += ",";
  param += getSignatureAlgorithm();
  param += ",";
  param += (verify() ? "signature verified" : "signature did not verify");
  param += "]";
  return param;
}
