#ifndef OPEN_SSL_CERTIFICATE_REQUEST_IMPL_HPP
#define OPEN_SSL_CERTIFICATE_REQUEST_IMPL_HPP

#include <openssl/x509.h>
#include "CertificateRequestImpl.hpp"

namespace Akenti {

/**
  * This class is a subclass of <i>CertificateRequestImpl</i>
  * and is a wrapper for the openssl X509_REQ structure. 
  * A certficateRequest holds allthe information needed to create an X509
  * identity certificate. It can be sent to a CA who generates and signs
  * the identity certficate. It provides
  * get methods to extract specific fields from the X509_REQ, such as
  * subject, publicKey, versionNumber, signatureAlgorithm and signature.
  * The corresponding set methods are in the 
  * OpenSSLCertificateRequestGeneratorImpl class.
  *
  * @see CertificateRequest
  * @see CertificateRequestImpl  
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */ 

class OpenSSLCertificateRequestImpl : public CertificateRequestImpl {

public:

  /**
    * Constructs the object with specified type
    */
  OpenSSLCertificateRequestImpl(X509_REQ *request);

  /**
    * Deletes the object
    */
  ~OpenSSLCertificateRequestImpl();

  /**
    * Returns a <i>DistinguishedName</i> object which includes the 
    * <i>DistinguishName</i> of the Certificate to be requested.
    */ 
  DistinguishedName getSubject() const;

  /** 
    * Returns a <i>PublicKey</i> object which refers to an OpenSSLPrivateKey
    * object containing the publicKey of the subject of the request.
    * 
    * @see KeyPair
    * @see KeyPair#getPublicKey()
    * @see KeyFactory
    * @see KeyFactoryImpl#getKeyEncoded(const string&, const PublicKey&)
    * @see KeyFactoryImpl#initPublicKey(const string&, const string&, 
    *   const string&)
    * @see KeyFactoryImpl#generateRSAKeyPair(int, unsigned long)
    * @see KeyFactoryImpl#generateDSAKeyPair(int, const string&)
    */
  PublicKey getPublicKey() const;

  /**
    * Returns the versionNumber of this CertificateRequest
    */ 
  long getVersionNumber() const;

  /**
    * Returns the signatureAlgorithm such as "RSA-SHA1" or "DSA-MD5"
    * that this  CertificateRequest was signed by.
    */ 
  string getSignatureAlgorithm() const;
 
  /**
    * Returns the part of the certficateRequest that was signed.
    *
    * @exception SecurityException: getSignable:failed
    */
  string getSignable() const;

  /**
    * Returns the signature of this CertificateRequest
    */
  string getSignature() const;

  /**
    * Returns true if the signature of the request verifies.
    */
  bool verify() const;

  /**
    * Displays the CertificateRequest in string for debugging purposes.
    */
  string paramString() const;

  /**
    * Creates and initializes an OpenSSLCertificateRequestImpl from a
    * DER or PEM encoded string representation of a Request.
    *
    * @param string type encoding used in string
    * @param encoding the string representation of a certificateRequest.
    */
  static OpenSSLCertificateRequestImpl*
  init(const string& type, const string& encoding);
  
  /**
    * Returns a DER or PEM encoded version of the CertificateRequest that
    * can be sent to a CA for signing.
    *
    * @param string type the type of encoding wanted: DER or PEM
    * @exception Error: i2d_X509_REQ returned a failure - should not happen.
    * @exception SecurityException - if type is not DER or PEM
    */
  string
  getEncoded(const string& type) const;

private:

  X509_REQ *request;
};

} // end namespace Akenti

#endif

