#include <sstream> 
#include <openssl/err.h>

#include "OpenSSLSecureSocketImpl.hpp"
#include "OpenSSLContextImpl.hpp"
#include "OpenSSLX509CertificateImpl.hpp"
#include "SecurityException.hpp"
#include "Error.hpp"
 
using std::ostringstream;

#define DEBUG 1

extern "C" {
  int close(int);
}

namespace Akenti {

class SecureIOHandlerImpl : public AbstractIOHandlerImpl {

public:

  SecureIOHandlerImpl(SSL* ssl) : ssl(ssl) {
  }
 
  ~SecureIOHandlerImpl() {
    close();
    SSL_free(ssl);
  }

  int getDescriptor() const {
    return SSL_get_fd(ssl);
  } 

  bool isOpen() const  {
    return getDescriptor() != -1;
  }

  bool isClosed() const {
    return !isOpen();
  }

  int read(char* buf, int num) {
    return SSL_read(ssl, buf, num);
  }

  int write(const char* buf, int num) {
    return SSL_write(ssl, buf, num);
  }

  // check to see if ssl has a close method
  int close() {
    int fd = getDescriptor();
    int ret = -1;

    if (fd != -1) {
      /**
        * NOTE: We are explicity doing a quiet shutdown on the ssl connection
        * The SSL_shutdown(), otherwise barks!
        */
      SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
      ret = SSL_shutdown(ssl) <= 0 ? -1 : 0;
      ::close(fd);
    }

    return ret;
  }

  string paramString() const {
    ostringstream os;

    os << '['
       << "ssl descriptor"
       << '=' 
       << getDescriptor()
       << ']';

    return os.str();
  }

private:

  SSL* ssl;

};

} // end namespace Akenti

typedef Akenti::OpenSSLSecureSocketImpl OSSSI;

OSSSI::OpenSSLSecureSocketImpl(const SSLContext& cntx) 
  : SecureSocketImpl(cntx) {
  
  OpenSSLContextImpl* 
    impl = dynamic_cast<OpenSSLContextImpl*> (cntx.get());

  if (!impl) {
    throw Error(
      "OpenSSLSecureSocketImpl:constructor:know nothing about this SSLContext"
    );
  }

  if (!(ssl = SSL_new(impl->ctx))) {
    throw Error("OpenSSLSecureSocketImpl:constructor:bad alloc");
  }
}

OSSSI::~OpenSSLSecureSocketImpl() {
  SSL_free(ssl);
}

Akenti::SecureSocketImpl* 
OSSSI::newInstance(const SSLContext& cntx) {
  return new OpenSSLSecureSocketImpl(cntx); 
}

Akenti::CertificateImpl*
OSSSI::doGetPeer() const {
  return new OpenSSLX509CertificateImpl(SSL_get_peer_certificate(ssl));
}

void
OSSSI::doConnect() {
  int sd = openConnection();

  SSL_set_fd(ssl, sd);  

  if (SSL_connect(ssl) <= 0) {
    ::close(sd);

#if DEBUG
    ERR_print_errors_fp(stderr);
    throw SecurityException("SecureSocketImpl:doConnect:handshake failed");
#endif

    long err_code = ERR_get_error();
    string error(250, '\0'); 
    ERR_error_string(err_code, (char *) error.data());
    throw SecurityException("SecureSocketImpl:doConnect:handshake failed:" + error);
  }
  
  ssl->references++;    // :) 
  fd = new SecureIOHandlerImpl(ssl); 
}

void
OSSSI::doAccept(SecureSocketImpl* impl) {
  OpenSSLSecureSocketImpl* 
    temp = dynamic_cast<OpenSSLSecureSocketImpl*> (impl);

  int sd = acceptConnection(temp);
   
  SSL_set_fd(temp->ssl, sd);
   
  if (SSL_accept(temp->ssl) <= 0) {
    ::close(sd);
    
#if DEBUG
    ERR_print_errors_fp(stderr);
    throw SecurityException("SecureSocketImpl:doAccept:handshake failed");
#endif

    long err_code = ERR_get_error();
    string error(250, '\0'); 
    ERR_error_string(err_code, (char *) error.data());
    throw SecurityException("SecureSocketImpl:doAccept:handshake failed:" + error);
  }

  temp->ssl->references++;
  temp->fd = new SecureIOHandlerImpl(temp->ssl);
}

