#include "OpenSSLX509CertificateChainImpl.hpp"
#include "OpenSSLX509CertificateImpl.hpp"
#include "Certificate.hpp"
#include "DistinguishedName.hpp"
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/asn1.h>
#include <openssl/x509v3.h>
#include "Error.hpp"

typedef Akenti::OpenSSLX509CertificateChainImpl OSXCHAIN;

OSXCHAIN::OpenSSLX509CertificateChainImpl(STACK_OF(X509) *chain_arg) :
  chain(chain_arg) {

  if ((chain_arg == NULL) || (sk_X509_num(chain_arg) == 0)) {
    throw Error(
      "OpenSSLX509CertificateChainImpl:constructor:invalid arg)"
    );
  }
}

OSXCHAIN::~OpenSSLX509CertificateChainImpl() {
  int size = sk_X509_num(chain);

  for (int i = 0; i < size; i++) {
    sk_X509_delete(chain, i);
  }
  sk_X509_free(chain);
}

vector<Akenti::Certificate>
OSXCHAIN::getCACertificates() const {
  vector<Certificate> caCerts;

  for (int i = 0, n = sk_X509_num(chain); i < n; i++) {
    X509 *x = sk_X509_value(chain, i);

    BASIC_CONSTRAINTS *bs =
    (BASIC_CONSTRAINTS *) 
    X509_get_ext_d2i( x, NID_basic_constraints, NULL, NULL);

    if (bs != NULL && bs->ca) {
      caCerts.push_back(Certificate(new OpenSSLX509CertificateImpl(x)));
      BASIC_CONSTRAINTS_free(bs);
    }
  }

  return caCerts;
}

Akenti::Certificate
OSXCHAIN::getSubjectCertificate() const {
  OpenSSLX509CertificateImpl *x509Cert = NULL;
  bool found = false;

  for (int i = 0, n = sk_X509_num(chain); !(found) && i < n; i++) {

    bool ignore = false;

    for (int j=0, k=ignoreIdxs.size(); !ignore && j < k; j++) {
      if (i == ignoreIdxs[j]) {
        ignore = true;
      }
    }

    if (ignore) {
      continue;
    }
         
    X509 *x = sk_X509_value(chain, i);

    BASIC_CONSTRAINTS *bs =
    (BASIC_CONSTRAINTS *) 
    X509_get_ext_d2i( x, NID_basic_constraints, NULL, NULL);

    /*
     * NOTE: You need add one more reference to x509 ptr
     *       as it is shared by both the stack and the ref_ptr
     *       Certificate.
     */
    if (bs != NULL) {
      if (!bs->ca) {
        CRYPTO_add(&x->references, 1, CRYPTO_LOCK_X509);
        x509Cert = new OpenSSLX509CertificateImpl(x);
        found = true;
      }
      BASIC_CONSTRAINTS_free(bs);
    } else {
      CRYPTO_add(&x->references, 1, CRYPTO_LOCK_X509);
      x509Cert = new OpenSSLX509CertificateImpl(x);
      found = true;
    }
  }

  return Certificate(x509Cert);
}

void
OSXCHAIN::ignoreCertificates(
  const vector< pair<string,string> >& pairs) {

  for (int i = 0, n = sk_X509_num(chain); i < n; i++) {
    X509 *x = sk_X509_value(chain, i);

    string name(1024, '\0' );

    X509_NAME_oneline(
    X509_get_subject_name(x), (char *) name.data(), name.length());

    BASIC_CONSTRAINTS *bs =
    (BASIC_CONSTRAINTS *) 
    X509_get_ext_d2i( x, NID_basic_constraints, NULL, NULL);

    if (bs == NULL || !(bs->ca)) {

      DistinguishedName dn(name.c_str());

      bool found = false;

      for (int j=0, k=pairs.size(); !found && j < k; j++) {
        if (dn.containsAttributeValue(
              pairs[j].first, pairs[j].second)) {
          ignoreIdxs.push_back(i);
          found = true;
        }
      }
    }
  }
}

Akenti::DistinguishedName
OSXCHAIN::getGuarantor() const {
  if (getSubjectCertificate()) {
    // XXX
  }
  return getSubjectCertificate()->getGuarantor();
}

Akenti::DistinguishedName
OSXCHAIN::getPrincipal() const {
  if (getSubjectCertificate()) {
    // XXX
  }
  return getSubjectCertificate()->getPrincipal();
}

string
OSXCHAIN::paramString() const {
  return "";
}
