#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/asn1.h>
#include "OpenSSLX509CertificateImpl.hpp"
#include "OpenSSLPublicKeyImpl.hpp"
#include "SecurityException.hpp"
#include "Error.hpp"
#include "PublicKey.hpp"
#include "DistinguishedName.hpp"
#include "Signature.hpp"
#include "UtcTime.hpp"

typedef Akenti::OpenSSLX509CertificateImpl OSXCI;

OSXCI::OpenSSLX509CertificateImpl(X509 *x509_arg) : x509(x509_arg) {
  if (!x509_arg) {
    throw Error("OpenSSLX509CertificateImpl:constructor:invalid arg");
  }
}

OSXCI::~OpenSSLX509CertificateImpl() {
  X509_free(x509); 
} 

string 
OSXCI::getType() const {
  return string("X509");
}

string 
OSXCI::getFormat() const {
  return string("DER"); 
}

Akenti::DistinguishedName
OSXCI::getGuarantor() const {
  // XXX
  string name(200, '\0');
 
  X509_NAME_oneline(
    X509_get_issuer_name(x509), 
    (char *) name.data(), 
    name.length()
  );

  return DistinguishedName(name.c_str());
}

Akenti::DistinguishedName 
OSXCI::getPrincipal() const {
  // XXX
  string name(200, '\0');
 
  X509_NAME_oneline(
    X509_get_subject_name(x509), 
    (char *) name.data(), 
    name.length()
  );

  return DistinguishedName(name.c_str());
}

Akenti::PublicKey
OSXCI::getSubjectPublicKey() const {
  return PublicKey( 
    new OpenSSLPublicKeyImpl(X509_get_pubkey(x509))
  );
}

long
OSXCI::getVersionNumber() const {
  return X509_get_version(x509); 
}

long
OSXCI::getSerialNumber() const {
 return ASN1_INTEGER_get(X509_get_serialNumber(x509)); 
}

Akenti::UtcTime
OSXCI::getNotBefore() const {
  ASN1_UTCTIME *utctime = X509_get_notBefore(x509);

  return UtcTime((char*) utctime->data);
}

Akenti::UtcTime
OSXCI::getNotAfter() const {
  ASN1_UTCTIME *utctime = X509_get_notAfter(x509);

  return UtcTime((char*) utctime->data);
}

string 
OSXCI::getSignature() const {
  return string(
    (char*) x509->signature->data, 
    x509->signature->length
  );
}

string 
OSXCI::getSignable() const {
  int len = i2d_X509_CINF(x509->cert_info, NULL);
  string signable = string(len, '\0');
  unsigned char *data = (unsigned char *) signable.data();

  i2d_X509_CINF(x509->cert_info, &data);
  return signable;
}

string 
OSXCI::getSignatureAlgorithm() const {
  return string(
    OBJ_nid2sn(OBJ_obj2nid(x509->sig_alg->algorithm))
  );
}

bool
OSXCI::verify(const PublicKey& publicKey) const {
  Signature sig = Signature::getInstance(getSignatureAlgorithm());

  sig->initVerify(publicKey);
  sig->update(getSignable());
  return sig->verify(getSignature());
}

string 
OSXCI::paramString() const {
  // XXX
  return toText();
}

long
OSXCI::hashCode() const {
  // XXX
  return X509_subject_name_hash(x509);
}

string 
OSXCI::getEncoded() const {
  return getEncoded(getFormat());
}

string 
OSXCI::getRevocationURL() const {
  throw Error("OpenSSLX509CertificateImpl:getRevocationURL:To be implemented");
}

string
OSXCI::getEncoded(const string& type) const {
  if (type == "DER") {
    int len = i2d_X509(x509, NULL);

    if (len <= 0) {
      throw Error("OpenSSLX509CertificateImpl:getEncoded:should never happen");
    }

    string encoding(len, '\0');
    unsigned char *temp = (unsigned char *) encoding.data();

    i2d_X509(x509, &temp);
    return encoding;
  } else if (type == "PEM") {
    BIO* bio = BIO_new(BIO_s_mem());

    PEM_write_bio_X509(bio, x509);

    char *pem = NULL;
    long len = BIO_get_mem_data(bio, &pem);
    string encoding(pem, len);

    // only bio needs to be freed :)
    BIO_free(bio);
    return encoding;
  }

  throw SecurityException("OpenSSLX509CertificateImpl:getEncoded:unsupported type:" + type);
}

string 
OSXCI::getHashString() const {
  X509_NAME* name = X509_get_subject_name(x509);
  int len = i2d_X509_NAME(name, NULL);

  if (len <= 0) {
    throw Error("OpenSSLX509CertificateImpl:getHashString:should never happen");
  }

  string ret(len, '\0');
  unsigned char *temp = (unsigned char *) ret.data();
  
  i2d_X509_NAME(name, &temp);
  return ret;
}

string 
OSXCI::toText() const {
  BIO* bio = BIO_new(BIO_s_mem());

  X509_print(bio, x509);

  char *text = NULL;
  long len = BIO_get_mem_data(bio, &text);
  string ret(text, len);

  BIO_free(bio);
  return ret;
}

const X509* 
OSXCI::getX509() const {
  return x509;
}
