#ifndef OPEN_SSL_X509_CERTIFICATE_IMPL_HPP
#define OPEN_SSL_X509_CERTIFICATE_IMPL_HPP

#include <openssl/x509.h>
#include "CertificateImpl.hpp"

namespace Akenti {

/**
  * This class is the subclass of <i>CertificateImpl</i>
  * and it provides all the services the opensl X509 
  * certificate supports. 
  *
  * This class contains a openssl X509 structure and basically acts as 
  * a wrapper to functions that act on that structure.
  * Within the Akenti code these certificates (actually a ref_ptr to the
  * certificateImpl) is kept in an AkentiX509CertificateImpl which is a
  * subclass of AkentiCertificateImpl and supports all the methods of the
  * superclass. Thus Akenti code normally refers to the AkentiX509Certificate 
  * rather than this class. Methods in this class are called by the 
  * AkentiX509CertificateImpl. 
  * 
  * As this class is part of the shared library, it may be used directly by
  * code other than Akenti.
  *
  * @see CertificateImpl
  * @see PublicKey
  * @see UtcTime
  * @see DistinguishedName 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */
class OpenSSLX509CertificateImpl : public CertificateImpl {

public:

  /**
    * Constructs this object takes <i>X509</i> as type .
    */ 
  OpenSSLX509CertificateImpl(X509 *x509);

  /**
    * Deletes this object.
    */ 
  ~OpenSSLX509CertificateImpl();

  /**
    * Returns the type <i>X509</i> of this certificate.
    *
    * @see param x509
    */  
  string getType() const;

  /**
    * Returns <i>publicKey</i> of this certificate.
    */
  PublicKey getSubjectPublicKey() const;

  /**
    * Returns the <i>versionNumber</i> of this certificate. 
    */
  long getVersionNumber() const;

  /**
    * Returns the earliest time that the certificate is valid.
    */ 
  UtcTime getNotBefore() const;

  /**
    * Returns the latest time that the certificate is valid.
    */
  UtcTime getNotAfter() const;

  /**
    * Returns the <i>serialNumber</i> of this certificate, which together with
    * the issuing CA's name, uniquely identifies this certificate.
    */
  long getSerialNumber() const;

  /**
    * Returns the <i>signature</i>  of this certificate. 
    */
  string getSignature() const;

  /**
    * Returns <i>signatureAlgorithm</i> of this certificate.
    */ 
  string getSignatureAlgorithm() const;

  /**
    * Returns <i>format</i> of this certificate.
    */ 
  string getFormat() const;

  /**
    * Returns <i>CA</i> that signed this certificate.
    */ 
  DistinguishedName getGuarantor() const;

  /**
    * Returns <i>subject</i> of this certificate.
    */ 
  DistinguishedName getPrincipal() const;

  /**
    * Returns <i>RevocationURL</i> for the CA that issued this certificate.
    */ 
  string getRevocationURL() const;

  /**
    * Returns the <i>PublicKey</i> stored in this certificate.
    */ 
  bool verify(const PublicKey& publicKey) const;

  /**
    * Returns a string represenation of this certificate.
    */ 
  string paramString() const;

  /**
    *
    */ 
  long hashCode() const;

  /**
    * Returns the part of this certificate that is signed.
    */ 
  string getSignable() const;

  /**
    * Returns the DER encoding for this certificate
    */ 
  string getEncoded() const;

  /**
    * Returns either the DER or PEM encoding of the certificate depending on
    * the value of type.
    */ 
  string getEncoded(const string& type) const;

  /**
    */
  string getHashString() const;

  /**
    */
  string toText() const;

private:

  friend class OpenSSLContextImpl;

  const X509* getX509() const;

  X509 *x509;

};

} // end namespace Akenti

#endif
