#ifndef AK_CERTIFICATE_IMPL_HPP
#define AK_CERTIFICATE_IMPL_HPP

#include <string>
#include "UtcTime.hpp"

using std::string;

namespace Akenti {

class PublicKey;
class DistinguishedName;

/**
  * This abstract class defines an interface to a generic signed certificate.
  * A certficate is constant once it has been generated, so this class only 
  * defines methods to get fields of the certificate such as type, 
  * versionNumber, serialNumber, signature that a <i> certificate</i> 
  * supports. A certificate is created by the CertificateGenerator class. 
  * OpenSSLX509Certificate is currently the only subclass of this class. 
  * The rest of the Akenti certificates are subclasses of AkentiCertificateImpl.
  *
  * @see UtcTime
  * @see PublicKey
  * @see DistinguishedName
  * @see Signature
  * @see AkentiCertificateImpl
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class CertificateImpl {

public:

  /**
    * Deletes this object.
    */ 
  virtual ~CertificateImpl() {
  }

  /**
    * Returns the type of this certificate.  
    */
  virtual string getType() const = 0;

  /**
    */
  virtual long hashCode() const = 0;

  /**
    * Encodes this certificate.
    */ 
  virtual string getEncoded() const = 0;

  /**
    * Returns the subject's publicKey of this certificate.
    */
  virtual PublicKey getSubjectPublicKey() const = 0;

  /**
    * Returns bool value to indicate the correctness of this certificate.
    */ 
  virtual bool verify(const PublicKey& publicKey) const = 0;

  /**
    * Returns format of this certificate.
    */
  virtual string getFormat() const = 0;

  /**
    * Returns guarantor of this certificate. 
    */
  virtual DistinguishedName getGuarantor() const = 0;

  /**
    * Returns principal of this certificate.
    */
  virtual DistinguishedName getPrincipal() const = 0;

  /** 
    * Returns serialNumber of this certificate.
    */
  virtual long getSerialNumber() const = 0;

  /**
    * Returns signature of this certificate.
    */
  virtual string getSignature() const = 0;

  /**
    * Returns the information that you need 
    * to sign.
    */
  virtual string getSignable() const = 0;

  /**
    * Returns the signature's algorithm name of this certificate.
    */  
  virtual string getSignatureAlgorithm() const = 0;

  /**
    * Returns the versionNumber of this certificate.
    */
  virtual long getVersionNumber() const = 0;

  /**
    * Returns GMT time for the beginning time of valid usage
    * of this certificate.
    */ 
  virtual UtcTime getNotBefore() const = 0;

  /**
    */
  virtual string getRevocationURL() const = 0;

  /**
    * Returns GMT time for the ending time of valid usage
    * of this certificate.
    */
  virtual UtcTime getNotAfter() const = 0;

  /**
    * Checks validity of this certificate comparing to current time.
    */
  bool checkValidity() const {
    return checkValidity(UtcTime());
  }
 
  /**
    * Checks validity of this certificate comparing to given time.
    */
  bool checkValidity(const UtcTime& utcTime) const {
    return utcTime >= getNotBefore() && utcTime <= getNotAfter();
  }

  /**
    */
  virtual string getHashString() const = 0;

  /**
    */
  virtual string toText() const = 0;

  /**
    * Returns the string of items related to this certificate
    * for correctness check.
    */
  virtual string paramString() const = 0;

protected:

  /**
    * Constructs this object.
    */
  CertificateImpl() {
  }

};

} // end of namespace Akenti

#endif
