#ifndef AK_SECURE_SOCKET_IMPL_HPP
#define AK_SECURE_SOCKET_IMPL_HPP

#include "SocketImpl.hpp"
#include "SSLContext.hpp"
#include "Certificate.hpp"

/**
  * Base abstract class for SecureSocket implementations - inherits 
  * from SocketImpl and adds a security context.
  *
  * @version 1.1 00/05/01 
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class SecureSocketImpl : public SocketImpl {

public:

  /**
    * Destructor
    */
  virtual ~SecureSocketImpl() {
  }

  /**
    * Wraps the underlying socket connection.
    */
  virtual void doConnect() = 0;

  /**
    * Wraps the underlying socket accept
    */
  virtual void doAccept(SecureSocketImpl* impl) = 0;

  /**
    *  Returns the security context associated with this SecureSocket.
    * 
    *  @return ctx - the SSLContext associated with the SecureSocket.
    */
  const SSLContext& getSSLContext() const {
    return ctx;
  }

  Certificate getPeerCertificate() const {
    return Certificate(doGetPeer());
  }

  /**
    * Creates a new instance of a concrete SecureSocketImpl. Used instead of 
    * a constructor in order to allow code to create a SecureSocketImpl without
    * knowing the concrete type. 
    *
    * @param ctx - the security context to be associated with the secure socket.
    * @return - a new SecureSocketImpl of the same time as this one.
    * @exception - Error if the SSLContext is not recognized.
    */
  virtual SecureSocketImpl* newInstance(const SSLContext& ctx) = 0;

  /**
    * Returns a string representation for SecureSocketImpl which is used for
    * debugging purposes.
    *
    * @return debugging info
    */
  virtual string paramString() const {
    const SSLContext& cntx = getSSLContext();

    return cntx->paramString();
  }


protected:

  /**
    * Construct a <i>SecureSocketImpl</i> given the <i>SSLContext</i>.
    *
    * @exception Error if cntx is not valid
    */ 
  SecureSocketImpl(const SSLContext& cntx) {
    if (!cntx) {
      throw Error("SecureSocketImpl:constructor:invalid arg");
    }
    
    this->ctx = cntx;
  }

  virtual CertificateImpl* doGetPeer() const = 0;

  SSLContext ctx;

};

} // end of namespace Akenti

#endif

